/**
 * \file landmark_poliyline_2D.h
 *
 *  Created on: May 26, 2016
 *      \author: jvallve
 */

#ifndef LANDMARK_POLYLINE_2D_H_
#define LANDMARK_POLYLINE_2D_H_

// Wolf
#include "core/landmark/landmark_base.h"
#include "laser/processor/polyline_2D_utils.h"

// yaml-cpp library
#include <yaml-cpp/yaml.h>

namespace wolf
{

//landmark classification
typedef enum
{
    UNCLASSIFIED,
    CONTAINER,        ///< A container 12.2 x 2.44 (m)
    SMALL_CONTAINER,  ///< A small container 6.1 x 2.44 (m)
    PALLET,           ///< A pallet box 0.9 x 1.2 (m)
    OTHER             ///< To be used for custom rectangles
} PolylineClassType;

struct PolylineRectangularClass
{
        PolylineClassType type; // name of the type
        Scalar L; // long side length
        Scalar W; // short side length
        Scalar D; // diagonal length

        PolylineRectangularClass() :
            type(UNCLASSIFIED),
            L(0),
            W(0),
            D(0)
        {
        };

        PolylineRectangularClass(const PolylineClassType& _type, const Scalar& _L, const Scalar& _W) :
            type(_type),
            L(_L),
            W(_W),
            D(sqrt(L*L+W*W))
        {
        };
};
struct PolylineClassContainer : public PolylineRectangularClass
{
        PolylineClassContainer() : PolylineRectangularClass(CONTAINER, 12.2, 2.44){};
};
struct PolylineClassSmallContainer : public PolylineRectangularClass
{
        PolylineClassSmallContainer() : PolylineRectangularClass(SMALL_CONTAINER, 6.1, 2.44){};
};
struct PolylineClassPallet : public PolylineRectangularClass
{
        PolylineClassPallet() : PolylineRectangularClass(PALLET, 1.2, 0.9){};
};

WOLF_PTR_TYPEDEFS(LandmarkPolyline2D);
WOLF_LIST_TYPEDEFS(LandmarkPolyline2D);

//class
class LandmarkPolyline2D : public LandmarkBase
{
    protected:
        std::map<int,StateBlockPtr> point_state_ptr_map_; ///< polyline points state blocks
        bool first_defined_;            ///< Wether the first point is an extreme of a line or the line may continue
        bool last_defined_;             ///< Wether the last point is an extreme of a line or the line may continue
        bool closed_;                   ///< Wether the polyline is closed or not
        PolylineRectangularClass classification_; ///< The classification of the landmark
        int version_;                   ///< Integer increased each time a modification in landmark occurs (added points, merged points, closed, classified)
        LandmarkPolyline2DPtr merged_in_lmk_;

    private:

        StateBlockPtr& firstStateBlock();
        StateBlockPtr& lastStateBlock();

    public:
        LandmarkPolyline2D(StateBlockPtr _p_ptr, StateBlockPtr _o_ptr, const Eigen::MatrixXs& _points, const bool _first_defined, const bool _last_defined, unsigned int _first_id = 0);
        LandmarkPolyline2D(StateBlockPtr _p_ptr, StateBlockPtr _o_ptr, const Eigen::MatrixXs& _points, const bool _first_defined, const bool _last_defined, unsigned int _first_id, PolylineRectangularClass _class);
        virtual ~LandmarkPolyline2D();

        /** \brief Gets a const reference to the point state block pointer vector
         **/
        std::map<int,StateBlockPtr>& getPointStatePtrMap();


        /** \brief Gets wether the first/last points are defined or not
         **/
        bool isFirstDefined() const;
        bool isLastDefined() const;

        /** \brief Gets whether the polyline is closed or not
         **/
        bool isClosed() const;

        /** \brief Get the landmark pointer in which was merged (nullptr in case of no merged)
         **/
        LandmarkPolyline2DPtr getMergedInLandmark() const;

        /** \brief Set the landmark pointer in which was merged
         **/
        void setMergedInLandmark(const LandmarkPolyline2DPtr& lmk_ptr);

        /** \brief Gets whether the given state block point is defined or not (assumes the state block is in the landmark)
         **/
        bool isDefined(StateBlockPtr _state_block) const;

        /** \brief Sets the first/last extreme point
         **/
        void setFirst(const Eigen::VectorXs& _point, bool _defined);
        void setLast(const Eigen::VectorXs& _point, bool _defined);

        int getNPoints() const;
        int getNDefinedPoints() const;
        std::vector<int> getValidIds() const;
        int id2idx(int) const;
		int getFirstId() const;
		int getLastId() const;
		bool isValidId(const int& i) const;
		int getNextValidId(const int& i) const;
		int getPrevValidId(const int& i) const;
		int getVersion() const;

        const Eigen::VectorXs getPointVector(int _i) const;

        Eigen::MatrixXs computePointsGlobal() const;
        Eigen::MatrixXs computePointsGlobal(const int& _from_id, const int& _to_id) const;
        Eigen::MatrixXs computePointsCovGlobal() const;

        StateBlockPtr getPointStateBlock(int _i);

        /** \brief Gets a vector of all state blocks pointers
         **/
        virtual std::vector<StateBlockPtr> getPointsStateBlockVector() const;

        /** \brief Adds a new point to the landmark
         * \param _point: the point to be added
         * \param _extreme: if its extreme or not
         * \param _back: if it have to be added in the back (true) or in the front (false)
         **/
        void addPoint(const Eigen::VectorXs& _point, const bool& _defined, const bool& _back);

        /** \brief Adds new points to the landmark
         * \param _points: a matrix containing points, some of them to be added
         * \param _idx: from wich position of the first point to be added
         * \param _extreme: if last point to be added is extreme or not
         * \param _back: if the points have to be added in the back (true) or in the front (false)
         **/
        void addPoints(const Eigen::MatrixXs& _points, const unsigned int& _idx, const bool& _defined, const bool& _back);

        /** \brief define extreme point
         **/
        virtual void defineExtreme(const bool _back);

        /** \brief define first point
         **/
        virtual void defineFirst();

        /** \brief define last point
         **/
        virtual void defineLast();

        /** \brief Try to close the polyline (if it has overlapped points)
         **/
        virtual bool tryClose(const Scalar& _dist_tol);

        /** \brief Set the polyline as closed
         **/
        virtual void setClosed();

        /** \brief merge points
         **/
        void mergePoints(int _remove_id, int _remain_id);

        /** \brief Classify as a known object
         **/
        void classify(PolylineRectangularClass _class);

        /** \brief Try to classify polyline as a known object
         **/
        bool tryClassify(const Scalar& _dist_tol, std::vector<PolylineRectangularClass> _classes);

        /** \brief Try to classify polyline as a known object
         **/
        void mergeLandmark(const LandmarkPolyline2DPtr _merged_lmk, const int& _from, const int& _to, const int& _merged_from, const int& _merged_to);

        /** \brief get classification
         **/
        PolylineRectangularClass getClassification() const;

        /** \brief Adds all stateBlocks of the frame to the wolfProblem list of new stateBlocks
         **/
//        virtual void registerNewStateBlocks();
//        virtual void removeStateBlocks();

        /** Factory method to create new landmarks from YAML nodes
         */
        static LandmarkBasePtr create(const YAML::Node& _lmk_node);

        YAML::Node saveToYaml() const;

        static void tryMergeLandmarks(LandmarkPolyline2DPtrList& _lmk_list,
                                      const Scalar& max_sq_error,
                                      const Scalar& min_overlapping_dist,
                                      const int& min_N_overlapped,
                                      const int& min_N_defined_overlapped);
};

// STL
#include <iterator>

inline StateBlockPtr& LandmarkPolyline2D::firstStateBlock()
{
    return point_state_ptr_map_.begin()->second;
}

inline StateBlockPtr& LandmarkPolyline2D::lastStateBlock()
{
    return point_state_ptr_map_.rbegin()->second;
}

inline std::map<int,StateBlockPtr>& LandmarkPolyline2D::getPointStatePtrMap()
{
    return point_state_ptr_map_;
}

inline bool LandmarkPolyline2D::isFirstDefined() const
{
    return first_defined_;
}

inline bool LandmarkPolyline2D::isLastDefined() const
{
    return last_defined_;
}

inline bool LandmarkPolyline2D::isValidId(const int& i) const
{
    return point_state_ptr_map_.count(i) != 0;
};

inline bool LandmarkPolyline2D::isClosed() const
{
    return closed_;
}

inline LandmarkPolyline2DPtr LandmarkPolyline2D::getMergedInLandmark() const
{
    // recursive call
    if (merged_in_lmk_ != nullptr && merged_in_lmk_->getMergedInLandmark() != nullptr)
        return merged_in_lmk_->getMergedInLandmark();
    return merged_in_lmk_;
}

inline bool LandmarkPolyline2D::isDefined(StateBlockPtr _state_block) const
{
    if (_state_block == point_state_ptr_map_.begin()->second)
        return first_defined_;

    if (_state_block == point_state_ptr_map_.rbegin()->second)
        return last_defined_;

    return true;
}

inline int LandmarkPolyline2D::getNPoints() const
{
    return (int)point_state_ptr_map_.size();
}

inline int LandmarkPolyline2D::getNDefinedPoints() const
{
    return (int)point_state_ptr_map_.size() - (first_defined_ ? 0 : 1) - (last_defined_ ? 0 : 1);
}

inline std::vector<int> LandmarkPolyline2D::getValidIds() const
{
    std::vector<int> valid_ids;
    for (auto st_pair : point_state_ptr_map_)
       valid_ids.push_back(st_pair.first);

    return valid_ids;
}

inline int LandmarkPolyline2D::id2idx(int id) const
{
    assert(isValidId(id));
    return std::distance(point_state_ptr_map_.begin(),point_state_ptr_map_.find(id));
}

inline int LandmarkPolyline2D::getFirstId() const
{
	return point_state_ptr_map_.begin()->first;
}

inline int LandmarkPolyline2D::getLastId() const
{
	return point_state_ptr_map_.rbegin()->first;
}


inline int LandmarkPolyline2D::getNextValidId(const int& i) const
{
    assert(!(i == getLastId() && !closed_) && "Calling getNextValidId of last_id in an open landmark");
    assert(isValidId(i) && "Calling getNextValidId of an invalid id");

    if (i == getLastId() && closed_)
        return getFirstId();
    return std::next(point_state_ptr_map_.find(i))->first;
}

inline int LandmarkPolyline2D::getPrevValidId(const int& i) const
{
    assert(!(i == getFirstId() && !closed_) && "Calling getPrevValidId of first_id in an open landmark");
    assert(isValidId(i) && "Calling getPrevValidId of an invalid id");

    if (i == getFirstId() && closed_)
        return getLastId();
    return std::prev(point_state_ptr_map_.find(i))->first;
}

inline int LandmarkPolyline2D::getVersion() const
{
    return version_;
}

inline std::vector<StateBlockPtr> LandmarkPolyline2D::getPointsStateBlockVector() const
{
    std::vector<StateBlockPtr> state_block_vec;
    for (auto st_pair : point_state_ptr_map_)
        state_block_vec.push_back(st_pair.second);

    return state_block_vec;
}

inline void LandmarkPolyline2D::classify(PolylineRectangularClass _class)
{
    classification_ = _class;
}

inline PolylineRectangularClass LandmarkPolyline2D::getClassification() const
{
    return classification_;
}

inline void LandmarkPolyline2D::setMergedInLandmark(const LandmarkPolyline2DPtr& lmk_ptr)
{
    merged_in_lmk_ = lmk_ptr;
}

} /* namespace wolf */

#endif /* LANDMARK_POLYLINE_2D_H_ */
