#include "laser/feature/feature_match_polyline_2D.h"
#include "laser/feature/feature_polyline_2D.h"
#include "laser/processor/polyline_2D_utils.h"

namespace wolf {

bool FeatureMatchPolyline2D::fromDefined() const
{
    assert(pl_last_ != nullptr);

    return (feature_last_from_id_      != 0 || pl_last_->isFirstDefined()) &&
           (feature_incoming_from_id_  != 0 || pl_incoming_->isFirstDefined());
}

bool FeatureMatchPolyline2D::toDefined() const
{
    assert(pl_incoming_ != nullptr);
    assert(pl_last_ != nullptr);

    return (feature_last_to_id_     != pl_last_->getNPoints()-1     || pl_last_->isLastDefined())  &&
           (feature_incoming_to_id_ != pl_incoming_->getNPoints()-1 || pl_incoming_->isLastDefined());
}

unsigned int FeatureMatchPolyline2D::getNPoints() const
{
    return feature_incoming_to_id_ - feature_incoming_from_id_ + 1;
}

unsigned int FeatureMatchPolyline2D::getNDefinedPoints() const
{
    return getNPoints() - (fromDefined() ? 0:1) - (toDefined() ? 0:1);
}

bool FeatureMatchPolyline2D::check(bool check_partial_match) const
{
    // feature last polyline type
    if (pl_last_ == nullptr)
    {
        WOLF_ERROR("FeatureMatchPolyline2D feature last polyline 2D nullptr");
        return false;
    }
    if (pl_last_->isRemoving())
    {
        WOLF_ERROR("FeatureMatchPolyline2D removed feature last");
        return false;
    }
    // feature polyline type
    if (pl_incoming_ == nullptr)
    {
        WOLF_ERROR("FeatureMatchPolyline2D feature incoming polyline 2D nullptr");
        return false;
    }
    if (pl_incoming_->isRemoving())
    {
        WOLF_ERROR("FeatureMatchPolyline2D removed feature incoming");
        return false;
    }
    // Check idxs
    if (feature_last_from_id_ > feature_last_to_id_)
    {
        WOLF_ERROR("FeatureMatchPolyline2D feature_last_from_id_ > feature_last_to_id_");
        return false;
    }
    if (feature_last_from_id_ < 0)
    {
        WOLF_ERROR("FeatureMatchPolyline2D feature_last_from_id_ < 0");
        return false;
    }
    if (feature_last_to_id_ > pl_last_->getNPoints()-1)
    {
        WOLF_ERROR("FeatureMatchPolyline2D feature_last_to_id_ < pl_last_->getNPoints()-1");
        return false;
    }
    if (feature_incoming_from_id_ > feature_incoming_to_id_)
    {
        WOLF_ERROR("FeatureMatchPolyline2D feature_incoming_from_id_ > feature_incoming_to_id_");
        return false;
    }
    if (feature_incoming_from_id_ < 0)
    {
        WOLF_ERROR("FeatureMatchPolyline2D feature_incoming_from_id_ < 0");
        return false;
    }
    if (feature_incoming_to_id_ > pl_incoming_->getNPoints()-1)
    {
        WOLF_ERROR("FeatureMatchPolyline2D feature_incoming_to_id_ < pl_incoming_->getNPoints()-1");
        return false;
    }
    // partial match
    if (check_partial_match)
    {
        if (feature_incoming_from_id_ != 0 && feature_last_from_id_ != 0)
        {
            WOLF_ERROR("FeatureMatchPolyline2D partial match in front");
            return false;
        }
        if (feature_incoming_to_id_ != pl_incoming_->getNPoints()-1 && feature_last_to_id_ != pl_last_->getNPoints()-1)
        {
            WOLF_ERROR("FeatureMatchPolyline2D partial match in back");
            return false;
        }
    }
    // correspondences
    if (feature_last_to_id_ - feature_last_from_id_  != feature_incoming_to_id_ - feature_incoming_from_id_)
    {
        WOLF_ERROR("FeatureMatchPolyline2D feature_last_to_id_ - feature_last_from_id_  != feature_incoming_to_id_ - feature_incoming_from_id_");
        return false;
    }

    return true;
}

Eigen::Array<Scalar,Eigen::Dynamic,1> FeatureMatchPolyline2D::computeSqDistArray() const
{
    assert(pl_incoming_ != nullptr);
    assert(pl_last_ != nullptr);

    return computeSqDist(T_incoming_last_ * pl_last_->getPoints(),                                                   //_points_A
                         feature_last_from_id_,                                                                      //from_A,
                         feature_last_to_id_,                                                                        //to_A,
                         feature_last_from_id_ != 0                        || pl_last_->isFirstDefined(),            //from_A_defined,
                         feature_last_to_id_   != pl_last_->getNPoints()-1 || pl_last_->isLastDefined(),             //to_A_defined,
                         pl_incoming_->getPoints(),                                                                  //_points_B
                         feature_incoming_from_id_,                                                                  //from_B,
                         feature_incoming_to_id_,                                                                    //to_B,
                         feature_incoming_from_id_ != 0                            || pl_incoming_->isFirstDefined(),//from_B_defined,
                         feature_incoming_to_id_   != pl_incoming_->getNPoints()-1 || pl_incoming_->isLastDefined(), //to_B_defined,
                         pl_last_->getNPoints()-1,                                                                   //last_A,
                         getNPoints());                                                                              //N_overlapped
}

void FeatureMatchPolyline2D::print() const
{
    assert(pl_incoming_ != nullptr);
    assert(pl_last_ != nullptr);

    std::cout << std::endl << "POLYLINE FEATURE MATCH" << std::endl;
    std::cout << "INCOMING: " << pl_incoming_->id() << ": " << std::endl;
    std::cout << "\tpoints " << pl_incoming_->getNPoints() << std::endl;
    std::cout << "\tfirst defined " << pl_incoming_->isFirstDefined() << std::endl;
    std::cout << "\tlast defined " << pl_incoming_->isLastDefined() << std::endl;
    std::cout << "\tpoints \n" << pl_incoming_->getPoints() << std::endl;

    std::cout << "LAST " << pl_last_->id() << ": " << std::endl;
    std::cout << "\tpoints " << pl_last_->getNPoints() << std::endl;
    std::cout << "\tfirst defined " << pl_last_->isFirstDefined() << std::endl;
    std::cout << "\tlast defined " << pl_last_->isLastDefined() << std::endl << std::endl;
    std::cout << "\tpoints \n" << pl_last_->getPoints() << std::endl;
    std::cout << "\tpoints (local) \n" << T_incoming_last_ * pl_last_->getPoints() << std::endl;

    std::cout << "MATCH " << std::endl;
    std::cout << "\tfrom feature incoming point " << feature_incoming_from_id_ << std::endl;
    std::cout << "\tto feature  incoming point " << feature_incoming_to_id_ << std::endl;
    std::cout << "\tfrom feature last point " << feature_last_from_id_ << std::endl;
    std::cout << "\tto feature last point " << feature_last_to_id_ << std::endl;
    std::cout << "\tT_incoming_last_ " << std::endl << T_incoming_last_ << std::endl;
}

} // namespace wolf
