/*
 * landmark_match_polyline_2D.cpp
 *
 *  Created on: Apr 2, 2019
 *      Author: jvallve
 */

#include "laser/landmark/landmark_match_polyline_2D.h"
#include "laser/landmark/landmark_polyline_2D.h"
#include "laser/feature/feature_polyline_2D.h"
#include "laser/processor/polyline_2D_utils.h"

namespace wolf {



bool LandmarkMatchPolyline2D::fromDefined() const
{
    assert(pl_landmark_ != nullptr);

    return (landmark_from_id_ != pl_landmark_->getFirstId() || pl_landmark_->isFirstDefined()) &&
           (feature_from_id_  != 0                          || pl_landmark_->isFirstDefined());
}

bool LandmarkMatchPolyline2D::toDefined() const
{
    assert(pl_feature_ != nullptr);
    assert(pl_landmark_ != nullptr);

    return (landmark_to_id_   != pl_landmark_->getLastId()     || pl_landmark_->isLastDefined())  &&
           (feature_to_id_    != pl_feature_->getNPoints()-1   || pl_feature_->isLastDefined());
}

unsigned int LandmarkMatchPolyline2D::getNPoints() const
{
    return feature_to_id_ - feature_from_id_ + 1;
}

unsigned int LandmarkMatchPolyline2D::getNDefinedPoints() const
{
    return getNPoints() - (fromDefined() ? 0:1) - (toDefined() ? 0:1);
}

bool LandmarkMatchPolyline2D::isDeprecated() const
{
    assert(pl_landmark_ != nullptr);

    return pl_landmark_->isRemoving() || landmark_version_ != pl_landmark_->getVersion() || pl_landmark_->getMergedInLandmark() != nullptr;
}

bool LandmarkMatchPolyline2D::isPartialFront() const
{
    assert(pl_landmark_ != nullptr);

    // exceeding points if: feature point is not the extreme && either landmark point is not the extreme or landmark is closed
    return (feature_from_id_ != 0 && (landmark_from_id_ != pl_landmark_->getFirstId() || pl_landmark_->isClosed()));
}

bool LandmarkMatchPolyline2D::isPartialBack() const
{
    assert(pl_feature_ != nullptr);
    assert(pl_landmark_ != nullptr);

    // exceeding points if: feature point is not the extreme && either landmark point is not the extreme or landmark is closed
    return (feature_to_id_ != pl_feature_->getNPoints()-1 && (landmark_to_id_ != pl_landmark_->getLastId() || pl_landmark_->isClosed()));
}


bool LandmarkMatchPolyline2D::check(bool check_partial_match) const
{
    // landmark polyline type
    if (pl_landmark_ == nullptr)
    {
        WOLF_ERROR("LandmarkMatchPolyline2D landmark polyline 2D nullptr");
        return false;
    }
    if (isDeprecated())
    {
        WOLF_ERROR("LandmarkMatchPolyline2D landmark polyline 2D was merged or changed the version");
        return false;
    }
    if (pl_landmark_->isRemoving())
    {
        WOLF_ERROR("LandmarkMatchPolyline2D removed landmark");
        return false;
    }
    // feature polyline type
    if (pl_feature_ == nullptr)
    {
        WOLF_ERROR("LandmarkMatchPolyline2D feature polyline 2D nullptr");
        return false;
    }
    if (pl_feature_->isRemoving())
    {
        WOLF_ERROR("LandmarkMatchPolyline2D removed feature");
        return false;
    }
    // Check idxs
    if (!pl_landmark_->isClosed() && landmark_from_id_ > landmark_to_id_)
    {
        WOLF_ERROR("LandmarkMatchPolyline2D lmk_from > lmk_to with a not closed lmk");
        return false;
    }
    if (!pl_landmark_->isClosed() && landmark_from_id_ < pl_landmark_->getFirstId())
    {
        WOLF_ERROR("LandmarkMatchPolyline2D lmk_from < lmk_first with a not closed lmk");
        return false;
    }
    if (!pl_landmark_->isClosed() && landmark_to_id_ > pl_landmark_->getLastId())
    {
        WOLF_ERROR("LandmarkMatchPolyline2D lmk_to < lmk_last with a not closed lmk");
        return false;
    }
    // partial match
    if (check_partial_match)
    {
        if (isPartialFront())
        {
            WOLF_ERROR("LandmarkMatchPolyline2D partial match in front");
            return false;
        }
        if (isPartialBack())
        {
            WOLF_ERROR("LandmarkMatchPolyline2D partial match in back");
            return false;
        }
    }
    // N points
    int landmark_to_idx     = pl_landmark_->id2idx(landmark_to_id_);
    int landmark_from_idx   = pl_landmark_->id2idx(landmark_from_id_);
    int N_landmark = (landmark_to_idx >= landmark_from_idx ?
                        landmark_to_idx - landmark_from_idx + 1 :
                        pl_landmark_->getNPoints() - (landmark_from_idx - landmark_to_idx) + 1);
    int N_feature = feature_to_id_ - feature_from_id_ + 1;
    if (N_landmark != N_feature)
    {
        WOLF_ERROR("LandmarkMatchPolyline2D different amount of matching points in landmark and feature",
                   "landmark_from_idx: ", pl_landmark_->id2idx(landmark_from_id_),
                   "landmark_to_idx: ", pl_landmark_->id2idx(landmark_to_id_),
                   "feature_from_id_: ", feature_from_id_,
                   "feature_to_id_: ", feature_to_id_);
        return false;
    }
    // correspondences
    int lmk_id = landmark_from_id_;
    int ftr_id = feature_from_id_;
    while (1)
    {
        if (lmk_id == landmark_to_id_)
        {
            if (ftr_id != feature_to_id_)
            {
                WOLF_ERROR("LandmarkMatchPolyline2D landmark_to_id_ does not correspond to feature_to_id_");
                return false;
            }
            break;
        }
        else if (lmk_id == pl_landmark_->getLastId() && !pl_landmark_->isClosed())
        {
            WOLF_ERROR("LandmarkMatchPolyline2D lmk_id == lmk_last_id but lmk_id != landmark_to_id_ in a not closed lmk");
            return false;
        }
        else
        {
            lmk_id = pl_landmark_->getNextValidId(lmk_id);
            ftr_id++;
        }
    }

    return true;
}

Eigen::Array<Scalar,Eigen::Dynamic,1> LandmarkMatchPolyline2D::computeSqDistArray() const
{
    assert(pl_feature_ != nullptr);
    assert(pl_landmark_ != nullptr);

    return computeSqDist(T_feature_landmark_ * pl_landmark_->computePointsGlobal(landmark_from_id_,landmark_to_id_),//_points_A
                         0,                                                                                         //from_A,
                         getNPoints()-1,                                                                            //to_A,
                         landmark_from_id_ != pl_landmark_->getFirstId() || pl_landmark_->isFirstDefined(),         //from_A_defined,
                         landmark_to_id_ != pl_landmark_->getLastId() || pl_landmark_->isLastDefined(),             //to_A_defined,
                         pl_feature_->getPoints(),                                                                  //_points_B
                         feature_from_id_,                                                                          //from_B,
                         feature_to_id_,                                                                            //to_B,
                         feature_from_id_ != 0 || pl_feature_->isFirstDefined(),                                    //from_B_defined,
                         feature_to_id_ != pl_feature_->getNPoints()-1 || pl_feature_->isLastDefined(),             //to_B_defined,
                         getNPoints()-1,                                                                            //last_A,
                         getNPoints());                                                                             //N_overlapped
}

void LandmarkMatchPolyline2D::print() const
{
    assert(pl_feature_ != nullptr);
    assert(pl_landmark_ != nullptr);

    std::cout << std::endl << "POLYLINE LANDMARK MATCH" << std::endl;
    std::cout << "FEATURE: " << pl_feature_->id() << ": " << std::endl;
    std::cout << "\tpoints " << pl_feature_->getNPoints() << std::endl;
    std::cout << "\tfirst defined " << pl_feature_->isFirstDefined() << std::endl;
    std::cout << "\tlast defined " << pl_feature_->isLastDefined() << std::endl;
    std::cout << "\tpoints \n" << pl_feature_->getPoints() << std::endl;

    std::cout << "LANDMARK " << pl_landmark_->id() << ": " << std::endl;
    std::cout << "\tpoints " << pl_landmark_->getNPoints() << " ( ";
    std::vector<int> lmk_ids = pl_landmark_->getValidIds();
    for (auto id : lmk_ids)
        std::cout << id << " ";
    std::cout << ")" << std::endl;
    std::cout << "\tfirst defined " << pl_landmark_->isFirstDefined() << std::endl;
    std::cout << "\tlast defined " << pl_landmark_->isLastDefined() << std::endl << std::endl;
    std::cout << "\tpoints \n" << pl_landmark_->computePointsGlobal() << std::endl;
    std::cout << "\tpoints (local) \n" << T_feature_landmark_ * pl_landmark_->computePointsGlobal() << std::endl;

    std::cout << "MATCH " << std::endl;
    std::cout << "\tfrom feature point " << feature_from_id_ << std::endl;
    std::cout << "\tto feature point " << feature_to_id_ << std::endl;
    std::cout << "\tfrom landmark point " << landmark_from_id_ << std::endl;
    std::cout << "\tto landmark point " << landmark_to_id_ << std::endl;
    std::cout << "\tT_feature_landmark_ " << std::endl << T_feature_landmark_ << std::endl;
}

} // namespace wolf
