/**
 * \file landmark_polyline_2D.cpp
 *
 *  Created on: May 26, 2016
 *      \author: jvallve
 */

#include "laser/feature/feature_polyline_2D.h"
#include "laser/landmark/landmark_polyline_2D.h"
#include "laser/state_block/local_parametrization_polyline_extreme.h"
#include "laser/factor/factor_point_2D.h"
#include "laser/factor/factor_point_to_line_2D.h"
#include "core/state_block/state_block.h"
#include "core/common/factory.h"
#include "core/yaml/yaml_conversion.h"

namespace wolf
{


LandmarkPolyline2D::LandmarkPolyline2D(StateBlockPtr _p_ptr, StateBlockPtr _o_ptr, const Eigen::MatrixXs& _points, const bool _first_defined, const bool _last_defined, unsigned int _first_id) :
        LandmarkPolyline2D(_p_ptr, _o_ptr, _points, _first_defined, _last_defined, _first_id, PolylineRectangularClass())
{
}

LandmarkPolyline2D::LandmarkPolyline2D(StateBlockPtr _p_ptr, StateBlockPtr _o_ptr, const Eigen::MatrixXs& _points, const bool _first_defined, const bool _last_defined, unsigned int _first_id, PolylineRectangularClass _class) :
        LandmarkBase("POLYLINE 2D", _p_ptr, _o_ptr),
		first_defined_(_first_defined),
		last_defined_(_last_defined),
		closed_(false),
		classification_(_class),
		version_(0),
		merged_in_lmk_(nullptr)
{
    //std::cout << "LandmarkPolyline2D::LandmarkPolyline2D" << std::endl;
    assert(_points.cols() >= 2 && "2 points at least needed.");

    for (auto i = 0; i < _points.cols(); i++)
    {
    	auto new_sb = std::make_shared<StateBlock>(_points.col(i).head<2>());
    	point_state_ptr_map_[i+_first_id] = new_sb;
    	getStateBlockVec().push_back(new_sb);
    }

    if (!first_defined_)
        firstStateBlock()->setLocalParametrization(std::make_shared<LocalParametrizationPolylineExtreme>(std::next(point_state_ptr_map_.begin())->second));
    if (!last_defined_)
        lastStateBlock()->setLocalParametrization(std::make_shared<LocalParametrizationPolylineExtreme>(std::next(point_state_ptr_map_.rbegin())->second));

    //std::cout << "LandmarkPolyline2D " << id() << " created. First: " << getFirstId() << " last: "<< getLastId() << std::endl;
}

LandmarkPolyline2D::~LandmarkPolyline2D()
{
    removeStateBlocks();
}

void LandmarkPolyline2D::setFirst(const Eigen::VectorXs& _point, bool _defined)
{
    //std::cout << "LandmarkPolyline2D::setFirst. Defined " << _defined << std::endl;
    assert(_point.size() >= 2 && "LandmarkPolyline2D::setFirstExtreme: bad point size");
    assert(!(!_defined && first_defined_) && "setting a defined extreme with a not defined point");
    firstStateBlock()->setState(_point.head(2));
    if (!first_defined_ && _defined)
    	defineFirst();
}

void LandmarkPolyline2D::setLast(const Eigen::VectorXs& _point, bool _defined)
{
    //std::cout << "LandmarkPolyline2D::setLast. Defined " << _defined << std::endl;
    assert(_point.size() >= 2 && "LandmarkPolyline2D::setLastExtreme: bad point size");
    assert(!(!_defined && last_defined_) && "setting a defined extreme with a not defined point");
    lastStateBlock()->setState(_point.head(2));
    if (!last_defined_ && _defined)
    	defineLast();
}

const Eigen::VectorXs LandmarkPolyline2D::getPointVector(int _i) const
{
	assert(point_state_ptr_map_.find(_i) != point_state_ptr_map_.end());
    return point_state_ptr_map_.at(_i)->getState();
}

StateBlockPtr LandmarkPolyline2D::getPointStateBlock(int _i)
{
    assert(point_state_ptr_map_.find(_i) != point_state_ptr_map_.end());
	return point_state_ptr_map_.at(_i);
}

Eigen::MatrixXs LandmarkPolyline2D::computePointsGlobal() const
{
    Eigen::MatrixXs points_global = Eigen::MatrixXs::Ones(3,getNPoints());
    Eigen::Matrix2s R_world_points = Eigen::Rotation2Ds(getO()->getState()(0)).matrix();
    Eigen::Vector2s t_world_points = getP()->getState();

    for (int i = 0, valid_id = getFirstId(); i < getNPoints(); i++, valid_id = getNextValidId(valid_id))
    {
        points_global.block(0,i,2,1) = getPointVector(valid_id);
        if (valid_id == getLastId())
            break;
    }

    if (classification_.type != UNCLASSIFIED) // points are in landmark PO reference only when clasified
        points_global.topRows(2) = (R_world_points * points_global.topRows(2)).colwise() + t_world_points;

    assert(points_global.isApprox(computePointsGlobal(getFirstId(), getLastId()),1e-12));

    return points_global;
}

Eigen::MatrixXs LandmarkPolyline2D::computePointsGlobal(const int& _from_id, const int& _to_id) const
{
    assert((_from_id <= _to_id || closed_) && "wrong input");

    int N_points = (_from_id <= _to_id ? id2idx(_to_id) - id2idx(_from_id) + 1 : getNPoints() - (id2idx(_from_id) - id2idx(_to_id)) + 1);

    Eigen::MatrixXs points_global = Eigen::MatrixXs::Ones(3,N_points);

    int i = 0, valid_id = _from_id;

    while (i < N_points)
    {
        points_global.block(0,i,2,1) = getPointVector(valid_id);

        if (i == N_points-1)
             break;

        i++;
        valid_id = getNextValidId(valid_id);
    }
    assert(valid_id == _to_id && i == N_points-1);

    if (classification_.type != UNCLASSIFIED) // points are in landmark PO reference only when clasified
    {
        Eigen::Matrix2s R_world_points = Eigen::Rotation2Ds(getO()->getState()(0)).matrix();
        points_global.topRows(2) = (R_world_points * points_global.topRows(2)).colwise() + getP()->getState();
    }

    return points_global;
}

Eigen::MatrixXs LandmarkPolyline2D::computePointsCovGlobal() const
{
    Eigen::MatrixXs points_cov_global = Eigen::MatrixXs::Zero(2,2*getNPoints());

    for (int i = 0, valid_id = getFirstId(); i < getNPoints(); i++, valid_id = getNextValidId(valid_id))
    {
        // TODO
        points_cov_global.middleCols(i*2, 2) = Eigen::MatrixXs::Identity(2,2);

        if (valid_id == getLastId())
            break;
    }

    return points_cov_global;
}

void LandmarkPolyline2D::addPoint(const Eigen::VectorXs& _point, const bool& _defined, const bool& _back)
{
    assert(!closed_ && "adding point to a closed polyline!");

	//std::cout << "LandmarkPolyline2D::addPoint. Defined " << _defined << std::endl;
    assert(_point.size() >= 2 && "bad point size");

    // define previous extreme if not defined
    if (_back ? !last_defined_ : !first_defined_)
    	defineExtreme(_back);

    // add new extreme
    if (_back)
    {
        StateBlockPtr new_sb_ptr = std::make_shared<StateBlock>(_point.head<2>(),
                                                                false,
                                                                (!_defined ?
                                                                        std::make_shared<LocalParametrizationPolylineExtreme>(lastStateBlock()) :
                                                                        nullptr));
        point_state_ptr_map_[getLastId()+1]= new_sb_ptr;
    	getStateBlockVec().push_back(new_sb_ptr);

        if (getProblem())
        	getProblem()->notifyStateBlock(new_sb_ptr,ADD);

        last_defined_ = _defined;
    }
    else
    {
        StateBlockPtr new_sb_ptr = std::make_shared<StateBlock>(_point.head<2>(),
                                                                false,
                                                                (!_defined ?
                                                                        std::make_shared<LocalParametrizationPolylineExtreme>(firstStateBlock()) :
                                                                        nullptr));
        point_state_ptr_map_[getFirstId()-1] = new_sb_ptr;
    	getStateBlockVec().push_back(new_sb_ptr);

        if (getProblem())
        	getProblem()->notifyStateBlock(new_sb_ptr,ADD);

        first_defined_ = _defined;
    }

    // new version
    version_++;
}

void LandmarkPolyline2D::addPoints(const Eigen::MatrixXs& _points, const unsigned int& _idx, const bool& _defined,
                                   const bool& _back)
{
    assert(!closed_ && "adding points to a closed polyline!");

    //std::cout << "LandmarkPolyline2D::addPoints from/to: " << _idx << " Defined " << _defined << std::endl;
    assert(_points.rows() >= 2 && "bad points size");
    assert(_idx < _points.cols() && "bad index!");

    // define previous extreme if not defined
    if (_back ? !last_defined_ : !first_defined_)
    	defineExtreme(_back);

    // add new extreme points
    if (_back)
    {
        for (int i = _idx; i < _points.cols(); i++)
        {
            StateBlockPtr new_sb_ptr = std::make_shared<StateBlock>(_points.block(0,i,2,1),
                                                                    false,
                                                                    (i == _points.cols()-1 && !_defined ?
                                                                            std::make_shared<LocalParametrizationPolylineExtreme>(lastStateBlock()) :
                                                                            nullptr));
            point_state_ptr_map_[getLastId()+1] = new_sb_ptr;
        	getStateBlockVec().push_back(new_sb_ptr);

        	if (getProblem())
        		getProblem()->notifyStateBlock(new_sb_ptr,ADD);
        }
        last_defined_ = _defined;
    }
    else
    {
        for (int i = _idx; i >= 0; i--)
        {
            StateBlockPtr new_sb_ptr = std::make_shared<StateBlock>(_points.block(0,i,2,1),
                                                                    false,
                                                                    (i == 0 && !_defined ?
                                                                            std::make_shared<LocalParametrizationPolylineExtreme>(firstStateBlock()) :
                                                                            nullptr));
            point_state_ptr_map_[getFirstId()-1] = new_sb_ptr;
        	getStateBlockVec().push_back(new_sb_ptr);

            if (getProblem())
        		getProblem()->notifyStateBlock(new_sb_ptr,ADD);
        }
		first_defined_ = _defined;
    }

    // new version
    version_++;
}

void LandmarkPolyline2D::defineFirst()
{
    defineExtreme(false);
}

void LandmarkPolyline2D::defineLast()
{
    defineExtreme(true);
}

void LandmarkPolyline2D::defineExtreme(const bool _back)
{
    StateBlockPtr extreme_state = (_back ? lastStateBlock() : firstStateBlock());
    assert((_back ? !last_defined_: !first_defined_) && "defining an already defined extreme");
    assert(extreme_state->hasLocalParametrization() && "not defined extreme without local parameterization");

    //std::cout << "Defining extreme --> Removing and adding state blocks and factors" << std::endl;

    // remove local parameterization
    extreme_state->removeLocalParametrization();

    // update boolean
    if (_back)
        last_defined_ = true;
    else
        first_defined_ = true;
}

bool LandmarkPolyline2D::tryClose(const Scalar& _dist_tol)
{
    //std::cout << "tryClose landmark " << id() << " with tolerance = " << _dist_tol << std::endl;

    if (closed_)
    {
        WOLF_WARN("trying to close an already closed polyline landmark!");
        return true;
    }
    if (getNDefinedPoints() < 2)
    {
        WOLF_WARN("trying to close an polyline landmark with less than 2 defined points!");
        return false;
    }
    if (getNDefinedPoints() == 2 && getNPoints() < 4)
    {
        WOLF_WARN("trying to close an polyline landmark with 2 defined points and less than 4 points!");
        return false;
    }

    WOLF_DEBUG("tryClose landmark ",id()," point ids: ");
    //for (auto id : getValidIds())
    //    std::cout << id << " ";
    //std::cout << std::endl;

    std::vector<std::pair<int, int> > overlapped_ids;

    // Check first defined point against the rest defined points
    int first_point_id = (first_defined_ ? getFirstId() : getNextValidId(getFirstId()));
    int with_point_id = (last_defined_ ? getLastId() : getPrevValidId(getLastId()));
    int last_with_id  = first_point_id; // limit for avoiding more than one overlapping per defined point (full overlapping of defined points)
    bool found = false;

    // Trying to close overlapping defined points
    if (getNDefinedPoints() > 2)
    {
        // Find the point that matches with the first defined point - the search is done until full overlapping
        // last_with_id is the point overlapped with the last defined point, it CAN'T be overlapped with with_point_id since it's already overlapped with first defined point
        while (with_point_id > last_with_id)
        {
            WOLF_DEBUG("first_point_id: ",first_point_id,"\nwith_point_id: ",with_point_id,"\nlast_with_id: ", last_with_id);
            if ((getPointVector(with_point_id)-getPointVector(first_point_id)).squaredNorm() < _dist_tol*_dist_tol)
            {
                found = true;
                break;
            }

            with_point_id=getPrevValidId(with_point_id);
            last_with_id=getNextValidId(last_with_id);
        }

        // IF MATCH FOUND CHECK THE WHOLE OVERLAPPING
        if (found)
        {
            WOLF_DEBUG("good point match first: ", first_point_id, " with: ", with_point_id);

            // COMPUTE OVERLAPPING POINTS
            // if first not defined -> add previous overlapping point
            if (!first_defined_)
            {
                first_point_id = getPrevValidId(first_point_id);
                with_point_id = getPrevValidId(with_point_id);
            }

            // overlapped points from corresponding first point match to match with last point
            while (with_point_id <= getLastId())
            {
                overlapped_ids.push_back(std::pair<int,int>(first_point_id, with_point_id));
                if (with_point_id == getLastId())
                    break;
                first_point_id = getNextValidId(first_point_id);
                with_point_id = getNextValidId(with_point_id);
            }

            // CHECK ALL OVERLAPPED POINTS if any overlapping candidate found (if just one overlapped point, it has already been checked)
            if (overlapped_ids.size() > 1)
            {
                // Check that the other overlapped points are close enough (0 already checked)
                for (auto i = 0; i < overlapped_ids.size(); i++)
                {
                    Scalar sq_dist=1e6;

                    // 2 defined points: point-point
                    if ( (overlapped_ids[i].first != getFirstId() || first_defined_) &&
                         (overlapped_ids[i].second != getLastId() || last_defined_) )
                        sq_dist = (getPointVector(overlapped_ids[i].first)-getPointVector(overlapped_ids[i].second)).squaredNorm();
                    // defined with not defined: not defined point-segment
                    else if (overlapped_ids[i].first != getFirstId() || first_defined_)
                        sq_dist = sqDistPoint2Segment(getPointVector(overlapped_ids[i].second), getPointVector(overlapped_ids[i].first), getPointVector(getNextValidId(overlapped_ids[i].first)));
                    // not defined with defined: not defined point-segment
                    else if (overlapped_ids[i].second != getLastId() || last_defined_)
                        sq_dist = sqDistPoint2Segment(getPointVector(overlapped_ids[i].first),getPointVector(overlapped_ids[i].second), getPointVector(getPrevValidId(overlapped_ids[i].second)));
                    else
                        throw std::runtime_error("when trying to close, overlapping two not defined points in a landmark");

                    if (sq_dist > _dist_tol*_dist_tol)
                    {
                        WOLF_DEBUG("\nBad overlapping: ", sq_dist,
                                   "\n\tpoint\t", overlapped_ids[i].first, " (", getPointVector(overlapped_ids[i].first).transpose(), ")", (overlapped_ids[i].first == getFirstId() && !first_defined_ ? "NOT DEFINED" : ""),
                                   "\n\twith \t", overlapped_ids[i].second," (", getPointVector(overlapped_ids[i].second).transpose(),")", (overlapped_ids[i].second == getLastId() && !last_defined_ ? "NOT DEFINED" : ""));
                        return false;
                    }
                    WOLF_DEBUG("\nGood overlapping: ", sq_dist,
                               "\n\tpoint\t", overlapped_ids[i].first, " (", getPointVector(overlapped_ids[i].first).transpose(), ")", (overlapped_ids[i].first == getFirstId() && !first_defined_ ? "NOT DEFINED" : ""),
                               "\n\twith \t", overlapped_ids[i].second," (", getPointVector(overlapped_ids[i].second).transpose(),")", (overlapped_ids[i].second == getLastId() && !last_defined_ ? "NOT DEFINED" : ""));
                }
            }
        }
    }

    // Trying to close overlapping both not defined extremes
    if (overlapped_ids.empty() && !first_defined_ && !last_defined_)
    {
        WOLF_DEBUG("No overlapping found with defined points, trying with not defined extremes");

        // Check for both not defined extremes that the distance to the segment of first and last defined points is below the threshold
        Eigen::Vector2s first_not_def   = getPointVector(getFirstId());
        Eigen::Vector2s last_not_def    = getPointVector(getLastId());
        Eigen::Vector2s first_def       = getPointVector(getNextValidId(getFirstId()));
        Eigen::Vector2s last_def        = getPointVector(getPrevValidId(getLastId()));

        if(sqDistPoint2Segment(first_not_def, last_def, first_def) < _dist_tol*_dist_tol &&
           sqDistPoint2Segment(last_not_def,  last_def, first_def) < _dist_tol*_dist_tol)
        {
            WOLF_DEBUG("The two not defined extremes close to the segment defined first_defined and last_defined:",
                       "\n\tfirst point:         ", first_not_def.transpose(),
                       "\n\tlast point:          ", last_not_def.transpose(),
                       "\n\tfirst defined point: ", first_def.transpose(),
                       "\n\tlast defined point:  ", last_def.transpose());

            // Also that the projection of them in the segment should be overlapped
            // It means that last_not_def projected onto the segment (first_def-last_def) is closer to first_def than the projection of first_not_def
            Scalar last_proj_2_first_def  = (last_def-first_def).dot(last_not_def -first_def) / (last_def-first_def).norm();
            Scalar first_proj_2_first_def = (last_def-first_def).dot(first_not_def-first_def) / (last_def-first_def).norm();

            if (last_proj_2_first_def < first_proj_2_first_def)
            {

                WOLF_DEBUG("The not defined extremes are overlapped",
                           "\n\tlast_proj_2_first_def:  ", last_proj_2_first_def,
                           "\n\tfirst_proj_2_first_def: ", first_proj_2_first_def);
                overlapped_ids.push_back(std::pair<int,int>(getFirstId(), getPrevValidId(getLastId())));
                overlapped_ids.push_back(std::pair<int,int>(getLastId(), getNextValidId(getFirstId())));
            }
        }
    }

    // IF ANY OVERLAPPING -> CLOSE AND MERGE OVERLAPPED POINTS
    if (!overlapped_ids.empty())
    {
        WOLF_DEBUG("All overlappings OK, merging ", overlapped_ids.size(), " points");

        // merge points
        while (!overlapped_ids.empty())
        {
            // not defined points are merged into defined ones
            if ((overlapped_ids.back().first == getFirstId() && !first_defined_) || (overlapped_ids.back().first == getLastId() && !last_defined_))
                mergePoints(overlapped_ids.back().first,overlapped_ids.back().second);
            else
                mergePoints(overlapped_ids.back().second,overlapped_ids.back().first);
            overlapped_ids.pop_back();
        }
        // Close
        setClosed();

        WOLF_DEBUG("Landmark ",id()," was closed.");

        return true;
    }
    return false;
}

void LandmarkPolyline2D::setClosed()
{
    assert(getNDefinedPoints()  >= 2 && "closing a polyline with less than 2 defined points");

    // set closed
    closed_ = true;

    // new version
    version_++;
}

void LandmarkPolyline2D::mergePoints(int _remove_id, int _remain_id)
{
    assert(point_state_ptr_map_.find(_remove_id) != point_state_ptr_map_.end() && "removing an inexistent point");
    assert(point_state_ptr_map_.find(_remain_id) != point_state_ptr_map_.end() && "merging an inexistent point");
    assert(!(_remain_id == getLastId() && !last_defined_) && "in merging points, the remaining point must be defined");
    assert(!(_remain_id == getFirstId() && !first_defined_) && "in merging points, the remaining point must be defined");

    //std::cout << "merge points: remove " << _remove_id << " and keep " << _remain_id << " (ids: " << getFirstId() << " to " << getLastId() << ")" << std::endl;

    StateBlockPtr remove_state = getPointStateBlock(_remove_id);
    //std::cout << "state block to remove " << remove_state->getState().transpose() << std::endl;

    // Change factors from remove_state to remain_state
    FactorBasePtrList old_factors_list = getConstrainedByList();
    //std::cout << "changing " << old_factors_list.size() << " factors." << std::endl;
    for (auto fac_ptr : old_factors_list)
    {
        auto fac_point_ptr = std::dynamic_pointer_cast<FactorPoint2D>(fac_ptr);
        auto fac_point_line_ptr = std::dynamic_pointer_cast<FactorPointToLine2D>(fac_ptr);
        if (fac_point_ptr)
        {
            // If landmark point constrained -> new factor
            if (fac_point_ptr->getLandmarkPointId() == _remove_id)
            {
                // emplace new factor
                FactorBase::emplace<FactorPoint2D>(fac_ptr->getFeature(),
                                                   std::static_pointer_cast<FeaturePolyline2D>(fac_ptr->getFeature()),
                                                   std::static_pointer_cast<LandmarkPolyline2D>(shared_from_this()),
                                                   fac_point_ptr->getProcessor(),
                                                   fac_point_ptr->getFeaturePointId(),
                                                   _remain_id,
                                                   fac_point_ptr->getApplyLossFunction(),
                                                   fac_point_ptr->getStatus());
                // remove old factor
                fac_ptr->remove();
            }
        }
        else if (fac_point_line_ptr)
        {
            // If landmark point constrained -> new factor
            if (fac_point_line_ptr->getLandmarkPointId() == _remove_id)
            {
                // emplace new factor
                FactorBase::emplace<FactorPointToLine2D>(fac_ptr->getFeature(),
                                                         std::static_pointer_cast<FeaturePolyline2D>(fac_ptr->getFeature()),
                                                         std::static_pointer_cast<LandmarkPolyline2D>(shared_from_this()),
                                                         fac_point_line_ptr->getProcessor(),
                                                         fac_point_line_ptr->getFeaturePointId(),
                                                         _remain_id,
                                                         fac_point_line_ptr->getLandmarkPointAuxId(),
                                                         fac_point_line_ptr->getApplyLossFunction(),
                                                         fac_point_line_ptr->getStatus());
                // remove old factor
                fac_ptr->remove();
            }
            // If landmark point is aux point -> new factor
            else if (fac_point_line_ptr->getLandmarkPointAuxId() == _remove_id)
            {
                // emplace new factor
                FactorBase::emplace<FactorPointToLine2D>(fac_ptr->getFeature(),
                                                         std::static_pointer_cast<FeaturePolyline2D>(fac_ptr->getFeature()),
                                                         std::static_pointer_cast<LandmarkPolyline2D>(shared_from_this()),
                                                         fac_point_line_ptr->getProcessor(),
                                                         fac_point_line_ptr->getFeaturePointId(),
                                                         fac_point_line_ptr->getLandmarkPointId(),
                                                         _remain_id,
                                                         fac_point_line_ptr->getApplyLossFunction(),
                                                         fac_point_line_ptr->getStatus());
                // remove old factor
                fac_ptr->remove();
            }
        }
        else
        {
            WOLF_ERROR("polyline factor of unknown type");
            throw std::runtime_error ("polyline factor of unknown type");
        }
    }

    // If removed was a not defined extreme, set defined extreme
    if (_remove_id == getFirstId() && !first_defined_)
        first_defined_ = true;
    if (_remove_id == getLastId() && !last_defined_)
        last_defined_ = true;

    // Remove remove_state
    if (getProblem() != nullptr)
        getProblem()->notifyStateBlock(remove_state, REMOVE);
    //std::cout << "state removed " << std::endl;

    // remove stateblock from map
    point_state_ptr_map_.erase(_remove_id);
    // remove state block from state_block_vec
    auto sb_it = std::find(getStateBlockVec().begin(), getStateBlockVec().end(), remove_state);
    assert(sb_it != getStateBlockVec().end());
	getStateBlockVec().erase(sb_it);

    //std::cout << "state removed from point vector " << std::endl;

    // new version
    version_++;
}

bool LandmarkPolyline2D::tryClassify(const Scalar& _dist_tol, std::vector<PolylineRectangularClass> _classes)
{
    //std::cout << "tryClassify landmark " << id() << " with " << getNDefinedPoints() << " defined points of " << getNPoints() << " - tol = " << _dist_tol << std::endl;

    if (classification_.type != UNCLASSIFIED)
    {
        WOLF_WARN("trying to classify a container already classified");
        return false;
    }

    if (closed_ && getNDefinedPoints() != 4)
    {
        WOLF_WARN("trying to classify a closed polyline of Npoints different than 4");
        return false;
    }

    if (getNDefinedPoints() < 3 && getNDefinedPoints() > 4)
    {
        WOLF_WARN("wrong number of defined points for classification of rectangular forms");
        return false;
    }

    assert(classification_.type == UNCLASSIFIED && "trying to classify a container already classified");
    assert((!closed_ || getNDefinedPoints() == 4) && "trying to classify a closed polyline of Npoints different than 4");
    assert(getNDefinedPoints() > 2 && getNDefinedPoints() < 5 && "wrong number of defined points for classification of rectangular forms");

    std::vector<int> points_ids = getValidIds();
    if (!first_defined_)
        points_ids.erase(points_ids.begin());
    if (!last_defined_)
        points_ids.pop_back();

    // consider 3 first defined points A-0, B-1, C-2 (and D-3)
    Scalar dAB = (getPointVector(points_ids[0]) - getPointVector(points_ids[1])).norm();
    Scalar dBC = (getPointVector(points_ids[1]) - getPointVector(points_ids[2])).norm();
    Scalar dAC = (getPointVector(points_ids[0]) - getPointVector(points_ids[2])).norm();

    //std::cout << "dAB = " << dAB << std::endl;
    //std::cout << "dBC = " << dBC << std::endl;
    //std::cout << "dAC = " << dAC << std::endl;

    PolylineRectangularClass pl_class;
    bool configuration = true;

    for (auto candidate_class : _classes)
    //for (unsigned int i = 0; i < _classes.size(); i++)
    {
        //std::cout << "Trying classification " << i << ":\n";
        //std::cout << "Configuration 1:\n\terror L: " << fabs(dAB-candidate_class.L) << " error W: " << fabs(dBC-candidate_class.W) << " error D: " << fabs(dAC-candidate_class.D) << std::endl;
        //std::cout << "Configuration 2:\n\terror L: " << fabs(dBC-candidate_class.L) << " error W: " << fabs(dAB-candidate_class.W) << " error D: " << fabs(dAC-candidate_class.D) << std::endl;

        // check configuration 1
        if(fabs(dAB-candidate_class.L) < _dist_tol &&
           fabs(dBC-candidate_class.W) < _dist_tol &&
           fabs(dAC-candidate_class.D) < _dist_tol)
        {
            configuration = true;
            pl_class = candidate_class;
        }

        // check configuration 2
        if(fabs(dAB-candidate_class.W) < _dist_tol &&
           fabs(dBC-candidate_class.L) < _dist_tol &&
           fabs(dAC-candidate_class.D) < _dist_tol)
        {
            configuration = false;
            pl_class = candidate_class;
        }

        // type
        if (pl_class.type != UNCLASSIFIED)
        {
            // Using 4th defined point to validate (if exists)
            if (getNDefinedPoints() == 4)
            {
                //std::cout << "checking with 4th point... " << std::endl;

                Scalar dAD = (getPointVector(points_ids[0]) - getPointVector(points_ids[3])).norm();
                Scalar dBD = (getPointVector(points_ids[1]) - getPointVector(points_ids[3])).norm();
                Scalar dCD = (getPointVector(points_ids[2]) - getPointVector(points_ids[3])).norm();

                // necessary conditions (rectangular shape) not fulfilled
                if (fabs(dAD-dBC) > _dist_tol ||
                    fabs(dBD-dAC) > _dist_tol ||
                    fabs(dCD-dAB) > _dist_tol)
                {
                    //std::cout << "4th point doesn't match, discarding classification" << std::endl;
                    pl_class = PolylineRectangularClass();
                    continue;
                }
            }
            break;
        }
    }

    // any container position fitted
    if (pl_class.type != UNCLASSIFIED)
    {
        //std::cout << "landmark " << id() << " classified as " << pl_class.type << " in configuration " << (configuration ? "1" : "2") << std::endl;
        //std::cout << "\tdefined points: " << getNDefinedPoints() << "- points: " << getNPoints() << std::endl;

        // ENSURE 4 DEFINITE POINTS
        // if not 4 defined add/define/merge points
        if (getNDefinedPoints() != 4 || getNPoints() != 4)
        {
            // Add a point if not enough points
            if (getNPoints() < 4)
            {
                addPoint(Eigen::Vector2s::Zero(), true, true);
                points_ids.push_back(getLastId());
                //std::cout << "ADDING POINT\n";
                //std::cout << "\tdefined points: " << getNDefinedPoints() << "- points: " << getNPoints() << std::endl;
            }
            assert(getNPoints() >= 4 && "classified landmark must have 4 points");

            // First
            if (!first_defined_)
            {
                //std::cout << "FIRST NOT DEFINED POINT\n";
                if (getNDefinedPoints() == 4) // merge
                {
                    //std::cout << "4 DEFINED POINTS\n";
                    if (!last_defined_)
                        mergePoints(getFirstId(), getPrevValidId(getLastId()));
                    else
                        mergePoints(getFirstId(), getLastId());
                }
                else // define
                {
                    defineFirst();
                    points_ids.push_back(getFirstId());
                }
                //std::cout << "\tdefined points: " << getNDefinedPoints() << "- points: " << getNPoints() << std::endl;
            }
            assert(first_defined_ && "first point should be defined at this point");

            // Last
            if (!last_defined_)
            {
                //std::cout << "FIRST NOT DEFINED POINT\n";
                if (getNDefinedPoints() == 4) // merge
                {
                    //std::cout << "4 DEFINED POINTS\n";
                    mergePoints(getLastId(), getFirstId());
                }
                else // define
                {
                    defineLast();
                    points_ids.push_back(getLastId()); // it is points_ids[3]
                }
                //std::cout << "\tdefined points: " << getNDefinedPoints() << "- points: " << getNPoints() << std::endl;
            }
            assert(last_defined_ && first_defined_ && "classified landmark must have all extremes defined");
            assert(getNDefinedPoints() == 4 && "classified landmark must have 4 defined points");
            assert(getNPoints() == 4 && "classified landmark must have 4 points");
            assert(points_ids.size() == 4 && "classified landmark must have 4 points");
        }

        // Close
        if (!closed_)
            setClosed();

        // Classify
        classify(pl_class);

        // Move origin to to the center
        getP()->setState((getPointVector(points_ids[0]) + getPointVector(points_ids[2])) / 2.0);
        Eigen::Vector2s frame_x = (configuration ? getPointVector(points_ids[0])-getPointVector(points_ids[1]) : getPointVector(points_ids[2])-getPointVector(points_ids[1]));
        getO()->setState(Eigen::Vector1s::Constant(atan2(frame_x(1),frame_x(0))));

        // Unfix origin
        getP()->unfix();
        getO()->unfix();

        //std::cout << "A: " << getPointVector(points_ids[0]).transpose() << std::endl;
        //std::cout << "B: " << getPointVector(points_ids[1]).transpose() << std::endl;
        //std::cout << "C: " << getPointVector(points_ids[2]).transpose() << std::endl;
        //std::cout << "frame_x:           " << frame_x.transpose() << std::endl;
        //std::cout << "frame position:    " << getP()->getState().transpose() << std::endl;
        //std::cout << "frame orientation: " << getO()->getState() << std::endl;

        // Set polyline points to its respective relative positions
        if (configuration)
        {
            getPointStateBlock(points_ids[0])->setState(Eigen::Vector2s( classification_.L / 2,-classification_.W / 2));
            getPointStateBlock(points_ids[1])->setState(Eigen::Vector2s(-classification_.L / 2,-classification_.W / 2));
            getPointStateBlock(points_ids[2])->setState(Eigen::Vector2s(-classification_.L / 2, classification_.W / 2));
            getPointStateBlock(points_ids[3])->setState(Eigen::Vector2s( classification_.L / 2, classification_.W / 2));
        }
        else
        {
            getPointStateBlock(points_ids[0])->setState(Eigen::Vector2s(-classification_.L / 2,-classification_.W / 2));
            getPointStateBlock(points_ids[1])->setState(Eigen::Vector2s(-classification_.L / 2, classification_.W / 2));
            getPointStateBlock(points_ids[2])->setState(Eigen::Vector2s( classification_.L / 2, classification_.W / 2));
            getPointStateBlock(points_ids[3])->setState(Eigen::Vector2s( classification_.L / 2,-classification_.W / 2));
        }

        for (auto st_pair : point_state_ptr_map_)
            st_pair.second->fix();

        return true;
    }
    return false;
}

void LandmarkPolyline2D::mergeLandmark(const LandmarkPolyline2DPtr _merged_lmk, const int& _from, const int& _to, const int& _merged_from, const int& _merged_to)
{
    // merge landmark into this landmark. Assumptions & ideas:
    // - if merged landmark is closed, this is also closed.
    // - if this is closed and the merged landmark has not defined extremes,
    //   then: if all merged defined points match with all points of this -> more than one point can be merged into the same point of this landmark.
    // - each merged landmark point is only merged with one point of this, not necessarily the opposite (previous statement).


    //std::cout << "merging landmark " << _merged_lmk->id() << " to landmark " << id() << std::endl;
    //std::cout << "\tmerged landmark from " << _merged_from << " to " << _merged_to << std::endl;
    //std::cout << "\tremaining landmark from " << _from << " to " << _to << std::endl;
    assert(!isRemoving() && !_merged_lmk->isRemoving() && "merging any landmark already merged or removed");
    assert(!(_merged_lmk->isClosed() && !this->isClosed()) && "merging a closed landmark into a not closed one");

    std::map<int,int> merged_id_to_id;

    // ------------------------ ADD POINTS (if not closed)
    if (!closed_)
    {
        // ADD FRONT POINTS
        if (_merged_from > _merged_lmk->getFirstId())
        {
            assert(_from == getFirstId());
            for (int merged_id = _merged_lmk->getPrevValidId(_merged_from); ; merged_id = _merged_lmk->getPrevValidId(merged_id))
            {
                //std::cout << "adding front point: " << _merged_lmk->getPointVector(merged_id).transpose() << std::endl;
                //std::cout << "Until current first point: " << getFirstId() << "\n";

                // Add point
                addPoint(_merged_lmk->getPointVector(merged_id), merged_id != _merged_lmk->getFirstId() || _merged_lmk->isFirstDefined() ,false);
                //std::cout << "added\n";
                //std::cout << "adding correspondence" << merged_id << " with " << getFirstId() << "\n";
                merged_id_to_id[merged_id] = getFirstId();
                //std::cout << "added correspondence\n";

                // exit loop
                if (merged_id == _merged_lmk->getFirstId())
                    break;
            }
        }
        // ADD BACK POINTS
        if (_merged_to < _merged_lmk->getLastId())
        {
            assert(_to == getLastId());
            for (int merged_id = _merged_lmk->getNextValidId(_merged_to); ; merged_id = _merged_lmk->getNextValidId(merged_id))
            {
                //std::cout << "adding back point: " << _merged_lmk->getPointVector(merged_id).transpose() << std::endl;
                //std::cout << "Until current last point: " << getLastId() << "\n";

                // Add point
                addPoint(_merged_lmk->getPointVector(merged_id), merged_id != _merged_lmk->getLastId() || _merged_lmk->isLastDefined() ,true);
                //std::cout << "added\n";
                //std::cout << "adding correspondence" << merged_id << " with " << getLastId() << "\n";
                merged_id_to_id[merged_id] = getLastId();
                //std::cout << "added correspondence\n";

                // exit loop
                if (merged_id == _merged_lmk->getLastId())
                    break;
            }
        }
    }
    // store id correspondences of overlapped points (also handles closed lmk)
    for (int merged_id = _merged_from, id = _from; ; merged_id = _merged_lmk->getNextValidId(merged_id), id = getNextValidId(id))
    {
        //std::cout << "correspondence points: " << id << " with " << merged_id << std::endl;

        WOLF_ERROR_COND(!(merged_id_to_id.find(merged_id) == merged_id_to_id.end()), "This point id was already indexed as merged:", merged_id);
        assert(merged_id_to_id.find(merged_id) == merged_id_to_id.end());
        merged_id_to_id[merged_id] = id;

        // Set Defined extremes if it's the case
        if (id == getFirstId() && !first_defined_ && (merged_id != _merged_lmk->getFirstId() || _merged_lmk->isFirstDefined()))
            defineFirst();
        if (id == getLastId() && !last_defined_ && (merged_id != _merged_lmk->getLastId() || _merged_lmk->isLastDefined()))
            defineLast();

        // exit loop
        if (merged_id == _merged_to)
        {
            WOLF_ERROR_COND(!(id == _to && merged_id == _merged_to), "Amount of merged points should be the same for remaining and merged landmarks.");
            assert(id == _to && merged_id == _merged_to);
            break;
        }
    }

    // ------------------------- COPY FACTORS
    FactorBasePtrList old_factors_list = _merged_lmk->getConstrainedByList();
    //std::cout << "\tchanging " << old_factors_list.size() << " factors." << std::endl;
    for (auto fac_ptr : old_factors_list)
    {
        //std::cout << "\t\tfactor " << fac_ptr->id() << " to landmark " << fac_ptr->getLandmarkOther()->id() << std::endl;
        assert(fac_ptr->getLandmarkOther() == _merged_lmk);

        auto fac_point_ptr = std::dynamic_pointer_cast<FactorPoint2D>(fac_ptr);
        auto fac_point_line_ptr = std::dynamic_pointer_cast<FactorPointToLine2D>(fac_ptr);

        // point 2 point
        if (fac_point_ptr)
        {
            //std::cout << "\t\tpoint-point factor to id: " << fac_point_ptr->getLandmarkPointId() << std::endl;
            assert(merged_id_to_id.find(fac_point_ptr->getLandmarkPointId()) != merged_id_to_id.end());

            // emplace new factor
            FactorBase::emplace<FactorPoint2D>(fac_ptr->getFeature(),
                                               std::static_pointer_cast<FeaturePolyline2D>(fac_ptr->getFeature()),
                                               std::static_pointer_cast<LandmarkPolyline2D>(shared_from_this()),
                                               fac_point_ptr->getProcessor(),
                                               fac_point_ptr->getFeaturePointId(),
                                               merged_id_to_id[fac_point_ptr->getLandmarkPointId()],
                                               fac_point_ptr->getApplyLossFunction(),
                                               fac_point_ptr->getStatus());
            // remove old factor
            fac_ptr->remove();

        }
        // point 2 line
        else if (fac_point_line_ptr)
        {
            //std::cout << "\t\tpoint-line factor to ids: " << fac_point_line_ptr->getLandmarkPointId() << ", " << fac_point_line_ptr->getLandmarkPointAuxId() << std::endl;
            assert(merged_id_to_id.find(fac_point_line_ptr->getLandmarkPointId()) != merged_id_to_id.end());
            assert(merged_id_to_id.find(fac_point_line_ptr->getLandmarkPointAuxId()) != merged_id_to_id.end());

            // emplace new factor
            FactorBase::emplace<FactorPointToLine2D>(fac_ptr->getFeature(),
                                                     std::static_pointer_cast<FeaturePolyline2D>(fac_ptr->getFeature()),
                                                     std::static_pointer_cast<LandmarkPolyline2D>(shared_from_this()),
                                                     fac_point_line_ptr->getProcessor(),
                                                     fac_point_line_ptr->getFeaturePointId(),
                                                     merged_id_to_id[fac_point_line_ptr->getLandmarkPointId()],
                                                     merged_id_to_id[fac_point_line_ptr->getLandmarkPointAuxId()],
                                                     fac_point_line_ptr->getApplyLossFunction(),
                                                     fac_point_line_ptr->getStatus());
            // remove old factor
            fac_ptr->remove();

        }
        else
        {
            WOLF_ERROR("polyline factor of unknown type");
            throw std::runtime_error ("polyline factor of unknown type");
        }
    }

    // ------------------------- REMOVE _merged_lmk
    _merged_lmk->setMergedInLandmark(std::static_pointer_cast<LandmarkPolyline2D>(shared_from_this()));
    _merged_lmk->remove();
    //std::cout << "\tLandmark deleted\n";
}

//void LandmarkPolyline2D::registerNewStateBlocks()
//{
//    LandmarkBase::registerNewStateBlocks();
//	if (getProblem())
//		for (auto state_it : point_state_ptr_map_)
//			getProblem()->notifyStateBlock(state_it.second, ADD);
//}

//void LandmarkPolyline2D::removeStateBlocks()
//{
//    auto sbp_pair = point_state_ptr_map_.begin();
//    while (sbp_pair != point_state_ptr_map_.end())
//    {
//        if (getProblem())
//            getProblem()->notifyStateBlock(sbp_pair->second, REMOVE);
//
//        sbp_pair = point_state_ptr_map_.erase(sbp_pair);
//    }
//    LandmarkBase::removeStateBlocks();
//}

// static
LandmarkBasePtr LandmarkPolyline2D::create(const YAML::Node& _lmk_node)
{
    // Parse YAML node with lmk info and data
    unsigned int    id              = _lmk_node["id"].as<unsigned int>();
    Eigen::VectorXs pos             = _lmk_node["position"].as<Eigen::VectorXs>();
    bool            pos_fixed       = true;//_lmk_node["position fixed"].as<bool>();
    Eigen::VectorXs ori             = _lmk_node["orientation"].as<Eigen::VectorXs>();
    bool            ori_fixed       = true;//_lmk_node["orientation fixed"].as<bool>();
    int             first_id        = _lmk_node["first_id"].as<int>();
    bool            first_defined   = _lmk_node["first_defined"].as<bool>();
    bool            last_defined    = _lmk_node["last_defined"].as<bool>();
    unsigned int    npoints         = _lmk_node["points"].size();
    int             classif_type    = _lmk_node["classification_type"].as<int>();
    Scalar          classif_L       = _lmk_node["classification_L"].as<Scalar>();
    Scalar          classif_W       = _lmk_node["classification_W"].as<Scalar>();

    // load points
    Eigen::MatrixXs points(2,npoints);
    for (unsigned int i = 0; i < npoints; i++)
        points.col(i)               = _lmk_node["points"][i].as<Eigen::Vector2s>();

    // Create a new landmark
    LandmarkPolyline2DPtr lmk_ptr = std::make_shared<LandmarkPolyline2D>(std::make_shared<StateBlock>(pos, pos_fixed),
                                                                         std::make_shared<StateBlock>(ori, ori_fixed),
                                                                         points,
                                                                         first_defined,
                                                                         last_defined,
                                                                         first_id,
                                                                         PolylineRectangularClass(static_cast<PolylineClassType>(classif_type),
                                                                                                  classif_L,
                                                                                                  classif_W));
    lmk_ptr->setId(id);

    // fix all points
    for (auto st_pair : lmk_ptr->getPointStatePtrMap())
        st_pair.second->fix();

    return lmk_ptr;
}

YAML::Node LandmarkPolyline2D::saveToYaml() const
{
    // First base things
    YAML::Node node = LandmarkBase::saveToYaml();

    // Then add specific things
    node["first_id"]            = getFirstId();
    node["first_defined"]       = first_defined_;
    node["last_defined"]        = last_defined_;
    node["classification_type"] = static_cast<int>(classification_.type);
    node["classification_L"]    = classification_.L;
    node["classification_W"]    = classification_.W;

    for (auto st : getPointsStateBlockVector())
        node["points"].push_back(st->getState());

    return node;
}

void LandmarkPolyline2D::tryMergeLandmarks(LandmarkPolyline2DPtrList& _lmk_list,
                                           const Scalar& max_sq_error,
                                           const Scalar& min_overlapping_dist,
                                           const int& min_N_overlapped,
                                           const int& min_N_defined_overlapped)
{
    //std::cout << "LandmarkPolyline2D::tryMergeLandmarks\n";
    assert(_lmk_list.size() >= 2);

    // 2 by 2
    LandmarkPolyline2DPtr remaining_lmk = _lmk_list.front();
    _lmk_list.pop_front();
    assert(!remaining_lmk->isRemoving());
    LandmarkPolyline2DPtr lmk_1, lmk_2, merged_lmk;
    int remaining_from, remaining_to, merged_from, merged_to;

    while (!_lmk_list.empty())
    {
        lmk_1 = remaining_lmk;
        lmk_2 = _lmk_list.front();
        _lmk_list.pop_front();

        if (lmk_2->isRemoving())
            continue;

        //std::cout << "Candidates to be merged: " << lmk_1->id() << " " << lmk_2->id() << "\n";

        // check best correspondence
        Eigen::MatrixXs points_1 = lmk_1->computePointsGlobal();
        Eigen::MatrixXs points_2 = lmk_2->computePointsGlobal();
        MatchPolyline2DPtr match = computeBestSqDist(points_1, lmk_1->isFirstDefined(), lmk_1->isLastDefined(), lmk_1->isClosed(),
                                                     points_2, lmk_2->isFirstDefined(), lmk_2->isLastDefined(), lmk_2->isClosed(),
                                                     max_sq_error,
                                                     min_overlapping_dist,
                                                     min_N_overlapped,
                                                     min_N_defined_overlapped);

        // match
        if ( match )
        {
            std::vector<int> valid_id_1 = lmk_1->getValidIds();
            std::vector<int> valid_id_2 = lmk_2->getValidIds();

            // decide which is the remaining landmark:
            // If only one closed, this one remains.
            // Otherwise (both closed or both not closed) the one with most factors remains
            if ((lmk_1->isClosed() && !lmk_2->isClosed()) ||
                (lmk_1->isClosed() == lmk_2->isClosed() && lmk_1->getHits() > lmk_2->getHits()))
            {
                remaining_lmk = lmk_1;
                merged_lmk    = lmk_2;
                remaining_from = valid_id_1.at(match->from_A_id_);
                remaining_to   = valid_id_1.at(match->to_A_id_);
                merged_from    = valid_id_2.at(match->from_B_id_);
                merged_to      = valid_id_2.at(match->to_B_id_);
            }
            else
            {
                remaining_lmk = lmk_2;
                merged_lmk    = lmk_1;
                remaining_from = valid_id_2.at(match->from_B_id_);
                remaining_to   = valid_id_2.at(match->to_B_id_);
                merged_from    = valid_id_1.at(match->from_A_id_);
                merged_to      = valid_id_1.at(match->to_A_id_);
            }

            // merge 2 landmarks
            remaining_lmk->mergeLandmark(merged_lmk, remaining_from, remaining_to, merged_from, merged_to);

            // reset pointers
            lmk_1.reset();
            lmk_2.reset();
            assert(merged_lmk->getHits() == 0);
            //std::cout << "merged\n";
        }
        // no match (remaining: most factors)
        else
            remaining_lmk = (lmk_1->getHits() > lmk_2->getHits() ? lmk_1 : lmk_2);
    }
    //std::cout << "tryMergeLandmarks::done\n";
}

// Register landmark creator
namespace
{
const bool WOLF_UNUSED registered_lmk_polyline_2D = LandmarkFactory::get().registerCreator("POLYLINE 2D", LandmarkPolyline2D::create);
}

} /* namespace wolf */
