#include "laser/processor/polyline_2D_utils.h"

namespace wolf {

Eigen::Matrix3s pose2T2D(const Eigen::Vector2s& t, const Scalar& theta)
{
    Eigen::Matrix3s T(Eigen::Matrix3s::Identity());

    T.topLeftCorner(2,2) = Eigen::Rotation2D<Scalar>(theta).matrix();
    T.topRightCorner(2,1) = t;

    return T;
}

Eigen::Matrix3s pose2T2D(const Eigen::Vector3s& pose)
{
    Eigen::Matrix3s T(Eigen::Matrix3s::Identity());

    T.topLeftCorner(2,2) = Eigen::Rotation2D<Scalar>(pose(2)).matrix();
    T.topRightCorner(2,1) = pose.head(2);

    return T;
}

Eigen::Vector3s T2pose2D(const Eigen::Matrix3s& T)
{
    Eigen::Vector3s pose;
    pose.head(2) = T.topRightCorner(2,1);
    pose(2) = atan2(T(1,0),T(0,0));

    return pose;
}

Eigen::Vector3s computeRigidTrans(const Eigen::MatrixXs& _points_A, const Eigen::MatrixXs& _points_B, const bool& first_defined, const bool& last_defined)
{
    assert(_points_A.cols() == _points_B.cols());
    assert(_points_A.rows() >= 2);
    assert(_points_B.rows() >= 2);

    int N_defined_points = _points_A.cols() - (first_defined ? 0 : 1) - (last_defined ? 0 : 1);
    bool all_defined = first_defined && last_defined;
    Eigen::Vector3s t_A_B;
    assert(N_defined_points > 0 && _points_A.cols() > 1);

    // Rotation
    Eigen::VectorXs angles_A_B(Eigen::VectorXs::Zero(_points_A.cols()));
    Eigen::Vector2s points_A_mean((all_defined ? _points_A.row(0).mean() : _points_A.row(0).segment(first_defined ? 0 : 1, N_defined_points).mean()),
                                  (all_defined ? _points_A.row(1).mean() : _points_A.row(1).segment(first_defined ? 0 : 1, N_defined_points).mean()));
    Eigen::Vector2s points_B_mean((all_defined ? _points_B.row(0).mean() : _points_B.row(0).segment(first_defined ? 0 : 1, N_defined_points).mean()),
                                  (all_defined ? _points_B.row(1).mean() : _points_B.row(1).segment(first_defined ? 0 : 1, N_defined_points).mean()));
    Eigen::MatrixXs points_A_centered = _points_A.topRows<2>().colwise()-points_A_mean;
    Eigen::MatrixXs points_B_centered = _points_B.topRows<2>().colwise()-points_B_mean;

    for (auto i = 0; i < _points_A.cols(); i++ )
        angles_A_B(i) = pi2pi(atan2(points_A_centered(1,i),points_A_centered(0,i)) - atan2(points_B_centered(1,i),points_B_centered(0,i)));

    // fix 2nd&3rd quadrant angles bad resulting mean:
    // if simultaneously in 2nd and 3rd quadrants, add 2PI in 3rd and 4th
    if (angles_A_B.maxCoeff() > M_PI / 2 && angles_A_B.minCoeff() < -M_PI / 2)
        angles_A_B = (angles_A_B.array() >= 0).select(angles_A_B, angles_A_B.array()+2*M_PI);

    // if only one defined point, mean of not defined points angles (mean point is pivot, atan2 is not defined)
    if (N_defined_points == 1)
        t_A_B(2) = ((first_defined ? 0 : angles_A_B(0)) + (last_defined ? 0 : angles_A_B(_points_A.cols()-1))) / (_points_A.cols() - 1);
    // use all points if all defined
    else if (all_defined)
        t_A_B(2) = angles_A_B.mean();
    // otherwise compute using defined points
    else
        t_A_B(2) = angles_A_B.segment(first_defined ? 0 : 1, N_defined_points).mean();

    // Translation
    t_A_B.head<2>() = points_A_mean - Eigen::Rotation2D<Scalar>(t_A_B(2))*points_B_mean;

    return t_A_B;
}

/** \brief Computes the squared distance from a point to a segment defined by two points
 **/
Scalar sqDistPoint2Segment(const Eigen::Vector2s& p, const Eigen::Vector2s& a, const Eigen::Vector2s& b)
{
    Scalar ap_sq = (p-a).squaredNorm();
    Scalar bp_sq = (p-b).squaredNorm();
    Scalar ab_sq = (b-a).squaredNorm();
    Scalar sqDist;

    // If the projection of p onto ab is inside the segment (acute triangle: all angles < 90º), return the sq distance point-line.
    if (ap_sq <= bp_sq + ab_sq && bp_sq <= ap_sq + ab_sq) // acute triangle: applying pitagoras twice, angles ABP and PAB <= 90º
        sqDist = ap_sq - std::pow((b-a).dot(p-a),2) / ab_sq;

    // Otherwise, return the sq distance to the closest segment point.
    else
        sqDist = std::min(bp_sq,ap_sq);

    return sqDist;
}

Scalar sqDistPointToLine(const Eigen::Vector3s& _A, const Eigen::Vector3s& _A_aux,
                         const Eigen::Vector3s& _B, bool _A_defined, bool _B_defined, bool strict)
{
    /* Squared distance from B to the line A_aux-A (match evaluated is A-B)
     *
     * No matter if A_aux is defined or not.
     *
     * Case 1: B not defined & A not defined
     *
     *      If projection of B over AAaux is in [A_aux, inf), return sq.dist. B-Line(AAux)
     *      Otherwise, return sqDist(Aaux-B) or 1e12 if strict=true.
     *
     *      Check: Angle BAauxA is <= 90º: (AB)² <= (AauxB)² + (AAaux)²
     *
     * Case 2: B not defined & A defined
     *
     *      If projection of B over AAaux is in [A_aux, A], return sq.dist. B-Line(AAux)
     *      Otherwise, return min(sqDist(Aaux-B),sqDist(A-B)) or 1e12 if strict=true.
     *
     *      Check: Angles BAauxA & BAAaux are <= 90º: (AB)² <= (AauxB)² + (AAaux)² & (AauxB)² <= (AB)² + (AAaux)²
     *
     * Case 3: B defined & A not defined
     *
     *      If projection of B over AAaux is in [A, inf), return sq.dist. B-Line(AAux)
     *      Otherwise, return sqDist(A-B)) or 1e12 if strict=true.
     *
     *      Check: Angle BAAaux is >= 90º: (AauxB)² >= (AB)² + (AAaux)²
     *
     * (both points defined is not point-to-line, is point-to-point)
     *
     */

    assert((!_A_defined || !_B_defined) && "ProcessorPolyline::sqDistPointToLine: at least one point must not be defined.");

    Scalar AB_sq = (_B-_A).head<2>().squaredNorm();
    Scalar AauxB_sq = (_B-_A_aux).head<2>().squaredNorm();
    Scalar AAaux_sq = (_A_aux-_A).head<2>().squaredNorm();

    // squared distance to line
    // Case 1
    if (!_B_defined && !_A_defined)
    {
        if (AB_sq <= AauxB_sq + AAaux_sq)
            return AB_sq - std::pow(((_A_aux-_A).head<2>()).dot((_B-_A).head<2>()),2) / AAaux_sq;
        else
            return (strict ? 1e12 : AauxB_sq);
    }

    // Case 2
    if (!_B_defined && _A_defined)
    {
        if (AB_sq <= AauxB_sq + AAaux_sq && AauxB_sq <= AB_sq + AAaux_sq)
            return AB_sq - std::pow(((_A_aux-_A).head<2>()).dot((_B-_A).head<2>()),2) / AAaux_sq;
        else
            return (strict ? 1e12 : std::min(AauxB_sq,AB_sq));
    }

    // Case 3
    if (_B_defined && !_A_defined)
    {
        if (AauxB_sq >= AB_sq + AAaux_sq)
            return AB_sq - std::pow(((_A_aux-_A).head<2>()).dot((_B-_A).head<2>()),2) / AAaux_sq;
        else
            return (strict ? 1e12 : AB_sq);
    }

    // Default return unused
    return 1e12;
}

MatchPolyline2DPtr computeBestSqDist(const Eigen::MatrixXs& _points_A, bool _first_defined_A, bool _last_defined_A, bool _closed_A,
                                     const Eigen::MatrixXs& _points_B, bool _first_defined_B, bool _last_defined_B, bool _closed_B,
                                     const Scalar& sq_error_max,
                                     const Scalar& overlap_dist_ratio_min,
                                     int overlap_N_min,
                                     int overlap_N_defined_min,
                                     bool both_exceeding_A_allowed_,
                                     bool both_exceeding_B_allowed_,
									 bool print)
{
    if (print)
    {
        std::cout << "computeBestSqDist...\n";
        std::cout << "\tA is " << (_closed_A ? "CLOSED" : "OPEN") << " has "<< _points_A.cols() << " points\n";
        std::cout << "\tB is " << (_closed_B ? "CLOSED" : "OPEN") << " has "<< _points_B.cols() << " points\n";
        std::cout << "\tA extremes defined: " << (_first_defined_A ? "1 " : "0 ") << (_last_defined_A ? "1" : "0") << std::endl;
        std::cout << "\tB extremes defined: " << (_first_defined_B ? "1 " : "0 ") << (_last_defined_B ? "1" : "0") << std::endl;
        std::cout << "\tpoints_A:\n" << _points_A.topRows(2) << std::endl;
        std::cout << "\tpoints_B:\n" << _points_B.topRows(2) << std::endl;
        std::cout << "\tsq_error_max: " << sq_error_max << std::endl;
        std::cout << "\toverlap_dist_ratio_min: " << overlap_dist_ratio_min << std::endl;
        std::cout << "\toverlap_N_min: " << overlap_N_min << std::endl;
        std::cout << "\toverlap_N_defined_min: " << overlap_N_defined_min << std::endl;
    }

    if (overlap_N_min < 2)
    {
    	WOLF_WARN("overlap_N_min should be at least 2! It is a sufficient condition (if enough overlapping distance ratio)");
    	overlap_N_min = 2;
    }
    if (overlap_N_defined_min < 1)
    {
    	WOLF_WARN("overlap_N_defined_min should be at least 1! It is a sufficient condition");
    	overlap_N_defined_min = 1;
    }

    // ASSERTIONS if closed -> all points defined
    assert((_closed_A &&_first_defined_A && _last_defined_A) || !_closed_A);
    assert((_closed_B &&_first_defined_B && _last_defined_B) || !_closed_B);

    /* ASSUMPTIONS:
     *
     *      IF BOTH NOT CLOSED:
     *          (a) - A has equal or more points than B         --> Otherwise: Call the function switching A<->B
     *
     *      IF ONE CLOSED:
     *          (b) - Should be A closed (not B)                --> Otherwise: Call the function switching A<->B
     *          (c) - B.defined_points <= A.(defined_)points    --> Otherwise: No matching
     *
     *      IF BOTH CLOSED:
     *          (d) - B.(defined_)points == A.(defined_)points  --> Otherwise: No matching
     */

    MatchPolyline2DPtr best_match = nullptr;

    // ----------------- Call switching A<->B
    if ( (!_closed_A && !_closed_B && _points_A.cols() < _points_B.cols()) || // (a)
         (!_closed_A &&  _closed_B) )                                         // (b)
    {
        //std::cout << "Auto-calling switching A<->B...\n";

        best_match = computeBestSqDist(_points_B, _first_defined_B, _last_defined_B, _closed_B,
                                                          _points_A, _first_defined_A, _last_defined_A, _closed_A,
                                                          sq_error_max,
                                                          overlap_dist_ratio_min,
                                                          overlap_N_min,
                                                          overlap_N_defined_min,
                                                          both_exceeding_B_allowed_,
                                                          both_exceeding_A_allowed_);
        // undo switching
        if (best_match != nullptr)
        {
            std::swap(best_match->from_A_id_, best_match->from_B_id_);
            std::swap(best_match->to_A_id_,   best_match->to_B_id_);
        }
        return best_match;
    }

    // ------------------ No matching (c) & (d)
    int N_defined_B = _points_B.cols() - (_first_defined_B ? 0 : 1) - (_last_defined_B ? 0 : 1);

    if ((_closed_A && !_closed_B && _points_A.cols() < N_defined_B) ||     // (c)
        (_closed_A &&  _closed_B && _points_A.cols() != _points_B.cols())) // (d)
        return nullptr;

    /* ------------------ Normal function call:
     * Search along all possible overlapping configurations between A&B points
     *      We define offset as the correspondence (if exists) of the first point_A to B's.
     *      Offset is within [min_offset, max_offset], we distinguish between 3 cases:
     *          1. None closed:  [-last_A, last_B]
     *          2. A closed:     [-last_A, 0]
     */

    int last_A, last_B, offset, max_offset, min_offset, from_A, to_A, from_B, to_B, N_overlap;
    last_A = _points_A.cols() - 1;
    last_B = _points_B.cols() - 1;

    if (!_closed_A)
    {
        min_offset = -last_A;
        max_offset = last_B;
    }
    else
    {
        min_offset = -last_A;
        max_offset = 0;
    }

    // Check all overlapping configurations
    for (offset = min_offset; offset <= max_offset; offset++)
    {
        from_A = std::max(0,-offset);
        from_B = std::max(0, offset);

        if (!_closed_A && !_closed_B)
            N_overlap = std::min(last_B - from_B, last_A - from_A)+1;
        else if(_closed_A)
            N_overlap = _points_B.cols();
        else
            std::runtime_error("Impossible state. Function should auto-called reversing A&B");

        to_A = from_A+N_overlap-1;
        to_B = from_B+N_overlap-1;

        while (to_A > last_A && _closed_A)
            to_A -= _points_A.cols();

        // exceeding points for both extremes allowed?
        if (!both_exceeding_A_allowed_ && from_A != 0 && to_A != last_A)
            continue;
        if (!both_exceeding_B_allowed_ && from_B != 0 && to_B != last_B)
            continue;

        if (print)
        {
            std::cout << "\toffset " << offset << std::endl;
            std::cout << "\t\tfrom_A          " << from_A << std::endl;
            std::cout << "\t\tto_A            " << to_A << std::endl;
            std::cout << "\t\tfrom_B          " << from_B << std::endl;
            std::cout << "\t\tto_B            " << to_B << std::endl;
            std::cout << "\t\tlast_A          " << last_A << std::endl;
            std::cout << "\t\tlast_B          " << last_B << std::endl;
            std::cout << "\t\tN_overlap    " << N_overlap << std::endl;
            std::cout << "\t\tfirst_defined_A " << _first_defined_A << std::endl;
            std::cout << "\t\tfirst_defined_B " << _first_defined_B << std::endl;
            std::cout << "\t\tlast_defined_A  " << _last_defined_A << std::endl;
            std::cout << "\t\tlast_defined_B  " << _last_defined_B << std::endl;
        }

        bool from_A_defined = (from_A != 0 || _first_defined_A) && (from_A != last_A || _last_defined_A);   //((from_A != 0 && from_A != last_A) || _first_defined_A);
        bool from_B_defined = (from_B != 0 || _first_defined_B) && (from_B != last_B || _last_defined_B);   //((from_B != 0 && from_B != last_B) || _first_defined_B);
        bool to_A_defined   = (to_A != 0 || _first_defined_A) && (to_A != last_A || _last_defined_A);       //((to_A   != 0 && to_A   != last_A) || _last_defined_A);
        bool to_B_defined   = (to_B != 0 || _first_defined_B) && (to_B != last_B || _last_defined_B);       //((to_B   != 0 && to_B   != last_B) || _last_defined_B);

        // SUFFICIENT CONDITIONS
        //      COND 1: sufficient overlapping defined points
        //      COND 2: sufficient overlapping points (if no defined points, min overlapping distance)
        int N_defined_overlap = std::max(0,N_overlap - (from_A_defined && from_B_defined ? 0 : 1) - (to_A_defined && to_B_defined ? 0 : 1));
		if ( N_defined_overlap < overlap_N_defined_min &&
			 N_overlap < overlap_N_min )
		{
			if (print)
				std::cout  << "\t\tnone of the sufficient conditions hold\n";
			continue;
		}
		// additional overlapping distance contidion for 2 not defined points
		if (N_defined_overlap == 0)
		{
			Scalar overlap_dist;
			WOLF_WARN_COND(N_overlap != 2, "THIS SHOULDN'T HAPPEN. \nN_defined_overlap: ", N_defined_overlap, "\nN_overlap: ", N_overlap, "\noverlap_N_defined_min: ", overlap_N_defined_min, "\noverlap_N_min: ", overlap_N_min)
			assert(N_overlap == 2);
			Eigen::Vector2s p_from_A = _points_A.col(from_A).head(2);
			Eigen::Vector2s p_to_A   = _points_A.col(to_A).head(2);
			Eigen::Vector2s p_from_B = _points_B.col(from_B).head(2);
			Eigen::Vector2s p_to_B   = _points_B.col(to_B).head(2);
			Eigen::Vector2s vA = p_to_A - p_from_A;
			Eigen::Vector2s vB = p_to_B - p_from_B;

			// THE MINIMUM OF ALL 6 possible overlap distances
			Eigen::Vector6s overlap_dists;
			// 1: fromA - toA
			overlap_dists(0) = vA.norm();
			// 2: fromB - toB
			overlap_dists(1) = vB.norm();
			// 3: fromA - projected(toB)
			overlap_dists(2) = std::abs((vA.normalized()).dot(p_to_B - p_from_A));
			// 4: fromB - projected(toA)
			overlap_dists(3) = std::abs((vB.normalized()).dot(p_to_A - p_from_B));
			// 5: projected(fromB) - toA
			overlap_dists(4) = std::abs((-vA.normalized()).dot(p_from_B - p_to_A));
			// 6: projected(fromA) - toB
			overlap_dists(5) = std::abs((-vB.normalized()).dot(p_from_A - p_to_B));

			//std::cout << "\tdist2.mean() = " << dist2.mean() << std::endl;
			//std::cout << "\tmax_sq_error - dist2.mean() = " << max_sq_error - dist2.mean() << std::endl;
			if (print)
				std::cout << "\toverlap_dists ratios = " << (overlap_dists/std::min(vA.norm(),vB.norm())).transpose() << std::endl;
			overlap_dist = overlap_dists.minCoeff();

			// if necessary overlapping condition does not hold, continue
			if (overlap_dist/std::min(vA.norm(),vB.norm()) < overlap_dist_ratio_min)
			{
				if (print)
					std::cout  << "\t\toverlapping distance ratio condition does not hold with 0 defined points overlapped\n";
				continue;
			}
		}

        assert(N_overlap <= _points_B.cols());
        assert(N_overlap <= _points_A.cols() || (_closed_A && N_defined_B < _points_B.cols())); // N_overlap can be > _points_A if all defined points of B match with A points and the not defined extremes should match too
        assert(from_A >= 0);
        assert(from_B >= 0);
        assert(from_B == 0 || !_closed_A);
        assert(to_B == _points_B.cols()-1 || !_closed_A);
        assert(to_A <= last_A);
        assert(to_B <= last_B);
        assert(to_A < _points_A.cols());
        assert(to_B < _points_B.cols());

        // NECESSARY COND: All squared distances should be lower than a threshold
        // squared distances for all overlapping points
        Eigen::Array<Scalar,Eigen::Dynamic,1> dist2 = computeSqDist(_points_A, from_A, to_A, from_A_defined, to_A_defined,
                                                                    _points_B, from_B, to_B, from_B_defined, to_B_defined,
                                                                    last_A, N_overlap );
        if (print)
            std::cout << "\t\tsquared distances = " << dist2.transpose() << std::endl;

        // if the necessary condition does not hold, continue
        if ((dist2 > sq_error_max).any())
        {
            if (print)
                std::cout  << "\t\ttoo much error\n";
            continue;
        }

        // COMPUTE SCORE (distance based)
        Scalar normalized_score = (sq_error_max - dist2.mean()) / sq_error_max;
        if (best_match != nullptr && print)
        {
            std::cout << "\tnormalized_score = " << normalized_score << std::endl;
            std::cout << "\tbest_match->normalized_score_ = " << best_match->normalized_score_ << std::endl;
        }

        // BEST MATCH
        if (best_match == nullptr || normalized_score > best_match->normalized_score_ )
        {
            if (print)
                std::cout << "\tBEST MATCH normalized_score = " << normalized_score << std::endl;

            if (best_match == nullptr)
                best_match = std::make_shared<MatchPolyline2D>();

            best_match->from_A_id_= from_A;
            best_match->to_A_id_= to_A;
            best_match->from_B_id_= from_B;
            best_match->to_B_id_= to_B;
            best_match->normalized_score_ = normalized_score;
            best_match->overlap_dist_ = 0;//overlap_dist;

            if (print)
                best_match->print();
        }
    }

    return best_match;
}

Eigen::Array<Scalar,Eigen::Dynamic,1> computeSqDist(const Eigen::MatrixXs& _points_A, const int& from_A, const int& to_A, bool from_A_defined, bool to_A_defined,
                                                    const Eigen::MatrixXs& _points_B, const int& from_B, const int& to_B, bool from_B_defined, bool to_B_defined,
                                                    const int& last_A, const int& N_overlap )
{
    // POINT-POINT DISTANCES for all overlap points (in case of closed A in two parts)
    // Fill with B points
    Eigen::ArrayXXd d(_points_B.block(0,from_B, 2, N_overlap).array());

    // Substract A points
    if (to_A >= from_A && N_overlap <= _points_A.cols())
        d -= _points_A.block(0,from_A, 2, N_overlap).array();
    else
    {
        d.leftCols(_points_A.cols()-from_A) -= _points_A.block(0,from_A, 2, _points_A.cols()-from_A).array();
        d.rightCols(to_A+1)                 -= _points_A.block(0, 0, 2, to_A+1).array();
    }
    Eigen::Array<Scalar,Eigen::Dynamic,1> dist2 = d.row(0).pow(2) + d.row(1).pow(2);

    // CORRECT POINT-LINE DISTANCES for not defined extremes
    // First
    if (!from_B_defined || !from_A_defined)
    {
        // take care of closed A: next of from_A
        int next_from_A = (from_A+1 > last_A ? 0 : from_A+1);
        //std::cout << "\t\t\tFirst feature not defined distance to line" << std::endl;

        // std::cout << "\t\t\tA   " << _points_A.col(from_A).transpose() << std::endl;
        // std::cout << "\t\t\tAaux" << _points_A.col(next_from_A).transpose() << std::endl;
        // std::cout << "\t\t\tB   " << _points_B.col(from_B).transpose() << std::endl;
        dist2(0) = sqDistPointToLine(_points_A.col(from_A),
                                     _points_A.col(next_from_A),
                                     _points_B.col(from_B),
                                     from_A_defined,
                                     from_B_defined);
    }
    // Last
    if (!to_B_defined || !to_A_defined)
    {
        // take care of closed A: next of to_A
        int prev_to_A = (to_A == 0 ? last_A : to_A-1);
        //std::cout << "\t\t\tLast feature not defined distance to line" << std::endl;

        // std::cout << "\t\t\tA   " << _points_A.col(to_A).transpose() << std::endl;
        // std::cout << "\t\t\tAaux" << _points_A.col(prev_to_A).transpose() << std::endl;
        // std::cout << "\t\t\tB   " << _points_B.col(to_B).transpose() << std::endl;
        dist2(N_overlap-1) = sqDistPointToLine(_points_A.col(to_A),
                                                  _points_A.col(prev_to_A),
                                                  _points_B.col(to_B),
                                                  to_A_defined,
                                                  to_B_defined);
    }
    //std::cout << "\t\tsquared distances = " << dist2.transpose() << std::endl;

    return dist2;
}

MatchPolyline2DList computeBestRigidTrans(const Eigen::MatrixXs& _points_A, bool _first_defined_A, bool _last_defined_A, bool _closed_A,
                                          const Eigen::MatrixXs& _points_B, bool _first_defined_B, bool _last_defined_B, bool _closed_B,
                                          const Scalar& max_sq_error )
{
    //std::cout << "computeBestRigidTrans...\n";

    // ASSERTIONS if closed -> all points defined
    assert((_closed_A &&_first_defined_A && _last_defined_A) || !_closed_A);
    assert((_closed_B &&_first_defined_B && _last_defined_B) || !_closed_B);

    /* ASSUMPTIONS:
     *
     *      IF BOTH NOT CLOSED:
     *          (a) - A has equal or more points than B         --> Otherwise: Call the function switching A<->B
     *
     *      IF ONE CLOSED:
     *          (b) - Should be A closed (not B)                --> Otherwise: Call the function switching A<->B
     *          (c) - B.defined_points <= A.(defined_)points    --> Otherwise: No matching
     *
     *      IF BOTH CLOSED:
     *          (d) - B.(defined_)points == A.(defined_)points  --> Otherwise: No matching
     */

    MatchPolyline2DList matches;
    auto N_defined_points_A = _points_A.cols() - (_first_defined_A ? 0 : 1) - (_last_defined_A  ? 0 : 1);
    auto N_defined_points_B = _points_B.cols() - (_first_defined_B ? 0 : 1) - (_last_defined_B  ? 0 : 1);

    // Not enough defined points
    if (N_defined_points_A <= 1 || N_defined_points_B <= 1 )
        return matches;

    //std::cout << "\tA is " << (_closed_A ? "CLOSED" : "NOT CLOSED") << " has "<< _points_A.cols() << " points (" << N_defined_points_A << " defined)\n";
    //std::cout << "\tB is " << (_closed_B ? "CLOSED" : "NOT CLOSED") << " has "<< _points_B.cols() << " points (" << N_defined_points_B << " defined)\n";
    //std::cout << "\tpoints_A:\n" << _points_A.topRows(2) << std::endl;
    //std::cout << "\tpoints_B:\n" << _points_B.topRows(2) << std::endl;

    // ----------------- Call switching A<->B
    if ( (!_closed_A && !_closed_B && _points_A.cols() < _points_B.cols()) || // (a)
         (!_closed_A &&  _closed_B) )                                         // (b)
    {
        //std::cout << "Auto-calling switching A<->B...\n";

        matches = computeBestRigidTrans(_points_B, _first_defined_B, _last_defined_B, _closed_B,
                                        _points_A, _first_defined_A, _last_defined_A, _closed_A, max_sq_error);
        // undo switching
        for (auto match : matches)
        {
            std::swap(match->from_A_id_, match->from_B_id_);
            std::swap(match->to_A_id_,   match->to_B_id_);
            match->T_A_B_ = match->T_A_B_.inverse().eval();
        }
        return matches;
    }

    // ------------------ No matching (c) & (d) return empty list
    int N_defined_B = _points_B.cols() - (_first_defined_B ? 0 : 1) - (_last_defined_B ? 0 : 1);

    if ((_closed_A && !_closed_B && _points_A.cols() < N_defined_B) ||     // (c)
        (_closed_A &&  _closed_B && _points_A.cols() != _points_B.cols())) // (d)
        return matches;

    /* ------------------ Normal function call:
     * Search along all possible overlapping configurations between A&B points
     *      We define offset as the correspondence (if exists) of the first point_A to B's.
     *      Offset is within [min_offset, max_offset], we distinguish between 3 cases:
     *          1. None closed:  [-last_A, last_B]
     *          2. A closed:     [-last_A, 0]
     */

    int last_A, last_B, offset, max_offset, min_offset, from_A, to_A, from_B, to_B, N_overlap(0);
    last_A = _points_A.cols() - 1;
    last_B = _points_B.cols() - 1;

    if (!_closed_A)
    {
        min_offset = -last_A;
        max_offset = last_B;
    }
    else
    {
        min_offset = -last_A;
        max_offset = 0;
    }

    // Check all overlapping configurations
    for (offset = min_offset; offset <= max_offset; offset++)
    {
        from_A = std::max(0,-offset);
        from_B = std::max(0, offset);

        if (!_closed_A && !_closed_B)
            N_overlap = std::min(last_B - from_B, last_A - from_A)+1;
        else if(_closed_A)
            N_overlap = _points_B.cols();
        else
            std::runtime_error("Impossible state. Function should auto-called reversing A&B");

        to_A = from_A+N_overlap-1;
        to_B = from_B+N_overlap-1;

        while (to_A > last_A && _closed_A)
            to_A -= _points_A.cols();

        //std::cout << "\toffset " << offset << std::endl;
        //std::cout << "\t\tfrom_A " << from_A << std::endl;
        //std::cout << "\t\tto_A   " << to_A << std::endl;
        //std::cout << "\t\tfrom_B " << from_B << std::endl;
        //std::cout << "\t\tto_B   " << to_B << std::endl;
        //std::cout << "\t\tlast_A  " << last_A << std::endl;
        //std::cout << "\t\tlast_B  " << last_B << std::endl;
        //std::cout << "\t\tN_overlap " << N_overlap << std::endl;

        bool from_A_defined = ((from_A != 0 && from_A != last_A) || _first_defined_A);
        bool from_B_defined = ((from_B != 0 && from_B != last_B) || _first_defined_B);
        bool to_A_defined   = ((to_A   != 0 && to_A   != last_A) || _last_defined_A);
        bool to_B_defined   = ((to_B   != 0 && to_B   != last_B) || _last_defined_B);

        //std::cout << "\t\tfrom_A_defined " << from_A_defined << (from_A == 0) << (from_A == last_A) << _first_defined_A << std::endl;
        //std::cout << "\t\tfrom_B_defined " << from_B_defined << (from_B == 0) << (from_B == last_B) << _first_defined_B << std::endl;
        //std::cout << "\t\tto_A_defined   " << to_A_defined   << (to_A == 0)   << (to_A == last_A)   << _last_defined_A << std::endl;
        //std::cout << "\t\tto_B_defined   " << to_B_defined   << (to_B == 0)   << (to_B == last_B)   << _last_defined_B << std::endl;

        // If only one overlap point, no rigid transformation possible
        if ( N_overlap == 1 )
            continue;
        // If only two overlap points, all should be defined
        if ( N_overlap == 2 &&
             (!from_A_defined || !from_B_defined || !to_A_defined || !to_B_defined ) )
            continue;

        assert(N_overlap <= _points_B.cols());
        assert(N_overlap <= _points_A.cols() || (_closed_A && N_defined_B < _points_B.cols())); // N_overlap can be > _points_A if all defined points of B match with A points and the not defined extremes should match too
        assert(from_A >= 0);
        assert(from_B >= 0);
        assert(from_B == 0 || !_closed_A);
        assert(to_B == _points_B.cols()-1 || !_closed_A);
        assert(to_A <= last_A);
        assert(to_B <= last_B);
        assert(to_A < _points_A.cols());
        assert(to_B < _points_B.cols());

        // RIGID TRANSFORMATION FOR DEFINED POINTS
        //std::cout << "\tcomputing rigid transformation...\n";
        int from_B_def = from_B_defined ? from_B : from_B+1;
        int from_A_def = from_A_defined ? from_A : from_A+1;
        int to_A_def   = to_A_defined ? to_A : to_A+1;
        int N_overlap_def = N_overlap - (from_B_defined && from_A_defined ? 0 : 1) - (to_B_defined && to_A_defined  ? 0 : 1);
        Eigen::MatrixXs points_B_def = _points_B.middleCols(from_B_def, N_overlap_def);
        Eigen::MatrixXs points_A_def(_points_A.rows(), N_overlap_def);
        if (to_A >= from_A)
            points_A_def = _points_A.middleCols(from_A_def, N_overlap_def);
        else
        {
            points_A_def.leftCols(_points_A.cols()-from_A_def) = _points_A.middleCols(from_A_def, _points_A.cols()-from_A_def);
            points_A_def.rightCols(to_A_def+1)                 = _points_A.middleCols(0, to_A_def+1);
        }
        //std::cout << "\tpoints filled...\n";

        //std::cout << "\tcomputeRigidTrans\n";
        Eigen::Vector3s pose_A_B = computeRigidTrans(points_A_def, points_B_def);
        Eigen::Matrix3s T_A_B = pose2T2D(pose_A_B.head(2), pose_A_B(2));

        // POINT-POINT DISTANCES of the matching of defined points
        Eigen::MatrixXs d = (points_A_def - T_A_B * points_B_def).topRows(2);
        Eigen::ArrayXd dist2(N_overlap);
        dist2.middleRows(from_B_defined && from_A_defined ? 0 : 1, N_overlap_def) = d.colwise().squaredNorm().transpose().array();

        // POINT-LINE DISTANCES for not defined extremes
        //std::cout << "\tPOINT-LINE DISTANCES\n";
        // First
        if (!from_B_defined || !from_A_defined)
        {
            // take care of closed A: next of from_A
            int next_from_A = (from_A+1 > last_A ? 0 : from_A+1);
            //std::cout << "\t\t\tFirst feature not defined distance to line" << std::endl;

            // std::cout << "\t\t\tA   " << _points_A.col(from_A).transpose() << std::endl;
            // std::cout << "\t\t\tAaux" << _points_A.col(next_from_A).transpose() << std::endl;
            // std::cout << "\t\t\tB   " << _points_B.col(from_B).transpose() << std::endl;
            dist2(0) = sqDistPointToLine(_points_A.col(from_A),
                                         _points_A.col(next_from_A),
                                         T_A_B * _points_B.col(from_B),
                                         from_A_defined,
                                         from_B_defined);
            assert(N_overlap > 1);
        }
        // Last
        if (!to_B_defined || !to_A_defined)
        {
            // take care of closed A: next of to_A
            int prev_to_A = (to_A == 0 ? last_A : to_A-1);
            //std::cout << "\t\t\tLast feature not defined distance to line" << std::endl;

            // std::cout << "\t\t\tA   " << _points_A.col(to_A).transpose() << std::endl;
            // std::cout << "\t\t\tAaux" << _points_A.col(prev_to_A).transpose() << std::endl;
            // std::cout << "\t\t\tB   " << _points_B.col(to_B).transpose() << std::endl;
            dist2(N_overlap-1) = sqDistPointToLine(_points_A.col(to_A),
                                                      _points_A.col(prev_to_A),
                                                      T_A_B * _points_B.col(to_B),
                                                      to_A_defined,
                                                      to_B_defined);
            assert(N_overlap > 1);
        }
        //std::cout << "\t\tsquared distances = " << dist2.transpose() << "( should be < " << max_sq_error << ")" << std::endl;

        // All squared distances should be within a threshold
        if ((dist2 < max_sq_error).all())
        {
            auto match = std::make_shared<MatchPolyline2D>();

            match->from_A_id_= from_A;
            match->to_A_id_= to_A;
            match->from_B_id_= from_B;
            match->to_B_id_= to_B;
            match->normalized_score_ = (max_sq_error - dist2.mean()) / max_sq_error;
            match->T_A_B_ = T_A_B;

            // Store matches ordered by the squared norm of the transformation (I know, mixing mts and rads..)
            matches.push_back(match);
            //std::cout << "match added to list of matches!\n";
        }
    }

    //std::cout << matches.size() << " rigid transformations found\n";
    return matches;
}

bool isProjectedOverSegment(const Eigen::Vector3s& _A, const Eigen::Vector3s& _A_aux, const Eigen::Vector3s& _B)
{
    /* The orthogonal projection of B over the line A-Aaux is in the segment [A,Aaux].
     *
     *      Check: the angles BAAaux and BAauxA are <= 90º:
     *          (BAaux)² <= (BA)² + (AAaux)² & (BA)² <= (BAaux)² + (AAaux)²
     *
     */

    Scalar AB_sq = (_B-_A).head<2>().squaredNorm();
    Scalar AauxB_sq = (_B-_A_aux).head<2>().squaredNorm();
    Scalar AAaux_sq = (_A_aux-_A).head<2>().squaredNorm();

    return (AauxB_sq <= AB_sq + AAaux_sq) && (AB_sq <= AauxB_sq + AAaux_sq);
}

Eigen::VectorXs computeSquaredMahalanobisDistances(const Eigen::Vector2s& _feature,
                                                   const Eigen::Matrix2s& _feature_cov,
                                                   const Eigen::Vector2s& _expected_feature,
                                                   const Eigen::Matrix2s& _expected_feature_cov,
                                                   const Eigen::MatrixXs& _mu)
{
    // ------------------------ d
    Eigen::Vector2s d = _feature - _expected_feature;

    // ------------------------ Sigma_d
    Eigen::Matrix2s iSigma_d = (_feature_cov + _expected_feature_cov).inverse();
    Eigen::VectorXs squared_mahalanobis_distances(_mu.cols());
    for (unsigned int i = 0; i < _mu.cols(); i++)
    {
        squared_mahalanobis_distances(i) = (d - _mu.col(i)).transpose() * iSigma_d * (d - _mu.col(i));
        //if ((d - _mu.col(i)).norm() < 1)
        //{
        //    std::cout << "distance:    " << (d - _mu.col(i)).norm() << std::endl;
        //    std::cout << "iSigma_d:    " << std::endl << iSigma_d << std::endl;
        //    std::cout << "mahalanobis: " << squared_mahalanobis_distances(i) << std::endl;
        //}
    }

    return squared_mahalanobis_distances;
}

} // namespace wolf
