#include "laser/processor/processor_odom_icp.h"

using namespace wolf;
using namespace laserscanutils;

ProcessorOdomICP::ProcessorOdomICP(ProcessorParamsOdomICPPtr _params):
                    ProcessorTracker("ODOM ICP", _params)
{
    proc_params_ = _params;


    icp_params_.use_point_to_line_distance  = proc_params_->use_point_to_line_distance;
    icp_params_.max_correspondence_dist     = proc_params_->max_correspondence_dist;
    icp_params_.max_iterations              = proc_params_->max_iterations;
    icp_params_.use_corr_tricks             = proc_params_->use_corr_tricks;
    icp_params_.outliers_maxPerc            = proc_params_->outliers_maxPerc;
    icp_params_.outliers_adaptive_order     = proc_params_->outliers_adaptive_order;
    icp_params_.outliers_adaptive_mult      = proc_params_->outliers_adaptive_mult;

    // ICP algorithm
    icp_tools_ptr_ = std::make_shared<ICP>();

    // Frame transforms
    trf_origin_last_.res_covar      = Eigen::Matrix3s::Identity();
    trf_origin_incoming_.res_covar  = Eigen::Matrix3s::Identity();
    trf_last_incoming_.res_covar    = Eigen::Matrix3s::Identity();
}

ProcessorOdomICP::~ProcessorOdomICP()
{

}

void ProcessorOdomICP::preProcess()
{

}



void ProcessorOdomICP::configure(SensorBasePtr _sensor)
{
    sensor_laser_       = std::static_pointer_cast<SensorLaser2D>(_sensor);
    laser_scan_params_  = sensor_laser_->getScanParams();
}



unsigned int ProcessorOdomICP::processKnown()
{
    // Match the incoming with the origin, passing the transform from origin to last as initialization

    if (origin_ptr_) // Make sure it's not the FIRST_TIME
    {
        CaptureLaser2DPtr origin_ptr    = std::static_pointer_cast<CaptureLaser2D>(origin_ptr_);
        CaptureLaser2DPtr incoming_ptr  = std::static_pointer_cast<CaptureLaser2D>(incoming_ptr_);

        trf_origin_incoming_ = icp_tools_ptr_->align(incoming_ptr->getScan(),
                                                     origin_ptr->getScan(),
                                                     this->laser_scan_params_,
                                                     this->icp_params_,
                                                     this->trf_origin_last_.res_transf); // Check order
    }

    return 0;
}

unsigned int ProcessorOdomICP::processNew(const int& _max_features)
{
    Eigen::Vector3s t_identity;
    t_identity << 0, 0, 0;

    // XXX: Dynamic or static? JS: static is OK.
    CaptureLaser2DPtr   incoming_ptr    = std::static_pointer_cast<CaptureLaser2D>(incoming_ptr_);
    CaptureLaser2DPtr   last_ptr        = std::static_pointer_cast<CaptureLaser2D>(last_ptr_);

    trf_last_incoming_ =  icp_tools_ptr_->align(incoming_ptr->getScan(),
                                                last_ptr->getScan(),
                                                this->laser_scan_params_,
                                                this->icp_params_,
                                                t_identity);
    return 0;
}

bool ProcessorOdomICP::voteForKeyFrame()
{
    return (voteForKeyFrameDistance() ||
            voteForKeyFrameAngle() ||
            voteForKeyFrameTime() ||
            voteForKeyFrameMatchQuality());
}

inline bool ProcessorOdomICP::voteForKeyFrameDistance()
{
    bool vote = (trf_origin_incoming_.res_transf.head<2>().norm() > proc_params_->vfk_min_dist);
    if (vote)
    {
        std::cout << "Distance True" << '\n';
    }

    return vote;
}

inline bool ProcessorOdomICP::voteForKeyFrameAngle()
{
    bool vote = (fabs(trf_origin_incoming_.res_transf(2)) > proc_params_->vfk_min_angle);
    if (vote)
    {
        std::cout << "Angle True" << '\n';
    }

    return vote;
}

inline bool ProcessorOdomICP::voteForKeyFrameTime()
{
    Scalar secs = incoming_ptr_->getTimeStamp() - origin_ptr_->getTimeStamp();

    bool vote = (secs > proc_params_->vfk_min_time);

    if (vote)
    {
        std::cout << "Time True" << '\n';
    }

    return vote;
}

inline bool ProcessorOdomICP::voteForKeyFrameMatchQuality()
{
    bool vote = (trf_origin_incoming_.error > proc_params_->vfk_min_error || trf_origin_incoming_.nvalid < proc_params_->vfk_max_points);
    if  (vote)
    {
        std::cout << "Quality True" << '\n';
    }
    return vote ;
}


void ProcessorOdomICP::advanceDerived()
{
    using namespace Eigen;

    // overwrite last frame's state

    Isometry2ds   w_T_ro = Translation2ds(origin_ptr_->getFrame()->getState().head(2)) * Rotation2Ds(origin_ptr_->getFrame()->getState()(2));
    Isometry2ds  ro_T_so = Translation2ds(origin_ptr_->getSensorP()->getState())       * Rotation2Ds(origin_ptr_->getSensorO()->getState()(0));
    Isometry2ds& ri_T_si = ro_T_so; // A reference just to have nice names without computing overhead

    // incoming

    Isometry2ds so_T_si = Translation2ds(trf_origin_incoming_.res_transf.head(2))     * Rotation2Ds(trf_origin_incoming_.res_transf(2));
    Isometry2ds  w_T_ri = w_T_ro * ro_T_so * so_T_si * ri_T_si.inverse();
    Vector3s x_inc; x_inc << w_T_ri.translation() , Rotation2Ds(w_T_ri.rotation()).angle();

    // Put the state of incoming in last
    last_ptr_->getFrame()->setState(x_inc);

    trf_origin_last_ = trf_origin_incoming_;
}

void ProcessorOdomICP::resetDerived()
{

    // Using processing_step_ to ensure that origin, last and incoming are different
    if (processing_step_ != FIRST_TIME_WITH_PACK && processing_step_ != FIRST_TIME_WITHOUT_PACK)
    {
        // Notation explanation:
        // x1_R_x2: Rotation from frame x1 to frame x2
        // x1_p_y1_y2: translation vector from y1 to y2 expressed in frame x1

        // Rotation composition
        Eigen::Rotation2D<Scalar>   w_R_ro  = Eigen::Rotation2D<Scalar>(origin_ptr_->getFrame()->getState()(2));
        Eigen::Rotation2D<Scalar>   r_R_s   = Eigen::Rotation2D<Scalar>(origin_ptr_->getSensorO()->getState()(0));
        Eigen::Rotation2D<Scalar>& ro_R_so  = r_R_s; // just a ref for bettter chaining trf. names
        Eigen::Rotation2D<Scalar>  so_R_sl  = Eigen::Rotation2D<Scalar>(trf_origin_last_.res_transf(2));

        // Eigen::Rotation2D<Scalar> w_R_rl = w_R_ro*r_R_s*so_R_sl*r_R_s.inverse();
        // ro_R_so = rl_R_sl = r_R_s
        // Planar rotations are commutative
        Eigen::Rotation2D<Scalar> w_R_rl = w_R_ro * so_R_sl;

        // Translation composition
        Eigen::Vector2s w_p_w_ro    = origin_ptr_->getFrame()->getState().head(2);
        Eigen::Vector2s w_p_ro_so   = w_R_ro * origin_ptr_->getSensorP()->getState();
        Eigen::Vector2s w_p_so_sl   = w_R_ro * ro_R_so * trf_origin_last_.res_transf.head(2);
        Eigen::Vector2s w_p_sl_rl   = w_R_rl * (-last_ptr_->getSensorP()->getState());

        Eigen::Vector2s w_p_w_rl = w_p_w_ro + w_p_ro_so + w_p_so_sl + w_p_sl_rl;

        Eigen::Vector3s curr_state;
        curr_state.head(2)  = w_p_w_rl;
        curr_state(2)       = origin_ptr_->getFrame()->getState()(2) + trf_origin_last_.res_transf(2);

        last_ptr_->getFrame()->setState(curr_state);

        trf_origin_last_ = trf_last_incoming_;
    }
}

void ProcessorOdomICP::establishFactors()
{
    if (last_ptr_ != origin_ptr_)
    {
        auto ftr_ptr = emplaceFeature(last_ptr_);
        emplaceFactor(ftr_ptr);
    }
}

FeatureBasePtr ProcessorOdomICP::emplaceFeature(CaptureBasePtr _capture_laser)
{
    CaptureLaser2DPtr capture_laser = std::static_pointer_cast<CaptureLaser2D>(_capture_laser);
    return FeatureBase::emplace<FeatureICPAlign>(capture_laser,
                                                trf_origin_last_);
}

FactorBasePtr ProcessorOdomICP::emplaceFactor(FeatureBasePtr _feature)
{
    return FactorBase::emplace<FactorOdom2D>(_feature, _feature, origin_ptr_->getFrame(),
                                                      shared_from_this());
}


// Register in the SensorFactory
#include "core/processor/processor_factory.h"
namespace wolf {
    WOLF_REGISTER_PROCESSOR("ODOM ICP", ProcessorOdomICP);
    WOLF_REGISTER_PROCESSOR_AUTO("ODOM ICP", ProcessorOdomICP);
} // namespace wolf
