/*
 * processor_tracker_feature_polyline_2D.cpp
 *
 *  Created on: Mar 11, 2019
 *      Author: jvallve
 */

#include "laser/processor/processor_tracker_feature_polyline_2D.h"

namespace wolf
{

ProcessorTrackerFeaturePolyline2D::ProcessorTrackerFeaturePolyline2D(ProcessorParamsTrackerFeaturePolyline2DPtr _params) :
     ProcessorTrackerFeature("TRACKER FEATURE POLYLINE",_params),
     params_tracker_feature_polyline_(_params),
     line_finder_(_params->line_finder_params),
     extrinsics_transformation_computed_(false)
{

}

ProcessorTrackerFeaturePolyline2D::~ProcessorTrackerFeaturePolyline2D()
{
}

unsigned int ProcessorTrackerFeaturePolyline2D::trackFeatures(const FeatureBasePtrList& _features_in,
                                                              const CaptureBasePtr& _capture,
                                                              FeatureBasePtrList& _features_out,
                                                              FeatureMatchMap& _feature_correspondences)
{
	WOLF_DEBUG("PTFP ", getName(), "::trackFeatures ", _features_in.size());

    //
    //    return 0;

    // TRACK FEATURES ==================================================
	if (!_features_in.empty())
	{
		// prior transformations
		Eigen::Matrix3s T_last_incoming_prior(Eigen::Matrix3s::Identity());
		T_last_incoming_prior.topLeftCorner(2,2)  = R_last_incoming_;
		T_last_incoming_prior.topRightCorner(2,1) = t_last_incoming_;

		// ALL AGAINST ALL: nearest neighbor matching (already detected incoming features stored in: untracked_features_incoming_)
		auto ftr_it = untracked_features_incoming_.begin();

		// iterate over all untracked_features_incoming_ looking for last features in _feature_in
		while (ftr_it != untracked_features_incoming_.end())
		{
			bool matched = false;
			auto pl_incoming = std::static_pointer_cast<FeaturePolyline2D>(*ftr_it);
			FeatureMatchPolyline2DScalarMap best_matches;

			WOLF_DEBUG("----tracking feature incoming ", pl_incoming->id(), " with last features");

			// Check matching with all features in last
			// Store all matches consistent with T_last_incoming_prior in best_matches sorted by normalized score
			for (auto ftr_last_ : _features_in)
				tryMatchWithFeature(best_matches,
									std::static_pointer_cast<FeaturePolyline2D>(ftr_last_),
									pl_incoming,
									T_last_incoming_prior);

			// Try the match for all candidates (sorted by normalized score, higher scores at the end!)
			for (auto best_ftr_match_pair_it = best_matches.rbegin(); best_ftr_match_pair_it != best_matches.rend(); best_ftr_match_pair_it++ )
			{
				auto best_ftr_match = best_ftr_match_pair_it->second;
				auto pl_last = std::static_pointer_cast<FeaturePolyline2D>(best_ftr_match->feature_ptr_);

				//std::cout << "\t\tconfirming match with feature last: " << pl_last->id() << " with score " << best_ftr_match->normalized_score_ << std::endl;

				// VALIDATE LAST-LANDMARK --------------------------------------------------------------
				// only if last feature not recently emplaced in processNew
				if(!std::binary_search(new_features_last_.begin(), new_features_last_.end(), pl_last))
				{
					// Removed match
					if(landmark_match_map_.count(pl_last) == 0)
					{
						WOLF_WARN("Match last-landmark was removed, discarding tracking with incoming");
						continue;
					}

					auto pl_lmk  = std::static_pointer_cast<LandmarkPolyline2D>(landmark_match_map_[pl_last]->landmark_ptr_);

					// Landmark changed or merged -> update match with last feature
					if (landmark_match_map_[pl_last]->isDeprecated())
					{
						//std::cout << "\t\tupdating match last-landmark...\n";
						if (!updateLandmarkMatch(landmark_match_map_[pl_last]))
						{
							// not successful update -> remove match and feature
							WOLF_WARN("Not success: removing last feature ", pl_last->id(), " and removing landmark match");
							landmark_match_map_.erase(pl_last);
							pl_last->remove();
							continue;
						}
					}
					// Landmark removed -> not valid match with landmark
					else if (pl_lmk->isRemoving())
					{
						WOLF_WARN("Removed landmark ", pl_lmk->id(), " removing last feature ", pl_last->id(), " and removing landmark match");
						landmark_match_map_.erase(pl_last);
						pl_last->remove();
						continue;
					}

				}
				//std::cout << "\t\tlast-landmark still valid...\n";

				// INCOMING-LANDMARK --------------------------------------------------------------
				// only if there is a landmark, so if last feature is not in new_features_last_
				if(!std::binary_search(new_features_last_.begin(), new_features_last_.end(), pl_last))
				{
					// try to concatenate landmark match
					auto pl_lmk_match_incoming = concatenateFeatureLandmarkMatches(best_ftr_match,
																				   landmark_match_map_[pl_last]);

					// if concatenation did not succees -> next ftr match
					if (pl_lmk_match_incoming == nullptr)
					{
						WOLF_WARN("PTFP ", getName(), "::trackFeatures: incoming-last + last-landmark could not be concatenated into incoming-landmark");
						continue;
					}
					//std::cout << "\t\tlast-incoming valid, storing...\n";

					// Add the incoming match to landmark_match_map
					landmark_match_map_[pl_incoming] = pl_lmk_match_incoming;
				}

				// STORE INCOMING-LAST --------------------------------------------------------------
				// add best match to feature-feature correspondences
				_feature_correspondences[pl_incoming] = best_ftr_match;

				// add feature to list of tracked features
				_features_out.push_back(pl_incoming);

				// link the incoming eature with the capture (since it was just created in preProcess())
				pl_incoming->link(_capture);

				// match for this feature has been found
				matched = true;

				WOLF_DEBUG("Tracked feature");
				//best_ftr_match->print();
				break;
			}

			// next feature
			if (matched)
				ftr_it = untracked_features_incoming_.erase(ftr_it); // remove feature from untracked features list
			else
				ftr_it++;
		}
	}

    // TESTING ********************************************************
    // TRACK FEATURES ==================================================
	if (!getProblem()->getMap()->getLandmarkList().empty())
	{
		// prior transformations
		Eigen::Matrix3s T_sensor_world_incoming(Eigen::Matrix3s::Identity());
		T_sensor_world_incoming.topLeftCorner(2,2)  = R_sensor_world_incoming_;
		T_sensor_world_incoming.topRightCorner(2,1) = t_sensor_world_incoming_;
		// iterate over all untracked_features_incoming_ looking for existing landmarks
		auto ftr_it = untracked_features_incoming_.begin();
		while (ftr_it != untracked_features_incoming_.end())
		{
			auto pl_incoming = std::static_pointer_cast<FeaturePolyline2D>(*ftr_it);
			LandmarkMatchPolyline2DScalarMap best_lmk_matches;

			WOLF_DEBUG("----tracking feature incoming ", pl_incoming->id(), " with landmarks");

			// SEARCH: Check matching with all existing landmarks
			for (auto lmk : getProblem()->getMap()->getLandmarkList())
				// Check only polyline landmarks that hasn't been just emplaced
				if (lmk->getType() == "POLYLINE 2D")
					// Store all matches consistent with T_sensor_world_incoming in best_lmk_matches sorted by squared distance
					tryMatchWithLandmark(best_lmk_matches, std::static_pointer_cast<LandmarkPolyline2D>(lmk), pl_incoming, T_sensor_world_incoming);

			// Matches!
			if (!best_lmk_matches.empty())
			{
				// store best match (highest score) in map
				landmark_match_map_[pl_incoming] = best_lmk_matches.rbegin()->second;

				WOLF_DEBUG("Tracked landmark");
				//landmark_match_map_[pl_incoming]->print();

				// store landmark candidates to be merged
				if (best_lmk_matches.size() > 1)
				{
					LandmarkPolyline2DPtrList merge_candidates;
					for (auto lmk_match_pair_it : best_lmk_matches)
						merge_candidates.push_back(lmk_match_pair_it.second->pl_landmark_);
					merge_candidates_list_.push_back(merge_candidates);
				}

				// link the incoming feature with the capture (since it was just created in preProcess())
				pl_incoming->link(_capture);

				// new track
				track_matrix_.newTrack(pl_incoming);

				// remove feature from untracked features list
				ftr_it = untracked_features_incoming_.erase(ftr_it);
			}
			else
				ftr_it++;
		}
	}

    WOLF_DEBUG(_feature_correspondences.size(), " features tracked");
    return _feature_correspondences.size();
}

bool ProcessorTrackerFeaturePolyline2D::correctFeatureDrift(const FeatureBasePtr _origin_feature,
                                                            const FeatureBasePtr _last_feature,
                                                            FeatureBasePtr _incoming_feature)
{
	WOLF_DEBUG("PTFP ", getName(), "::correctFeatureDrift: incoming feature ", _incoming_feature->id());
	WOLF_DEBUG("PTFP ", getName(), "::correctFeatureDrift: last feature ", _last_feature->id());

    /** After finding a valid feature match last-incoming, we have to check that the concatenated match
     *  incoming-landmark is also valid.
     *
     *  We will only have to check IN BOTH EXTREMES if:
     *  1. There are exceeding points in incoming-last match.
     *  AND
     *  2. Landmark has exceeding points in the same extreme OR is closed.
     *
     *  It this is the case, we have to check that the exceeding points match with the corresponding existing landmark points.
     */

    assert(matches_last_from_incoming_.count(_incoming_feature) != 0);
    assert(landmark_match_map_.count(_incoming_feature) != 0);

    auto lmk_inc_match = std::static_pointer_cast<LandmarkMatchPolyline2D>(landmark_match_map_[_incoming_feature]);
    auto pl_lmk        = lmk_inc_match->pl_landmark_;
    auto pl_inc        = lmk_inc_match->pl_feature_;
    assert(pl_lmk != nullptr);
    assert(pl_inc != nullptr);
    assert(lmk_inc_match != nullptr);

    // WOLF_DEBUG("LANDMARK-INCOMING:");
    // lmk_inc_match->print();

    assert(lmk_inc_match->check(false) && "before correct drift, check not passed");//without checking partial match

    // COMPLETE PARTIAL MATCH
    if (!tryCompletePartialMatch(lmk_inc_match))
    {
        WOLF_WARN("correctFeatureDrift: tryCompletePartialMatch didn't success, removing..", "last feature ", _last_feature->id(), " incoming feature ", _incoming_feature->id(), " landmark ", pl_lmk->id());
        landmark_match_map_.erase(_incoming_feature);
        return false;
    }

    // FILTER OUT matches that does not pass check
    if (!lmk_inc_match->check())
    {
        WOLF_WARN("correctFeatureDrift: match didn't pass the check, removing..", "last feature ", _last_feature->id(), " incoming feature ", _incoming_feature->id(), " landmark ", pl_lmk->id());
        landmark_match_map_.erase(_incoming_feature);
        return false;
    }

    // CORRECT DRIFT
    // feature landmark transformation
    WOLF_DEBUG("try update match...");
    tryUpdateMatchTransformation(lmk_inc_match);

    // RETRY LANDMARK-MATCH IF TOO MUCH ERROR
    if ((lmk_inc_match->computeSqDistArray() >= 0.5).any())
    {
        WOLF_DEBUG("too much error: ", lmk_inc_match->computeSqDistArray().transpose(), " retrying match landmark ", pl_lmk->id(), " with feature ", _incoming_feature->id() );
        LandmarkMatchPolyline2DScalarMap lmk_matches;
        // try with concatenated transformation
        tryMatchWithLandmark(lmk_matches, pl_lmk, pl_inc, lmk_inc_match->T_feature_landmark_);

        // not succeed
        if (lmk_matches.empty())
        {
			WOLF_DEBUG("correctFeatureDrift: after retrying tryMatchWithLandmark with concatenated transformation didn't success, trying with SLAM transformation..", "last feature ", _last_feature->id(), " incoming feature ", _incoming_feature->id(), " landmark ", pl_lmk->id());

        	// try with SLAM transformation
            Eigen::Matrix3s T_incoming_map_prior(Eigen::Matrix3s::Identity());
            T_incoming_map_prior.topLeftCorner(2,2)  = R_sensor_world_incoming_;
            T_incoming_map_prior.topRightCorner(2,1) = t_sensor_world_incoming_;
        	tryMatchWithLandmark(lmk_matches, pl_lmk, pl_inc, T_incoming_map_prior);
        	if (lmk_matches.empty())
			{
				WOLF_WARN("correctFeatureDrift: after retrying tryMatchWithLandmark didn't success, removing..", "last feature ", _last_feature->id(), " incoming feature ", _incoming_feature->id(), " landmark ", pl_lmk->id());
				landmark_match_map_.erase(_incoming_feature);
				return false;
			}
        }
        // override if succed
		lmk_inc_match = lmk_matches.rbegin()->second;
		landmark_match_map_[_incoming_feature] = lmk_inc_match;
    }

    WOLF_DEBUG("Validated match:");
    //lmk_inc_match->print();

    assert(lmk_inc_match->check() && "after correct drift, check not passed"); // checking partial match

    return true;
}

bool ProcessorTrackerFeaturePolyline2D::tryCompletePartialMatch(LandmarkMatchPolyline2DPtr& lmk_inc_match) const
{
    auto pl_lmk = lmk_inc_match->pl_landmark_;
    auto pl_inc = lmk_inc_match->pl_feature_;

    // FRONT POINTS
    while (lmk_inc_match->isPartialFront())
    {
        //WOLF_INFO("Partial match in front...");
        //lmk_inc_match->print();

        auto exceeding_inc = lmk_inc_match->feature_from_id_-1;
        auto exceeding_lmk = pl_lmk->getPrevValidId(lmk_inc_match->landmark_from_id_);

        // check good overlap
        Eigen::Vector3s lmk_point_expected = lmk_inc_match->T_feature_landmark_ * pl_lmk->computePointsGlobal(exceeding_lmk,exceeding_lmk);
        Eigen::Vector3s ftr_point = pl_inc->getPoints().col(exceeding_inc);

        Scalar sq_dist = (lmk_point_expected.head(2) - ftr_point.head(2)).squaredNorm();

        // exceeding point not matched -> return not succeed
        if (sq_dist > params_tracker_feature_polyline_->match_landmark_position_sq_norm_max)
        {
            WOLF_WARN("correctFeatureDrift: any point didn't match while checking incoming-lmk front exceeding point: feature point ", exceeding_inc, " - landmark point ", exceeding_lmk);
            return false;
        }
        // add exceeding point to the landmark match
        else
        {
            lmk_inc_match->feature_from_id_  = exceeding_inc;
            lmk_inc_match->landmark_from_id_ = exceeding_lmk;
        }
    }
    // BACK POINTS
    while (lmk_inc_match->isPartialBack())
    {
        //WOLF_INFO("Partial match in back...");
        //lmk_inc_match->print();

        auto exceeding_inc = lmk_inc_match->feature_to_id_+1;
        auto exceeding_lmk = pl_lmk->getNextValidId(lmk_inc_match->landmark_to_id_);

        // check good overlap
        Eigen::Vector3s lmk_point_expected = lmk_inc_match->T_feature_landmark_ * pl_lmk->computePointsGlobal(exceeding_lmk,exceeding_lmk);
        Eigen::Vector3s ftr_point = pl_inc->getPoints().col(exceeding_inc);

        Scalar sq_dist = (lmk_point_expected.head(2) - ftr_point.head(2)).squaredNorm();

        // exceeding point not matched -> return not succeed
        if (sq_dist > params_tracker_feature_polyline_->match_landmark_position_sq_norm_max)
        {
            WOLF_WARN("correctFeatureDrift: any point didn't match while checking incoming-lmk back exceeding point: feature point ", exceeding_inc, " - landmark point ", exceeding_lmk);
            return false;
        }
        // add exceeding point to the landmark match
        else
        {
            lmk_inc_match->feature_to_id_  = exceeding_inc;
            lmk_inc_match->landmark_to_id_ = exceeding_lmk;
        }
    }

    return true;
}

bool ProcessorTrackerFeaturePolyline2D::voteForKeyFrame()
{
    //WOLF_DEBUG("PTFP ", getName(), "::voteForKeyFrame: ");
    // TODO
    /* IDEAS:
     *   - Any feature incoming derived too much from projected Landmark
     *   - Any feature incoming overlap with Landmark less than param (1 or 2) defined points
     *   - Any feature last would define extreme or add points to the Landmark
     *   - Some (more than param) tracks lost
     */
    return false;
}

unsigned int ProcessorTrackerFeaturePolyline2D::processNew(const int& _max_features)
{
	WOLF_DEBUG("PTFP ", getName(), "::processNew: ");

    // PT::processNew() ==========================================================
    unsigned int n = ProcessorTrackerFeature::processNew(_max_features); // implicit call to detectNewFeatures() and trackFeatures()

    WOLF_DEBUG("Processing ", n, " new last features");

    // prior transformations
    Eigen::Matrix3s T_sensor_world_last(Eigen::Matrix3s::Identity());
    T_sensor_world_last.topLeftCorner(2,2)  = R_sensor_world_last_;
    T_sensor_world_last.topRightCorner(2,1) = t_sensor_world_last_;

    // MATCH or EMPLACE LANDMARKS ==========================================================
    // For each new last feature: Either emplace a landmark or match with an existent landmark
    for (auto ftr : new_features_last_)
    {
        // not matched feature
        assert(landmark_match_map_.count(ftr) == 0 && "feature in new_features_last is already in landmark_match_map_");

        auto pl_ftr = std::static_pointer_cast<FeaturePolyline2D>(ftr);
        LandmarkMatchPolyline2DScalarMap best_lmk_matches;
        // WOLF_DEBUG("Processing feature: ", pl_ftr->id());

        // SEARCH: Check matching with all existing landmarks
        for (auto lmk : getProblem()->getMap()->getLandmarkList())
            // Check only polyline landmarks that hasn't been just emplaced
            if (lmk->getType() == "POLYLINE 2D" && new_landmarks_.count(lmk) == 0)
                // Store all matches consistent with T_sensor_world_last in best_lmk_matches sorted by squared distance
                tryMatchWithLandmark(best_lmk_matches, std::static_pointer_cast<LandmarkPolyline2D>(lmk), pl_ftr, T_sensor_world_last);

        // EMPLACE: If not matched with any existing landmark: emplace a new one
        if (best_lmk_matches.empty())
        {
            // emplace a landmark
            auto new_lmk_ptr = emplaceLandmark(pl_ftr);

            // Add a new match to landmark_match_map_
            auto new_lmk_match = std::make_shared<LandmarkMatchPolyline2D>();
            new_lmk_match->landmark_ptr_ = new_lmk_ptr;
            new_lmk_match->pl_landmark_ = std::static_pointer_cast<LandmarkPolyline2D>(new_lmk_ptr);
            new_lmk_match->pl_feature_ = pl_ftr;
            new_lmk_match->landmark_version_ = std::static_pointer_cast<LandmarkPolyline2D>(new_lmk_ptr)->getVersion();
            new_lmk_match->feature_from_id_= 0;                         // all points match
            new_lmk_match->landmark_from_id_ = 0;                       // all points match
            new_lmk_match->feature_to_id_= pl_ftr->getNPoints()-1;      // all points match
            new_lmk_match->landmark_to_id_ = pl_ftr->getNPoints()-1;    // all points match
            new_lmk_match->normalized_score_ = 1.0;                     // max score
            new_lmk_match->T_feature_landmark_ = Eigen::Matrix3s::Identity();
            new_lmk_match->T_feature_landmark_.topLeftCorner(2,2) = R_sensor_world_last_;
            new_lmk_match->T_feature_landmark_.topRightCorner(2,1) = t_sensor_world_last_;

            WOLF_ERROR_COND((new_lmk_match->computeSqDistArray() >= Constants::EPS).any(),
                            "Created a landmark not perfectly aligned with feature: ",
                            new_lmk_match->computeSqDistArray().transpose() );
            assert((new_lmk_match->computeSqDistArray() < Constants::EPS).all());

            // Store in new_landmarks
            new_landmarks_.insert(new_lmk_ptr);

            // store match in map
            landmark_match_map_[pl_ftr] = new_lmk_match;

            WOLF_DEBUG("New landmark ", new_lmk_ptr->id(), " created. Matching with feature ", pl_ftr->id());

            //WOLF_WARN("New match, last with new landmark:");
            //new_lmk_match->print();
        }
        // MATCH
        else
        {
            // store best match (highest score) in map
            landmark_match_map_[pl_ftr] = best_lmk_matches.rbegin()->second;

            //WOLF_DEBUG("New match, last with existing landmark:");
            //landmark_match_map_[pl_ftr]->print();

            // store landmark candidates to be merged
            if (best_lmk_matches.size() > 1)
            {
                LandmarkPolyline2DPtrList merge_candidates;
                for (auto lmk_match_pair_it : best_lmk_matches)
                    merge_candidates.push_back(lmk_match_pair_it.second->pl_landmark_);
                merge_candidates_list_.push_back(merge_candidates);
            }
        }
    }

    // INCOMING-LANDMARK MATCH ==========================================================
    // store in landmark_match_map_ the incoming features
    auto ftr_it = new_features_incoming_.begin();
    while (ftr_it != new_features_incoming_.end())
    {
        auto pl_incoming = std::static_pointer_cast<FeaturePolyline2D>(*ftr_it);
        assert(matches_last_from_incoming_.count(*ftr_it) && "last-incoming match not found");
        auto pl_ftr_match = std::static_pointer_cast<FeatureMatchPolyline2D>(matches_last_from_incoming_[*ftr_it]);
        assert(landmark_match_map_.count(pl_ftr_match->feature_ptr_) && "last-lmk match not found");

        // try to concatenate incoming-last + last-landmark into incoming-landmark
        auto pl_lmk_match_incoming = concatenateFeatureLandmarkMatches(pl_ftr_match,
                                                                       landmark_match_map_[pl_ftr_match->feature_ptr_]);

        // Matched
        if (pl_lmk_match_incoming != nullptr)
        {
        	WOLF_DEBUG("New match, incoming with new landmark:");
        	//pl_lmk_match_incoming->print();
            assert(landmark_match_map_.count(pl_incoming) == 0);
            landmark_match_map_[pl_incoming] = pl_lmk_match_incoming;
            ftr_it++;
        }
        // incoming-landmark did not match
        else
        {
            WOLF_DEBUG("PTFP ", getName(), "::processNew: incoming-last + last-landmark could not be concatenated into incoming-landmark");
            matches_last_from_incoming_.erase(pl_incoming);
            // remove feature (unlink not allowed, we could create a new one via copy..)
            pl_incoming->remove();
            ftr_it = new_features_incoming_.erase(ftr_it);
        }
    }
    return n;
}

unsigned int ProcessorTrackerFeaturePolyline2D::detectNewFeatures(const int& _max_new_features,
                                                                  const CaptureBasePtr& _capture,
                                                                  FeatureBasePtrList& _features_out)
{
	WOLF_DEBUG("PTFP ", getName(), "::detectNewFeatures (", untracked_features_last_.size(), " in untracked_features_last_)");

    // move all already created features (in preProcess())
    _features_out = std::move(untracked_features_last_);

    // resize to max_new_features
    if (_max_new_features != -1 && _features_out.size() > _max_new_features)
        _features_out.resize(_max_new_features);

    // link detected features (only created in preProcess())
    for (auto ftr : _features_out)
    	ftr->link(_capture);

    // WOLF_DEBUG(_features_out.size(), " were provided");

    return _features_out.size();
}

void ProcessorTrackerFeaturePolyline2D::establishFactors()
{
	WOLF_DEBUG("PTFP ", getName(), "::establishFactors: ");
    unsigned int N_factors = 0;

    // Create a factor for each match in last features
    FeatureBasePtrList last_features = last_ptr_->getFeatureList();
    for (auto ftr : last_features)
    {
        assert(landmark_match_map_.count(ftr) && "feature without landmark match in last features");
        auto lmk_match = landmark_match_map_[ftr];
        auto pl_lmk = lmk_match->pl_landmark_;
        auto pl_ftr = lmk_match->pl_feature_;
        assert(pl_ftr->id() == ftr->id());
        WOLF_DEBUG("Establishing factors for last feature: ", ftr->id(), " and landmark ", pl_lmk->id());

        // DEPRECATED MATCH: update match
        if (lmk_match->isDeprecated())
        {
            WOLF_DEBUG("\tLandmark was changed, merged or removed, trying to update...");
            // not successful update
            if (!updateLandmarkMatch(lmk_match))
            {
                WOLF_DEBUG("\tUpdate didn't success, removing feature...");
                // remove from match map
                landmark_match_map_.erase(ftr);
                // remove feature
                ftr->remove();
                continue;
            }
            else // landmark could be merged into another one, update the pointer
            	pl_lmk = lmk_match->pl_landmark_;
        }

        // CHECK
        if (!lmk_match->check())
		{
        	WOLF_DEBUG("\tMatch check not passed, removing feature...");
			// remove from match map
			landmark_match_map_.erase(ftr);
			// remove feature
			ftr->remove();
			continue;
        }

        // MODIFY LANDMARK (only for not closed and not recently emplaced)
        // WOLF_DEBUG("\tModifying..");
        if (!pl_lmk->isClosed() && new_landmarks_.count(pl_lmk) == 0)
        {
        	// If modified, add lmk to modified_lmks_ list
            if (modifyLandmarkAndMatch(lmk_match))
            	modified_lmks_.push_back(pl_lmk);

            //std::cout << "after modifyLandmarkAndMatchlandmark " << pl_lmk->id() << (pl_lmk->isClosed() ? " is closed" : " is new") << " matched with feature " << ftr->id() << std::endl;

            assert(lmk_match->feature_from_id_ == 0 && "Landmark didn't grow properly");
            assert(lmk_match->feature_to_id_ == pl_ftr->getNPoints()-1 && "Landmark didn't grow properly");
            assert(pl_lmk->getNPoints() >= pl_ftr->getNPoints() && "Landmark didn't grow properly");
            assert(lmk_match->check() && "Landmark didn't grow properly");
        }
        //else
        //	std::cout << "landmark " << pl_lmk->id() << (pl_lmk->isClosed() ? " is closed" : " is new") << " matched with feature " << ftr->id() << std::endl;

        // ESTABLISH FACTORS
        // WOLF_DEBUG("\tEstablishing factors..");
        // checks
        assert(lmk_match->feature_from_id_ == 0 && "Landmark didn't grow properly");
        assert(lmk_match->feature_to_id_ == pl_ftr->getNPoints()-1 && "Landmark didn't grow properly");
        assert(pl_lmk->getNPoints() >= pl_ftr->getNPoints() && "Landmark didn't grow properly");
        assert(lmk_match->check() && "Landmark didn't grow properly");

        // std::cout << "Establishing factor between" << std::endl;
        // std::cout << "\tfeature " << pl_ftr->id() << " is liked to capture? " << (pl_ftr->getCapture() ? pl_ftr->getCapture()->id() : 9999999999999999) << " is linked to problem? " << (pl_ftr->getProblem() ? "YES" : "NO") << std::endl;
        // std::cout << "\tlandmark " << pl_lmk->id() << std::endl;
        // std::cout << "\tmatch from feature point " << lmk_match->feature_from_id_ << std::endl;
        // std::cout << "\tmatch to feature point " << lmk_match->feature_to_id_ << std::endl;
        // std::cout << "\tmatch from landmark point " << lmk_match->landmark_from_id_ << std::endl;
        // std::cout << "\tmatch to landmark point " << lmk_match->landmark_to_id_ << std::endl;

        // Factors for all feature points
        int lmk_point_id = lmk_match->landmark_from_id_;
        for (int ftr_point_id = 0; ftr_point_id < pl_ftr->getNPoints(); ftr_point_id++)
        {
            // std::cout << "feature point " << ftr_point_id << std::endl;
            // std::cout << "landmark point " << lmk_point_id << std::endl;

            // First not defined point
            if (ftr_point_id == 0 && !pl_ftr->isFirstDefined())
                // first point to line factor
            {
                // std::cout << "point-line: landmark points " << lmk_point_id << ", " << pl_lmk->getNextValidId(lmk_point_id) << std::endl;
                assert(!(lmk_point_id == pl_lmk->getLastId() && !pl_lmk->isClosed()) && "Trying to establish a factor for first not defined point of a feature with the last landmark point without being closed. A landmark didn't grow correctly.");

                // emplace factor
                emplaceFactorPointToLine(pl_ftr, pl_lmk, ftr_point_id, lmk_point_id, pl_lmk->getNextValidId(lmk_point_id));
                N_factors++;
                // std::cout << "factor added" << std::endl;
            }

            // Last not defined point
            else if (ftr_point_id == pl_ftr->getNPoints()-1 && !pl_ftr->isLastDefined())
                // last point to line factor
            {
                // std::cout << "point-line: landmark points " << lmk_point_id << ", " << pl_lmk->getPrevValidId(lmk_point_id) << std::endl;
                assert(!(lmk_point_id == pl_lmk->getFirstId() && !pl_lmk->isClosed()) && "Trying to establish a factor for last not defined point of a feature with the first landmark point without being closed. A landmark didn't grow correctly.");

                // emplace factor
                emplaceFactorPointToLine(pl_ftr, pl_lmk, ftr_point_id, lmk_point_id, pl_lmk->getPrevValidId(lmk_point_id));
                N_factors++;
                // std::cout << "factor added" << std::endl;
            }

            // Defined point
            else
                // point to point factor
            {
                // std::cout << "point-point: landmark point " << lmk_point_id << std::endl;
                //std::cout << "landmark first id:" << pl_lmk->getFirstId() << std::endl;
                //std::cout << "landmark last id:" << pl_lmk->getLastId() << std::endl;
                //std::cout << "landmark n points:" << pl_lmk->getNPoints() << std::endl;
                emplaceFactorPoint(pl_ftr, pl_lmk, ftr_point_id, lmk_point_id);
                assert(pl_lmk->getHits() != 0);
                N_factors++;
                // std::cout << "factor added" << std::endl;
            }

            // Next lmk point
            if (ftr_point_id < pl_ftr->getNPoints()-1)
                lmk_point_id=pl_lmk->getNextValidId(lmk_point_id);
        }
        WOLF_DEBUG("landmark ", pl_lmk->id(), " has now ", pl_lmk->getHits(), " hits");
    }
    // std::cout << N_factors << " factors established" << std::endl;
}

void ProcessorTrackerFeaturePolyline2D::emplaceFactorPointToLine(FeaturePolyline2DPtr _polyline_feature,
                                                                 LandmarkPolyline2DPtr _polyline_landmark,
                                                                 const int& _ftr_point_id,
                                                                 const int& _lmk_point_id,
                                                                 const int& _lmk_prev_point_id)
{
    assert(_polyline_landmark->isValidId(_lmk_point_id) && _polyline_landmark->isValidId(_lmk_prev_point_id));

    FactorBasePtr new_fac = FactorBase::emplace<FactorPointToLine2D>(_polyline_feature,
                                                                     _polyline_feature,
                                                                     _polyline_landmark,
                                                                     shared_from_this(),
                                                                     _ftr_point_id,
                                                                     _lmk_point_id,
                                                                     _lmk_prev_point_id);
}

void ProcessorTrackerFeaturePolyline2D::emplaceFactorPoint(FeaturePolyline2DPtr _polyline_feature,
                                                               LandmarkPolyline2DPtr _polyline_landmark,
                                                               const int& _ftr_point_id,
                                                               const int& _lmk_point_id)
{
    assert(_polyline_landmark->isValidId(_lmk_point_id));

    FactorBasePtr new_fac = FactorBase::emplace<FactorPoint2D>(_polyline_feature,
                                                               _polyline_feature,
                                                               _polyline_landmark,
                                                               shared_from_this(),
                                                               _ftr_point_id,
                                                               _lmk_point_id);
}

LandmarkBasePtr ProcessorTrackerFeaturePolyline2D::emplaceLandmark(FeatureBasePtr _feature_ptr)
{
    WOLF_DEBUG("PTFP ", getName(), "::emplaceLandmark: ");
    //std::cout << "Robot global pose: " << t_world_robot_.transpose() << std::endl;
    //std::cout << "Sensor global pose: " << t_world_sensor_.transpose() << std::endl;
    assert(_feature_ptr->getType() == "POLYLINE 2D");
    assert(std::find(last_ptr_->getFeatureList().begin(),last_ptr_->getFeatureList().end(),_feature_ptr) != last_ptr_->getFeatureList().end() && "creating a landmark for a feature which is not in last capture");


    FeaturePolyline2DPtr polyline_ptr = std::static_pointer_cast<FeaturePolyline2D>(_feature_ptr);
    // compute feature global pose
    Eigen::MatrixXs points_global = R_world_sensor_last_ * polyline_ptr->getPoints().topRows<2>() +
                                    t_world_sensor_last_ * Eigen::VectorXs::Ones(polyline_ptr->getNPoints()).transpose();

    //std::cout << "Feature local points: " << std::endl << polyline_ptr->getPoints().topRows<2>() << std::endl;
    //std::cout << "Landmark global points: " << std::endl << points_global << std::endl;
    //std::cout << "New landmark: extremes defined " << polyline_ptr->isFirstDefined() << " "<< polyline_ptr->isLastDefined() << std::endl;

    // Create new landmark
    return LandmarkBase::emplace<LandmarkPolyline2D>(getProblem()->getMap(),
                                                     std::make_shared<StateBlock>(Eigen::Vector2s::Zero(), true),
                                                     std::make_shared<StateBlock>(Eigen::Vector1s::Zero(), true),
                                                     points_global,
                                                     polyline_ptr->isFirstDefined(),
                                                     polyline_ptr->isLastDefined());
    //std::cout << "done" << std::endl;
}

bool ProcessorTrackerFeaturePolyline2D::modifyLandmarkAndMatch(LandmarkMatchPolyline2DPtr& lmk_match)
{
	WOLF_DEBUG("PTFP ", getName(), "::modifyLandmarkAndMatch: ");
    //lmk_match->print();
    auto pl_lmk = lmk_match->pl_landmark_;
    auto pl_ftr = lmk_match->pl_feature_;
    bool print = false;
    bool modified = false;

    assert(lmk_match->check() && "Bad landmark match provided");

    Eigen::MatrixXs points_global2 = R_world_sensor_last_ * pl_ftr->getPoints().topRows<2>() +
                                     t_world_sensor_last_ * Eigen::VectorXs::Ones(pl_ftr->getNPoints()).transpose();
    Eigen::MatrixXs points_global = lmk_match->T_feature_landmark_.inverse() * pl_ftr->getPoints();
    if (print)
    {
    	std::cout << "------ modifying landmark: " << pl_lmk->id() << "matched with feature " << pl_ftr->id() << std::endl;
        std::cout << "\tlandmark points" << std::endl << pl_lmk->computePointsGlobal() << std::endl;
        std::cout << "\tfeature points_global " << std::endl << points_global << std::endl;
        //std::cout << "points_global2 " << std::endl << points_global2 << std::endl;
        std::cout << "\tlandmark points local" << std::endl << lmk_match->T_feature_landmark_ * pl_lmk->computePointsGlobal() << std::endl;
        std::cout << "\tfeature points " << std::endl << pl_ftr->getPoints() << std::endl;
        //std::cout << "\tfeature local global" << std::endl << lmk_match->T_feature_landmark_ * points_global << std::endl;
        std::cout << "\tlmk_match->T_feature_landmark_" << std::endl << lmk_match->T_feature_landmark_ << std::endl;
    }

    // MODIFY LANDMARK
    // -----------------GROW Front-----------------
    // Add new front points (if any feature point not matched)
    if ( lmk_match->feature_from_id_ > 0 ) // && !pl_lmk->isClosed()
    {
        assert(!pl_lmk->isClosed() && "feature has not matched points in a closed landmark");
        if (print)
        {
            std::cout << "Add new front points. Defined: " << pl_ftr->isFirstDefined() << std::endl;
            std::cout << "\tfeat from " << lmk_match->feature_from_id_ << std::endl;
            std::cout << "\tfeat to   " << lmk_match->feature_to_id_ << std::endl;
            std::cout << "\tland from " << lmk_match->landmark_from_id_ << std::endl;
            std::cout << "\tland to   " << lmk_match->landmark_to_id_ << std::endl;
        }
        pl_lmk->addPoints(points_global, // points matrix in global coordinates
                          lmk_match->feature_from_id_-1, // last feature point index to be added
                          pl_ftr->isFirstDefined(), // is defined
                          false); // front

        lmk_match->landmark_from_id_ = pl_lmk->getFirstId();
        lmk_match->feature_from_id_ = 0;
        lmk_match->landmark_version_ = pl_lmk->getVersion();
        modified = true;
        if (print)
        {
            std::cout << "\tfeat from " << lmk_match->feature_from_id_ << std::endl;
            std::cout << "\tfeat to   " << lmk_match->feature_to_id_ << std::endl;
            std::cout << "\tland from " << lmk_match->landmark_from_id_ << std::endl;
            std::cout << "\tland to   " << lmk_match->landmark_to_id_ << std::endl;
        }

        // checks
        assert(lmk_match->feature_from_id_ == 0 && "Landmark didn't grow properly");
        assert(lmk_match->landmark_from_id_ == pl_lmk->getFirstId() && "Landmark didn't grow properly");
        assert(lmk_match->check() && "Landmark didn't grow properly");
    }
    // -----------------CHANGE Front-----------------
    // Change first not defined point if landmark first not defined matched with first feature point that:
    // a. is defined
    // b. would extend the line (check if angle is bigger than 90º)
    else if (lmk_match->landmark_from_id_ == pl_lmk->getFirstId() && !pl_lmk->isFirstDefined()) // && pl_match->feature_from_id_ == 0
    {
        if (// case a
            pl_ftr->isFirstDefined() ||
            // case b
            (points_global.topLeftCorner(2,1)-pl_lmk->getPointVector(pl_lmk->getNextValidId(pl_lmk->getFirstId()))).squaredNorm() >
            (points_global.topLeftCorner(2,1)-pl_lmk->getPointVector(pl_lmk->getFirstId())).squaredNorm() +
            (pl_lmk->getPointVector(pl_lmk->getNextValidId(pl_lmk->getFirstId()))-pl_lmk->getPointVector(pl_lmk->getFirstId())).squaredNorm())
        {
            if (print)
            {
                std::cout << "Change first point. Defined: " << pl_ftr->isFirstDefined() << std::endl;
                std::cout << "\tpoint " << pl_lmk->getFirstId() << ": " << pl_lmk->getPointVector(pl_lmk->getFirstId()).transpose() << std::endl;
                std::cout << "\tpoint " << pl_lmk->getFirstId()+1 << ": " << pl_lmk->getPointVector(pl_lmk->getNextValidId(pl_lmk->getFirstId())).transpose() << std::endl;
                std::cout << "\tfeature point: " << points_global.topLeftCorner(2,1).transpose() << std::endl;
            }

            pl_lmk->setFirst(points_global.col(0), pl_ftr->isFirstDefined());
            modified = true;
            lmk_match->landmark_version_ = pl_lmk->getVersion();

            // checks
            assert(lmk_match->feature_from_id_ == 0 && "Landmark didn't grow properly");
            assert(lmk_match->landmark_from_id_ == pl_lmk->getFirstId() && "Landmark didn't grow properly");
            assert(lmk_match->check() && "Landmark didn't grow properly");
        }
    }
    // -----------------GROW Back-----------------
    // Add new back points (if any feature point not matched)
    if (lmk_match->feature_to_id_ < pl_ftr->getNPoints()-1)
    {
        assert(!pl_lmk->isClosed() && "feature not matched points in a closed landmark");
        if (print)
        {
            std::cout << "Add back points. Defined: " << pl_ftr->isLastDefined() << std::endl;
            std::cout << "\tfeat from " << lmk_match->feature_from_id_ << std::endl;
            std::cout << "\tfeat to   " << lmk_match->feature_to_id_ << std::endl;
            std::cout << "\tland from " << lmk_match->landmark_from_id_ << std::endl;
            std::cout << "\tland to   " << lmk_match->landmark_to_id_ << std::endl;
        }
        pl_lmk->addPoints(points_global, // points matrix in global coordinates
                          lmk_match->feature_to_id_+1, // last feature point index to be added
                          pl_ftr->isLastDefined(), // is defined
                          true); // back

        lmk_match->landmark_to_id_ = pl_lmk->getLastId();
        lmk_match->feature_to_id_ = pl_ftr->getNPoints()-1;
        lmk_match->landmark_version_ = pl_lmk->getVersion();
        modified = true;
        if (print)
        {
            std::cout << "\tfeat from " << lmk_match->feature_from_id_ << std::endl;
            std::cout << "\tfeat to   " << lmk_match->feature_to_id_ << std::endl;
            std::cout << "\tland from " << lmk_match->landmark_from_id_ << std::endl;
            std::cout << "\tland to   " << lmk_match->landmark_to_id_ << std::endl;
        }
        // checks
        assert(lmk_match->feature_to_id_ == pl_ftr->getNPoints()-1 && "Landmark didn't grow properly");
        assert(lmk_match->landmark_to_id_ == pl_lmk->getLastId() && "Landmark didn't grow properly");
        assert(lmk_match->check() && "Landmark didn't grow properly");
    }
    // -----------------CHANGE Back-----------------
    // Change last not defined point if landmark last not defined matched with last feature point that:
    // a. is defined
    // b. would extend the line (check if angle is bigger than 90º)
    else if (lmk_match->landmark_to_id_ == pl_lmk->getLastId() && !pl_lmk->isLastDefined()) //&& pl_match->feature_to_id_ == pl_ftr->getNPoints()-1
    {
        if (// case a
            pl_ftr->isLastDefined() ||
            // case b
            (points_global.topRightCorner(2,1)-pl_lmk->getPointVector(pl_lmk->getPrevValidId(pl_lmk->getLastId()))).squaredNorm() >
            (points_global.topRightCorner(2,1)-pl_lmk->getPointVector(pl_lmk->getLastId())).squaredNorm() +
            (pl_lmk->getPointVector(pl_lmk->getPrevValidId(pl_lmk->getLastId()))-pl_lmk->getPointVector(pl_lmk->getLastId())).squaredNorm())
        {
            if (print)
            {
                std::cout << "Change last point. Defined: " << (pl_ftr->isLastDefined() ? 1 : 0) << std::endl;
                std::cout << "\tpoint " << pl_lmk->getLastId() << ": " << pl_lmk->getPointVector(pl_lmk->getLastId()).transpose() << std::endl;
                std::cout << "\tpoint " << pl_lmk->getLastId()-1 << ": " << pl_lmk->getPointVector(pl_lmk->getPrevValidId(pl_lmk->getLastId())).transpose() << std::endl;
                std::cout << "\tfeature point: " << points_global.topRightCorner(2,1).transpose() << std::endl;
            }

            pl_lmk->setLast(points_global.rightCols(1), pl_ftr->isLastDefined());
            lmk_match->landmark_version_ = pl_lmk->getVersion();
            modified = true;

            // checks
            assert(lmk_match->feature_to_id_ == pl_ftr->getNPoints()-1 && "Landmark didn't grow properly");
            assert(lmk_match->landmark_to_id_ == pl_lmk->getLastId() && "Landmark didn't grow properly");
            assert(lmk_match->check() && "Landmark didn't grow properly");
        }
    }
    // all checks
    assert(lmk_match->feature_from_id_ == 0 && "Landmark didn't grow properly");
    assert(lmk_match->feature_to_id_ == pl_ftr->getNPoints()-1 && "Landmark didn't grow properly");
    assert(pl_lmk->getNPoints() >= pl_ftr->getNPoints() && "Landmark didn't grow properly");
    assert(lmk_match->check() && "Landmark didn't grow properly (should have stopped in a previous assert)");

    //if (print)
    //    lmk_match->print();

//    WOLF_ERROR_COND((lmk_match->computeSqDistArray() >= 0.5).any(),
//                    "landmark not properly aligned with feature: ",
//                    lmk_match->computeSqDistArray().transpose() );
//    assert((lmk_match->computeSqDistArray() < 0.5).all());

    return modified;
}

void ProcessorTrackerFeaturePolyline2D::advanceDerived()
{
	WOLF_DEBUG("PTFP ", getName(), "::advanceDerived: ");
    // remove landmark matches with last features
    if (last_ptr_ != nullptr)
        for (auto ftr : last_ptr_->getFeatureList())
        {
            assert(landmark_match_map_.count(ftr));
            landmark_match_map_.erase(ftr);
        }

    // check all not tracked features are not linked
    assert(std::all_of(untracked_features_last_.begin(), untracked_features_last_.end(), [](FeatureBasePtr f){return f->getCapture() == nullptr;}) &&
           "any linked feature in untracked_features_last_");
    assert(std::all_of(untracked_features_incoming_.begin(), untracked_features_incoming_.end(), [](FeatureBasePtr f){return f->getCapture() == nullptr;}) &&
           "any linked feature in untracked_features_incoming_");

    // remove untracked features
    //WOLF_DEBUG("PTF ", getName(), ": ", "removing ", untracked_features_last_.size() , " features in untracked_features_last_");
    untracked_features_last_.clear();

    // move untracked features incoming to last
    untracked_features_last_.splice(untracked_features_last_.end(),untracked_features_incoming_);
    //WOLF_DEBUG("PTF ", getName(), ": untracked_features_last_ has ", untracked_features_last_.size() , " features (prev. untracked_features_incoming_)");
    //WOLF_DEBUG_COND(last_ptr_!= nullptr, "PTF ", getName(), ": ", last_ptr_->getFeatureList().size(), " in last_ptr_)");

    // clear new_landmarks_
    new_landmarks_.clear();

    ProcessorTrackerFeature::advanceDerived();
}

void ProcessorTrackerFeaturePolyline2D::resetDerived()
{
	WOLF_DEBUG("PTFP ", getName(), "::resetDerived: ");
    // remove landmark matches with last features
    if (last_ptr_ != nullptr)
        for (auto ftr : last_ptr_->getFeatureList())
        {
            assert(landmark_match_map_.count(ftr));
            landmark_match_map_.erase(ftr);
        }

    // check all not tracked features are not linked
    assert(std::all_of(untracked_features_last_.begin(), untracked_features_last_.end(), [](FeatureBasePtr f){return f->getCapture() == nullptr;}) &&
           "any linked feature in untracked_features_last_");
    assert(std::all_of(untracked_features_incoming_.begin(), untracked_features_incoming_.end(), [](FeatureBasePtr f){return f->getCapture() == nullptr;}) &&
           "any linked feature in untracked_features_incoming_");

    // remove untracked features
    //WOLF_DEBUG("PTF ", getName(), ": ", "removing ", untracked_features_last_.size() , " features in untracked_features_last_");
    untracked_features_last_.clear();

    // move untracked features incoming to last
    untracked_features_last_.splice(untracked_features_last_.end(),untracked_features_incoming_);
    //WOLF_DEBUG("PTF ", getName(), ": untracked_features_last_ has ", untracked_features_last_.size() , " features (prev. untracked_features_incoming_)");
    //WOLF_DEBUG_COND(last_ptr_!= nullptr, "PTF ", getName(), ": ", last_ptr_->getFeatureList().size(), " in last_ptr_)");

    // clear new_landmarks_
    new_landmarks_.clear();

    ProcessorTrackerFeature::resetDerived();
}

void ProcessorTrackerFeaturePolyline2D::preProcess()
{
    //WOLF_DEBUG("PTFP ", getName(), "::extractPolylines: ");
    // Extract all polylines from incoming_
    // TODO: sort corners by bearing
    std::list<laserscanutils::Polyline> polylines;

    line_finder_.findPolylines(std::static_pointer_cast<CaptureLaser2D>(incoming_ptr_)->getScan(), (std::static_pointer_cast<SensorLaser2D>(getSensor()))->getScanParams(), polylines);
    WOLF_DEBUG("PTFP ", getName(), " Polylines extracted: ", polylines.size());

    for (auto&& pl : polylines)
    {
        //WOLF_DEBUG("new polyline detected: Defined", pl.first_defined_ , pl.last_defined_ );
        //std::cout << "covs: " << std::endl << pl.covs_ << std::endl;
        auto points = pl.points_;
        points.bottomRows(1) = Eigen::MatrixXs::Ones(1,points.cols());
        untracked_features_incoming_.push_back(std::make_shared<FeaturePolyline2D>(points, pl.covs_, pl.first_defined_, pl.last_defined_));
        WOLF_DEBUG("new polyline in untracked_features_incoming_: ",untracked_features_incoming_.back()->id());
    }

    // compute transformations
    if (last_ptr_ != nullptr && incoming_ptr_ != nullptr)
        computeTransformations();

    //WOLF_DEBUG("PTF ", getName(), ": ", "untracked_features_last_ has ", untracked_features_last_.size() , " features");
    //WOLF_DEBUG_COND(last_ptr_!=nullptr, "PTF ", getName(), ": ", last_ptr_->getFeatureList().size(), " in last_ptr_)");
}

void ProcessorTrackerFeaturePolyline2D::postProcess()
{
	WOLF_DEBUG("PTFP ", getName(), "::postProcess: ");
    // Try to close & classify modified landmarks
    while (!modified_lmks_.empty())
    {
    	WOLF_DEBUG("classifying and closing modified landmarks");
        auto lmk_ptr = modified_lmks_.front();
        modified_lmks_.pop_front();

        assert(!lmk_ptr->isRemoving());

        // try classify
        if (lmk_ptr->getClassification().type == UNCLASSIFIED && lmk_ptr->getNDefinedPoints() >= 3 && lmk_ptr->getNDefinedPoints() <= 4)
            lmk_ptr->tryClassify(params_tracker_feature_polyline_->class_position_error_th, params_tracker_feature_polyline_->polyline_classes);
        // try close
        if (!lmk_ptr->isClosed() && (lmk_ptr->getNDefinedPoints() >= 4 || lmk_ptr->getNPoints() >= 5))
            lmk_ptr->tryClose(params_tracker_feature_polyline_->class_position_error_th);
    }
    WOLF_DEBUG("closing and classify done");

    // Try to merge landmarks
    while (!merge_candidates_list_.empty())
    {
    	WOLF_DEBUG("merging landmarks...");
        // load next list of candidates
        LandmarkPolyline2DPtrList merge_candidates = std::move(merge_candidates_list_.front());
        merge_candidates_list_.pop_front();

        // change already merged lmks with the corresponding remaining lmk
        auto lmk_it = merge_candidates.begin();
        while (lmk_it != merge_candidates.end())
        {
        	if ((*lmk_it)->isRemoving())
        	{
        		lmk_it = merge_candidates.erase(lmk_it);
        		continue;
        	}
			if ((*lmk_it)->getMergedInLandmark() != nullptr)
				(*lmk_it) = (*lmk_it)->getMergedInLandmark();
			lmk_it++;
        }

        // remove repeated lmks
        merge_candidates.sort();
        merge_candidates.unique();

        if (merge_candidates.size() < 2)
            continue;

        // Merge landmarks candidates and accumulate the correspondence of merged to the remaining ones
        //FIXME: special parameter for merging landmarks?
        LandmarkPolyline2D::tryMergeLandmarks(merge_candidates,
                                              params_tracker_feature_polyline_->match_landmark_position_sq_norm_max,
                                              params_tracker_feature_polyline_->match_landmark_overlap_dist_min,
                                              params_tracker_feature_polyline_->match_landmark_overlap_n_min,
                                              params_tracker_feature_polyline_->match_landmark_overlap_n_defined_min);
    }
    WOLF_DEBUG("merging done");

    // Remove "wrong" landmarks
    WOLF_DEBUG("remove 'wrong' landmarks...");
    LandmarkBasePtrList lmk_list = getProblem()->getMap()->getLandmarkList();
    for (auto lmk_it = lmk_list.begin(); lmk_it != lmk_list.end(); lmk_it++)
        if ((*lmk_it)->getType() == "POLYLINE 2D")
        {
            auto points_map = std::static_pointer_cast<LandmarkPolyline2D>(*lmk_it)->getPointStatePtrMap();

            for (auto point_pair = points_map.begin(); point_pair != points_map.end(); point_pair++)
                if (point_pair->second->getState().maxCoeff() > 1e6 || point_pair->second->getState().minCoeff() < -1e6)
                {
                    (*lmk_it)->remove();
                    break;
                }
        }

    WOLF_DEBUG("PTF ", getName(), ": ", "untracked_features_last_ has ", untracked_features_last_.size() , " features");
    //WOLF_DEBUG_COND(last_ptr_!=nullptr,"PTF ", getName(), ": ", last_ptr_->getFeatureList().size(), " in last_ptr_)");
}

void ProcessorTrackerFeaturePolyline2D::tryMatchWithFeature(FeatureMatchPolyline2DScalarMap& ftr_matches,
                                                            const FeaturePolyline2DPtr& _ftr_L,
                                                            const FeaturePolyline2DPtr& _ftr_I,
                                                            const Eigen::Matrix3s& _T_last_incoming_prior) const
{
	WOLF_DEBUG("tryMatchWithFeature: incoming ", _ftr_I->id(), " last ", _ftr_L->id());

    MatchPolyline2DPtr best_match = nullptr;

    if (params_tracker_feature_polyline_->use_probabilistic_match)
    {
        // Compute Covariance of projected points (only considering noise in T_last_incoming) and computing Jacobian using the mean of the points
        Eigen::Vector2s mean_last_points = _ftr_L->getPoints().rowwise().mean().head(2);
        Eigen::Vector2s J_th;
        Scalar cos_th = _T_last_incoming_prior(0,0);
        Scalar sin_th = _T_last_incoming_prior(1,0);
        J_th << -sin_th * mean_last_points(0) - cos_th * mean_last_points(1),
                 cos_th * mean_last_points(0) - sin_th * mean_last_points(1);
        Eigen::Matrix2s Sigma = Eigen::Matrix2s::Identity()*params_tracker_feature_polyline_->match_feature_position_sq_norm_max +
                                J_th*params_tracker_feature_polyline_->match_feature_orientation_sq_norm_max*J_th.transpose();

        // sq root of inverse
        Eigen::Matrix2s sqrtOmega = Eigen::Matrix2s(Sigma.llt().matrixL()).inverse();

        // last expected points
        Eigen::Matrix<Scalar, 3, Eigen::Dynamic> last_expected_points = _T_last_incoming_prior.inverse() * _ftr_L->getPoints();

        // Shape points in the equiprobabilistic space
        Eigen::Matrix<Scalar, 3, Eigen::Dynamic> last_shaped = last_expected_points;
        Eigen::Matrix<Scalar, 3, Eigen::Dynamic> inc_shaped  = _ftr_I->getPoints();
        last_shaped.topRows<2>() = sqrtOmega*last_shaped.topRows<2>();
        inc_shaped.topRows<2>()  = sqrtOmega*inc_shaped.topRows<2>();

        // compute best sq Mahalanobis distance matching
        best_match = computeBestSqDist(inc_shaped, // <--feature incoming points
                                       _ftr_I->isFirstDefined(),
                                       _ftr_I->isLastDefined(),
                                       false,                     // <--feature is not closed for sure
                                       last_shaped, // <--feature last points
                                       _ftr_L->isFirstDefined(),
                                       _ftr_L->isLastDefined(),
                                       false,                     // <--feature is not closed for sure
                                       1,
                                       params_tracker_feature_polyline_->match_feature_overlap_dist_min,
                                       params_tracker_feature_polyline_->match_feature_overlap_n_min,
                                       params_tracker_feature_polyline_->match_feature_overlap_n_defined_min);
    }
    else // not probabilistic match
    {
        // last expected points
        Eigen::Matrix<Scalar, 3, Eigen::Dynamic> last_expected_points = _T_last_incoming_prior.inverse() * _ftr_L->getPoints();

        // compute best sq Mahalanobis distance matching
        best_match = computeBestSqDist(_ftr_I->getPoints(),       // <--feature incoming points
                                       _ftr_I->isFirstDefined(),
                                       _ftr_I->isLastDefined(),
                                       false,                     // <--feature is not closed for sure
                                       last_expected_points,      // <--feature last points
                                       _ftr_L->isFirstDefined(),
                                       _ftr_L->isLastDefined(),
                                       false,                     // <--feature is not closed for sure
                                       params_tracker_feature_polyline_->match_feature_position_sq_norm_max,
                                       params_tracker_feature_polyline_->match_feature_overlap_dist_min,
                                       params_tracker_feature_polyline_->match_feature_overlap_n_min,
                                       params_tracker_feature_polyline_->match_feature_overlap_n_defined_min,
									   true,true, // exceeding points allowed for both features
									   false); //print
    }

    //valid match
    if (best_match != nullptr)
    {
    	WOLF_DEBUG("match found:");

        auto ftr_match = std::make_shared<FeatureMatchPolyline2D>();
        // feature pointers
        ftr_match->feature_ptr_ = _ftr_L;
        ftr_match->pl_last_     = _ftr_L;
        ftr_match->pl_incoming_ = _ftr_I;
        // feature incoming point id's
        ftr_match->feature_incoming_from_id_ = best_match->from_A_id_;
        ftr_match->feature_incoming_to_id_   = best_match->to_A_id_;
        // feature last point id's
        ftr_match->feature_last_from_id_ = best_match->from_B_id_;
        ftr_match->feature_last_to_id_   = best_match->to_B_id_;
        // incoming last transformation
        if (ftr_match->getNDefinedPoints() > 0 && ftr_match->getNPoints() > 1)
        {
            ftr_match->T_incoming_last_ = pose2T2D(computeRigidTrans(_ftr_I->getPoints().middleCols(ftr_match->feature_incoming_from_id_,ftr_match->getNPoints()),
                                                                     _ftr_L->getPoints().middleCols(ftr_match->feature_last_from_id_,ftr_match->getNPoints()),
                                                                     ftr_match->fromDefined(),
                                                                     ftr_match->toDefined()));
            //ftr_match->print();
        }
        else
            ftr_match->T_incoming_last_ = _T_last_incoming_prior;
        // score
        ftr_match->normalized_score_ = best_match->normalized_score_;

        assert(ftr_match->feature_incoming_from_id_ == 0 || ftr_match->feature_last_from_id_ == 0);
        assert(ftr_match->feature_incoming_to_id_ == _ftr_I->getNPoints() -1 || ftr_match->feature_last_to_id_ == _ftr_L->getNPoints() -1);

        // store in list
        ftr_matches[ftr_match->normalized_score_] = ftr_match;
        //WOLF_INFO("match stored:");
        //ftr_match->print();
        assert(ftr_match->check() && "tryMatchWithFeature: wrong match");
    }

}

void ProcessorTrackerFeaturePolyline2D::tryMatchWithLandmark(LandmarkMatchPolyline2DScalarMap& lmk_matches,
                                                             const LandmarkPolyline2DPtr& _lmk_ptr,
                                                             const FeaturePolyline2DPtr& _feat_ptr,
                                                             const Eigen::Matrix3s& _T_feature_landmark_prior) const
{
	WOLF_DEBUG("tryMatchWithLandmark: feature ", _feat_ptr->id(), " landmark ", _lmk_ptr->id());

    MatchPolyline2DPtr best_match = nullptr;

    if (params_tracker_feature_polyline_->use_probabilistic_match)
    {
        // Compute Covariance of projected points (only considering noise in _T_feature_landmark_prior) and computing Jacobian using the mean of the points
        Eigen::Vector2s mean_last_points = _lmk_ptr->computePointsGlobal().rowwise().mean().head(2);
        Eigen::Vector2s J_th;
        Scalar cos_th = _T_feature_landmark_prior(0,0);
        Scalar sin_th = _T_feature_landmark_prior(1,0);
        J_th << -sin_th * mean_last_points(0) - cos_th * mean_last_points(1),
                 cos_th * mean_last_points(0) - sin_th * mean_last_points(1);
        Eigen::Matrix2s Sigma = Eigen::Matrix2s::Identity()*params_tracker_feature_polyline_->match_landmark_position_sq_norm_max +
                                J_th*params_tracker_feature_polyline_->match_landmark_orientation_sq_norm_max*J_th.transpose();

        // sq root of inverse
        Eigen::Matrix2s sqrtOmega = Eigen::Matrix2s(Sigma.llt().matrixL()).inverse();

        // landmark expected points
        Eigen::MatrixXs lmk_expected_points = _T_feature_landmark_prior * _lmk_ptr->computePointsGlobal();

        // Shape points in the equiprobabilistic space
        Eigen::Matrix<Scalar, 3, Eigen::Dynamic> lmk_shaped = lmk_expected_points;
        Eigen::Matrix<Scalar, 3, Eigen::Dynamic> feat_shaped  = _feat_ptr->getPoints();
        lmk_shaped.topRows<2>() = sqrtOmega*lmk_shaped.topRows<2>();
        feat_shaped.topRows<2>()  = sqrtOmega*feat_shaped.topRows<2>();

        // compute best sq Mahalanobis distance matching
        best_match = computeBestSqDist(lmk_shaped, // <--landmark points in local coordinates
                                       _lmk_ptr->isFirstDefined(),
                                       _lmk_ptr->isLastDefined(),
                                       _lmk_ptr->isClosed(),
                                       feat_shaped,             // <--feature last points
                                       _feat_ptr->isFirstDefined(),
                                       _feat_ptr->isLastDefined(),
                                       false,                     // <--feature is not closed for sure
                                       1,
                                       params_tracker_feature_polyline_->match_landmark_overlap_dist_min,
                                       params_tracker_feature_polyline_->match_landmark_overlap_n_min,
                                       params_tracker_feature_polyline_->match_landmark_overlap_n_defined_min,
                                       true, // partial match lmk allowed
                                       false);// partial match ftr allowed
    }
    else // not probabilistic match
    {
        // landmark expected points
        Eigen::MatrixXs lmk_expected_points = _T_feature_landmark_prior * _lmk_ptr->computePointsGlobal();

        // compute best sq distance matching
        best_match = computeBestSqDist(lmk_expected_points,  // <--landmark points in local coordinates
                                       _lmk_ptr->isFirstDefined(),
                                       _lmk_ptr->isLastDefined(),
                                       _lmk_ptr->isClosed(),
                                       _feat_ptr->getPoints(),        // <--feature points
                                       _feat_ptr->isFirstDefined(),
                                       _feat_ptr->isLastDefined(),
                                       false,                      // <--feature is not closed for sure
                                       params_tracker_feature_polyline_->match_landmark_position_sq_norm_max,
                                       params_tracker_feature_polyline_->match_landmark_overlap_dist_min,
                                       params_tracker_feature_polyline_->match_landmark_overlap_n_min,
                                       params_tracker_feature_polyline_->match_landmark_overlap_n_defined_min,
                                       true, // exceeding points in both extremes allowed for landmark
                                       false);// exceeding points in both extremes not allowed for feature
    }

    //valid match
    if (best_match != nullptr)
    {
    	WOLF_DEBUG("match found!");
        auto lmk_match = std::make_shared<LandmarkMatchPolyline2D>();

        // landmark
        lmk_match->landmark_ptr_=_lmk_ptr;
        lmk_match->pl_landmark_=_lmk_ptr;
        lmk_match->landmark_version_=_lmk_ptr->getVersion();
        // landmark point id's
        std::vector<int> lmk_valid_ids = _lmk_ptr->getValidIds();
        lmk_match->landmark_from_id_ = lmk_valid_ids[best_match->from_A_id_];
        lmk_match->landmark_to_id_   = lmk_valid_ids[best_match->to_A_id_];
        // feature
        lmk_match->pl_feature_=_feat_ptr;
        // feature point id's
        lmk_match->feature_from_id_ = best_match->from_B_id_;
        lmk_match->feature_to_id_   = best_match->to_B_id_;

        // feature landmark transformation
        lmk_match->T_feature_landmark_ = _T_feature_landmark_prior;
        tryUpdateMatchTransformation(lmk_match);

        // score
        lmk_match->normalized_score_ = best_match->normalized_score_;

        assert(lmk_match->feature_from_id_ == 0 || !_lmk_ptr->isClosed());
        assert(lmk_match->feature_to_id_ == _feat_ptr->getNPoints()-1 || !_lmk_ptr->isClosed());
        assert(lmk_match->feature_from_id_ == 0 || lmk_match->feature_to_id_ == _feat_ptr->getNPoints()-1);

        // store in list
        if (lmk_match->check())
        {
            lmk_matches[lmk_match->normalized_score_]= lmk_match;
            WOLF_DEBUG("match stored!");
            assert(lmk_match->check() && "tryMatchWithLandmark: wrong match");
        }
    }
}

/* updateLandmarkMatch
 *
 * landmark changed (grew,closed,merged) and the match should be updated
 */
bool ProcessorTrackerFeaturePolyline2D::updateLandmarkMatch(LandmarkMatchPolyline2DPtr& _lmk_match) const
{
	WOLF_DEBUG("PTFP ", getName(), "::updateLandmarkMatch: ");

    auto pl_lmk = _lmk_match->pl_landmark_;
    auto pl_ftr = _lmk_match->pl_feature_;

    // merged: update the pl_lmk pointer
    if (pl_lmk->getMergedInLandmark() != nullptr)
        pl_lmk = pl_lmk->getMergedInLandmark();
    // removed: not valid match
    else if (pl_lmk->isRemoving())
    {
        WOLF_DEBUG("updateLandmarkMatch: landmark was removed, returning false");
    	return false;
    }

    // try match with _feat_ptr again
    LandmarkMatchPolyline2DScalarMap new_lmk_matches;
    tryMatchWithLandmark(new_lmk_matches,pl_lmk,pl_ftr,_lmk_match->T_feature_landmark_);

    // not valid match
    if (new_lmk_matches.empty())
    {
        WOLF_DEBUG("updateLandmarkMatch: did not match with landmark, returning false");
        return false;
    }

    // valid match: overwrite match
    _lmk_match = new_lmk_matches.rbegin()->second; // the last is the one with highest score
    assert(_lmk_match->check() && "updateLandmarkMatch: wrongly updated");

    return true;
}

LandmarkMatchPolyline2DPtr ProcessorTrackerFeaturePolyline2D::concatenateFeatureLandmarkMatches(FeatureMatchPolyline2DPtr ftr_match,
                                                                                                LandmarkMatchPolyline2DPtr lmk_match_last) const
{
    //WOLF_DEBUG("PTFP ", getName(), "::concatenateFeatureLandmarkMatches: ");
    assert(lmk_match_last->check(false) && "input lmk_match_last wrong");
    //WOLF_DEBUG("INCOMING-LAST MATCH:");
    //ftr_match->print();
    //WOLF_DEBUG("LANDMARK-LAST MATCH:");
    //lmk_match_last->print();

    //WOLF_WARN_COND((lmk_match_last->computeSqDistArray() >= 0.5).any(),
    //                "before concatenateFeatureLandmarkMatches too much error: ",
    //                lmk_match_last->computeSqDistArray().transpose() );
    //assert((lmk_match_last->computeSqDistArray() < 0.5).all());

    auto lmk_match_incoming = std::make_shared<LandmarkMatchPolyline2D>();
    auto lmk_ptr = lmk_match_last->pl_landmark_;

    // feature correspondence from&to
    lmk_match_incoming->feature_from_id_  = ftr_match->feature_incoming_from_id_;
    lmk_match_incoming->feature_to_id_    = ftr_match->feature_incoming_to_id_;

    // landmark correspondence from&to
    // Not adding new points nor checks (done in correctFeatureDrift)
    int from_offset = ftr_match->feature_last_from_id_ - lmk_match_last->feature_from_id_;
    int to_offset   = ftr_match->feature_last_to_id_   - lmk_match_last->feature_to_id_;
//    std::cout << "\nfrom_offset:         " << from_offset;
//    std::cout << "\nto_offset:           " << from_offset;
    lmk_match_incoming->landmark_from_id_ = lmk_match_last->landmark_from_id_;
    lmk_match_incoming->landmark_to_id_   = lmk_match_last->landmark_to_id_;
    while (from_offset > 0)
    {
        if (lmk_match_incoming->landmark_from_id_ == lmk_match_incoming->landmark_to_id_ && !lmk_ptr->isClosed())
            return nullptr;
        lmk_match_incoming->landmark_from_id_ = lmk_ptr->getNextValidId(lmk_match_incoming->landmark_from_id_);
        from_offset--;
    }
    while (from_offset < 0)
    {
        lmk_match_incoming->feature_from_id_++;
        from_offset++;
    }
    while (to_offset < 0)
    {
        if (lmk_match_incoming->landmark_from_id_ == lmk_match_incoming->landmark_to_id_ && !lmk_ptr->isClosed())
            return nullptr;
        lmk_match_incoming->landmark_to_id_ = lmk_ptr->getPrevValidId(lmk_match_incoming->landmark_to_id_);
        to_offset++;
    }
    while (to_offset > 0)
    {
        lmk_match_incoming->feature_to_id_--;
        to_offset--;
    }
//    std::cout << "\nlmk_match_incoming:";
//    std::cout << "\n\tincoming_from:    " << lmk_match_incoming->feature_from_id_;
//    std::cout << "\n\tincoming_to:      " << lmk_match_incoming->feature_to_id_;
//    std::cout << "\n\tlmk_from:         " << lmk_match_incoming->landmark_from_id_;
//    std::cout << "\n\tlmk_to:           " << lmk_match_incoming->landmark_to_id_ << std::endl;

    // other match attributes
    lmk_match_incoming->landmark_ptr_     = lmk_ptr;
    lmk_match_incoming->pl_landmark_      = lmk_ptr;
    lmk_match_incoming->landmark_version_ = lmk_ptr->getVersion();
    lmk_match_incoming->pl_feature_       = ftr_match->pl_incoming_;
    lmk_match_incoming->normalized_score_ = ftr_match->normalized_score_;
    lmk_match_incoming->T_feature_landmark_ = ftr_match->T_incoming_last_ * lmk_match_last->T_feature_landmark_;

    assert(lmk_match_incoming->check(false) && "wrongly concatenated"); // check without partial match

    //WOLF_DEBUG("LANDMARK-INCOMING MATCH:");
    //lmk_match_incoming->print();

    //WOLF_WARN_COND((lmk_match_incoming->computeSqDistArray() >= 0.5).any(),
    //                "after concatenateFeatureLandmarkMatches too much error: ",
    //                lmk_match_incoming->computeSqDistArray().transpose() );
    //assert((lmk_match_incoming->computeSqDistArray() < 0.5).all());

    return lmk_match_incoming;
}

bool ProcessorTrackerFeaturePolyline2D::tryUpdateMatchTransformation(LandmarkMatchPolyline2DPtr& lmk_match) const
{
    //WOLF_DEBUG("PTFP ", getName(), "::tryUpdateMatchTransformation: ");
    if (!lmk_match->check())
        return false;

    assert(lmk_match->check());

    if (lmk_match->getNDefinedPoints() > 0 && lmk_match->getNPoints() > 1)
    {
        //auto T_feature_landmark_prev = lmk_match->T_feature_landmark_;
        auto prev_sqdist = lmk_match->computeSqDistArray();

        lmk_match->T_feature_landmark_ = pose2T2D(computeRigidTrans(lmk_match->pl_feature_->getPoints().middleCols(lmk_match->feature_from_id_,lmk_match->getNPoints()),
                                                                    lmk_match->pl_landmark_->computePointsGlobal(lmk_match->landmark_from_id_,lmk_match->landmark_to_id_),
                                                                    lmk_match->fromDefined(),
                                                                    lmk_match->toDefined()));

		//WOLF_WARN_COND((T2pose2D(T_feature_landmark_prev)-T2pose2D(lmk_match->T_feature_landmark_)).head(2).norm() > 0.1 ||
		//			   abs(pi2pi((T2pose2D(T_feature_landmark_prev)-T2pose2D(lmk_match->T_feature_landmark_))(2))) > 0.1,
		//			   "tryUpdateMatchTransformation: lmk_match->T_feature_landmark_ changed significantly. \n\tPrev pose: ",
		//			   T2pose2D(T_feature_landmark_prev).transpose(),
		//			   "\n\tUpdated pose:",
		//			   T2pose2D(lmk_match->T_feature_landmark_).transpose());

        //WOLF_WARN_COND(lmk_match->computeSqDistArray().maxCoeff() > prev_sqdist.maxCoeff() + 1e-9,
        //                "after tryUpdateMatchTransformation more error! ",
        //                "\nprev:", prev_sqdist.transpose(),
		//				"\nnow: ", lmk_match->computeSqDistArray().transpose());

        return true;
    }
    //WOLF_WARN("tryUpdateMatchTransformation: Not enough points: ",lmk_match->getNPoints(), " (should be > 1) or defined points: ", lmk_match->getNDefinedPoints(), " (should be > 0)");
    return false;
}

// MATH ///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void ProcessorTrackerFeaturePolyline2D::computeTransformations()
{
    //WOLF_DEBUG("ProcessorTrackerFeaturePolyline::computeTransformations: ");
    Eigen::Vector3s vehicle_pose_incoming = getProblem()->getState(incoming_ptr_->getTimeStamp());
    Eigen::Vector3s vehicle_pose_last     = getProblem()->getState(last_ptr_->getTimeStamp());
    WOLF_DEBUG("ProcessorTrackerFeaturePolyline::computeTransformations:",
               "\n\tvehicle_pose_incoming: ", vehicle_pose_incoming.transpose(), " at ", incoming_ptr_->getTimeStamp(),
               "\n\tvehicle_pose_last:     ", vehicle_pose_last.transpose(), " at ", last_ptr_->getTimeStamp());

    // world_robot
    Eigen::Matrix2s R_world_robot_incoming = Eigen::Rotation2Ds(vehicle_pose_incoming(2)).matrix();
    Eigen::Matrix2s R_world_robot_last     = Eigen::Rotation2Ds(vehicle_pose_last(2)).matrix();

    // robot_sensor (to be computed once if extrinsics are fixed and not dynamic)
    if (getSensor()->isExtrinsicDynamic() ||
        !getSensor()->getP()->isFixed() ||
        !getSensor()->getO()->isFixed() ||
        !extrinsics_transformation_computed_)
    {
        t_robot_sensor_ = getSensor()->getP()->getState();
        R_robot_sensor_ = Eigen::Rotation2Ds(getSensor()->getO()->getState()(0)).matrix();
        extrinsics_transformation_computed_ = true;
    }

    // INCOMING
    // global_sensor
    R_world_sensor_incoming_ = R_world_robot_incoming * R_robot_sensor_;
    t_world_sensor_incoming_ = vehicle_pose_incoming.head<2>() + R_world_robot_incoming * t_robot_sensor_;

    // sensor_global
    R_sensor_world_incoming_ = R_world_sensor_incoming_.transpose();
    t_sensor_world_incoming_ = -R_sensor_world_incoming_ * t_world_sensor_incoming_;

    // LAST
    // global_sensor
    R_world_sensor_last_ = R_world_robot_last * R_robot_sensor_;
    t_world_sensor_last_ = vehicle_pose_last.head<2>() + R_world_robot_last * t_robot_sensor_;

    // sensor_global
    R_sensor_world_last_ = R_world_sensor_last_.transpose();
    t_sensor_world_last_ = -R_sensor_world_last_ * t_world_sensor_last_;

    // INCOMING-LAST
    R_incoming_last_ = R_sensor_world_incoming_ * R_world_sensor_last_;
    t_incoming_last_ = t_sensor_world_incoming_ + R_sensor_world_incoming_ * t_world_sensor_last_;

    // last_incoming
    R_last_incoming_ = R_world_sensor_last_.transpose() * R_world_sensor_incoming_;
    t_last_incoming_ = t_sensor_world_last_ + R_sensor_world_last_ * t_world_sensor_incoming_;

    //std::cout << "t_robot_sensor_ " << t_robot_sensor_.transpose() << std::endl;
    //std::cout << "R_robot_sensor_ " << std::endl << R_robot_sensor_ << std::endl;
    //std::cout << "t_world_sensor_incoming_ " << t_world_sensor_incoming_.transpose() << std::endl;
    //std::cout << "R_world_sensor_incoming_ " << std::endl << R_world_sensor_incoming_ << std::endl;
    //std::cout << "t_sensor_world_incoming_ " << t_sensor_world_incoming_.transpose() << std::endl;
    //std::cout << "R_sensor_world_incoming_ " << std::endl << R_sensor_world_incoming_ << std::endl;
    //std::cout << "t_world_sensor_last_ " << t_world_sensor_last_.transpose() << std::endl;
    //std::cout << "R_world_sensor_last_ " << std::endl << R_world_sensor_last_ << std::endl;
    //std::cout << "t_sensor_world_last_ " << t_sensor_world_last_.transpose() << std::endl;
    //std::cout << "R_sensor_world_last_ " << std::endl << R_sensor_world_last_ << std::endl;
    //std::cout << "t_incoming_last_ " << t_incoming_last_.transpose() << std::endl;
    //std::cout << "R_incoming_last_ " << std::endl << R_incoming_last_ << std::endl;
    //std::cout << "t_last_incoming_ " << t_last_incoming_.transpose() << std::endl;
    //std::cout << "R_last_incoming_ " << std::endl << R_last_incoming_ << std::endl;
}


} /* namespace wolf */

// Register in the SensorFactory
#include "core/processor/processor_factory.h"
namespace wolf {
    WOLF_REGISTER_PROCESSOR("TRACKER FEATURE POLYLINE", ProcessorTrackerFeaturePolyline2D);
    WOLF_REGISTER_PROCESSOR_AUTO("TRACKER FEATURE POLYLINE", ProcessorTrackerFeaturePolyline2D);
} // namespace wolf
