#include <core/utils/utils_gtest.h>
#include "laser/landmark/landmark_polyline_2D.h"

using namespace wolf;
using namespace Eigen;

class LandmarkPolylineTest : public testing::Test
{
    public:

        LandmarkPolyline2DPtr lmk_ptr;
        std::vector<LandmarkPolyline2DPtr> lmks_class_list;
        std::vector<VectorXs> lmks_groundtruth;
        std::vector<MatrixXs> points_groundtruth;
        std::vector<PolylineRectangularClass> classification_groundtruth;
        std::vector<PolylineRectangularClass> classification_candidates;

        virtual void SetUp()
        {
            classification_candidates.push_back(PolylineClassContainer());
            classification_candidates.push_back(PolylineClassSmallContainer());
            classification_candidates.push_back(PolylineClassPallet());
            classification_candidates.push_back(PolylineRectangularClass(OTHER,45.6, 3.1));

            //std::vector<Scalar> object_L({12, 5.9, 1.2});
            //std::vector<Scalar> object_W({2.345, 2.345, 0.9});
            //std::vector<PolylineClassType> object_class({CONTAINER, SMALL_CONTAINER, PALLET});

            StateBlockPtr lmk_p_ptr = std::make_shared<StateBlock>(Vector2s::Zero());
            StateBlockPtr lmk_o_ptr = std::make_shared<StateBlock>(Vector1s::Zero());

            MatrixXs points = MatrixXs::Zero(2,5);
            points << 1, 1,-1,-1, 1,
                      2,-2,-2, 2, 2;

            lmk_ptr = std::make_shared<LandmarkPolyline2D>(lmk_p_ptr, lmk_o_ptr, points, true, true);

            // classification test
            MatrixXs class_points = MatrixXs::Zero(2,4);

            for (auto configuration = 1; configuration <= 2; configuration++)
                for (auto candidate_class : classification_candidates)
                //for (unsigned int classification = 0; classification < object_L.size(); classification++)
                {
                    const Scalar& L = candidate_class.L;
                    const Scalar& W = candidate_class.W;
                    //const Scalar& L = object_L[classification];
                    //const Scalar& W = object_W[classification];

                    // create trivial points in corresponding configuration
                    if (configuration == 1)
                        class_points << L/2,-L/2,-L/2, L/2,
                                       -W/2,-W/2, W/2, W/2;
                    else
                        class_points <<-L/2,-L/2, L/2, L/2,
                                       -W/2, W/2, W/2,-W/2;


                    // rotate and translate "randomly"
                    Vector3s center_pose; center_pose << 4, -9, M_PI/5;
                    Matrix2s R = Rotation2D<Scalar>(center_pose(2)).matrix();
                    class_points = R * class_points;
                    class_points += center_pose.head<2>() * VectorXs::Ones(4).transpose();

                    // store object and points groundtruth
                    lmks_groundtruth.push_back(center_pose);
                    points_groundtruth.push_back(class_points);

                    // store classification groundtruth
                    classification_groundtruth.push_back(candidate_class);

                    // create a landmark polyline using the points
                    StateBlockPtr p_ptr = std::make_shared<StateBlock>(center_pose.head<2>());
                    StateBlockPtr o_ptr = std::make_shared<StateBlock>(center_pose.tail<1>());
                    lmks_class_list.push_back(std::make_shared<LandmarkPolyline2D>(p_ptr, o_ptr, class_points, true, true));
                }
        }
};

TEST_F(LandmarkPolylineTest, Merge)
{
    ASSERT_EQ(lmk_ptr->getNPoints(), 5);

    lmk_ptr->mergePoints(0, 4);

    ASSERT_EQ(lmk_ptr->getNPoints(), 4);
    ASSERT_EQ(lmk_ptr->getFirstId(), 1);

//  PRINTF("All good at TestTest::DummyTestExample !\n");
}

TEST_F(LandmarkPolylineTest, Close)
{
    ASSERT_TRUE(lmk_ptr->tryClose(0.1));

    ASSERT_EQ(lmk_ptr->getNPoints(), 4);

//  PRINTF("All good at TestTest::DummyTestExample !\n");
}

TEST_F(LandmarkPolylineTest, Classify)
{
    for (unsigned int i = 0; i < lmks_class_list.size(); i++)
    {
        LandmarkPolyline2DPtr lmk_ptr = lmks_class_list[i];

        // test classify
        ASSERT_TRUE(lmk_ptr->tryClassify(0.1, classification_candidates));

        // test classification
        ASSERT_EQ(lmk_ptr->getClassification().type,    classification_groundtruth[i].type);
        ASSERT_EQ(lmk_ptr->getClassification().L,       classification_groundtruth[i].L);
        ASSERT_EQ(lmk_ptr->getClassification().W,       classification_groundtruth[i].W);
        ASSERT_EQ(lmk_ptr->getClassification().D,       classification_groundtruth[i].D);

        // test center pose
        ASSERT_MATRIX_APPROX(lmk_ptr->getState(), lmks_groundtruth[i], 1e-9);

        // test closed
        ASSERT_TRUE(lmk_ptr->isClosed());

        // test points fixed and center unfixed
        ASSERT_FALSE(lmk_ptr->getP()->isFixed());
        ASSERT_FALSE(lmk_ptr->getO()->isFixed());
        for (auto st_pair : lmk_ptr->getPointStatePtrMap())
            ASSERT_TRUE(st_pair.second->isFixed());

        // test points are in the same place (compoping with object pose)
        Matrix2s R = Rotation2D<Scalar>(lmk_ptr->getO()->getState()(0)).matrix();
        VectorXs t = lmk_ptr->getP()->getState();
        for (unsigned int point_i = 0; point_i < 4; point_i++)
            ASSERT_MATRIX_APPROX(R*lmk_ptr->getPointVector(point_i)+t, points_groundtruth[i].col(point_i), 1e-9);
    }

//  PRINTF("All good at TestTest::DummyTestExample !\n");
}

int main(int argc, char **argv)
{
  testing::InitGoogleTest(&argc, argv);
  return RUN_ALL_TESTS();
}
