#include <stdlib.h>
#include <core/utils/utils_gtest.h>
#include "laser/processor/polyline_2D_utils.h"

using namespace wolf;
using namespace Eigen;

class Polyline2DTest : public testing::Test
{
    public:

        Matrix3s T_map_sens1, T_map_sens2, T_map_polyline_center, T_sens1_sens2;
        MatrixXs points_map, points_sens1, points_sens2;

        virtual void SetUp()
        {
        }

        void SetUpPoints(int n, bool first_defined=true, bool last_defined=true)
        {
            // points
            MatrixXs points = MatrixXs::Random(3,n);
            points.bottomRows(1) = MatrixXs::Ones(1,n);
            T_map_polyline_center = pose2T2D(10*Vector2s::Random(), (rand() % 2000*M_PI)*0.001);
            points_map = T_map_polyline_center * points;

            //std::cout << "points_map: \n" << points_map << std::endl;

            // sens1
            T_map_sens1 = pose2T2D(5*Vector2s::Random(), (rand() % 2000*M_PI)*0.001);
            points_sens1 = T_map_sens1.inverse() * points_map;
            // move first point along the segment from the first to the second
            if (!first_defined)
            {
                Scalar rand_scale = (rand() % 2000 + 1)*0.001;
                Eigen::Vector2s new_first_point = points_sens1.col(1).head(2) + (points_sens1.col(0).head(2) - points_sens1.col(1).head(2)) * rand_scale;

                std::cout << "old first point: " << points_sens1.col(0).head(2).transpose() << std::endl;
                std::cout << "new first point: " << new_first_point.transpose() << std::endl;
                std::cout << "scale: " << rand_scale << std::endl;
                std::cout << "div vector: " << ((new_first_point - points_sens1.col(1).head(2)).array() / (points_sens1.col(0).head(2) - points_sens1.col(1).head(2)).array()).transpose() << std::endl;

                points_sens1.col(0).head(2) = new_first_point;
            }
            // move last point along the segment from the last to the second last
            if (!last_defined)
                points_sens1.col(n-1).head(2) = points_sens1.col(n-2).head(2) + (points_sens1.col(n-1).head(2)-points_sens1.col(n-2).head(2)) * (rand() % 2000 + 1)*0.001;
            //std::cout << "T_map_sens1: \n" << T_map_sens1 << std::endl;
            //std::cout << "points_sens1: \n" << points_sens1 << std::endl;

            // sens2
            T_map_sens2 = pose2T2D(5*Vector2s::Random(), (rand() % 2000*M_PI)*0.001);
            points_sens2 = T_map_sens2.inverse() * points_map;
            T_sens1_sens2 = T_map_sens1.inverse() * T_map_sens2;

            //std::cout << "T_map_sens2: \n" << T_map_sens2 << std::endl;
            //std::cout << "points_sens2: \n" << points_sens2 << std::endl;
        }
};

TEST_F(Polyline2DTest, Transformations)
{
    ASSERT_MATRIX_APPROX(Matrix3s::Identity(), pose2T2D(Vector2s::Zero(),0), 1e-6);
    ASSERT_MATRIX_APPROX(Matrix3s::Identity(), pose2T2D(Vector3s::Zero()), 1e-6);
    ASSERT_MATRIX_APPROX(Vector3s::Zero(), T2pose2D(Matrix3s::Identity()), 1e-6);

    Vector3s v_rand, v_subs;
    v_rand << Vector2s::Random(), pi2pi((rand() % 2000*M_PI)*0.001);
    v_subs = v_rand - T2pose2D(pose2T2D(v_rand));
    v_subs(2) = pi2pi(v_subs(2));
    ASSERT_MATRIX_APPROX(v_subs, Vector3s::Zero(), 1e-6);

    Matrix3s T_rand(Matrix3s::Identity());
    T_rand.topLeftCorner(2,2) = Rotation2D<Scalar>(v_rand(2)).matrix();
    T_rand.topRightCorner(2,1) = v_rand.head(2);
    ASSERT_MATRIX_APPROX(T_rand, pose2T2D(v_rand), 1e-6);
}

TEST_F(Polyline2DTest, computeRigidTrans8points)
{
    bool first_defined = true;
    bool last_defined = true;

    // Random points
    for (auto i = 0; i < 1e3; i++)
    {
        SetUpPoints(8, first_defined, last_defined);

        // compute rigid transformation
        Vector3s v_sens1_sens2 = computeRigidTrans(points_sens1, points_sens2, first_defined, last_defined);

        ASSERT_MATRIX_APPROX(pose2T2D(v_sens1_sens2), T_sens1_sens2, 1e-6);
        auto sq_error = computeSqDist(points_sens1,
                                      0,
                                      points_sens1.cols()-1,
                                      first_defined,
                                      last_defined,
                                      pose2T2D(v_sens1_sens2)*points_sens2,
                                      0,
                                      points_sens2.cols()-1,
                                      first_defined,
                                      last_defined,
                                      points_sens1.cols()-1,
                                      points_sens1.cols());
        ASSERT_TRUE((sq_error < 1e-6).all());

        // compute rigid transformation
        Vector3s v_sens1_map = computeRigidTrans(points_sens1, points_map);

        ASSERT_MATRIX_APPROX(pose2T2D(v_sens1_map), T_map_sens1.inverse(), 1e-6);
        sq_error = computeSqDist(points_sens1,
                                  0,
                                  points_sens1.cols()-1,
                                  first_defined,
                                  last_defined,
                                  pose2T2D(v_sens1_map)*points_map,
                                  0,
                                  points_map.cols()-1,
                                  first_defined,
                                  last_defined,
                                  points_sens1.cols()-1,
                                  points_sens1.cols());
        ASSERT_TRUE((sq_error < 1e-6).all());
    }
}

TEST_F(Polyline2DTest, computeRigidTrans2points2defined)
{
    bool first_defined = true;
    bool last_defined = true;

    // Random points
    for (auto i = 0; i < 1e3; i++)
    {
        SetUpPoints(2, first_defined, last_defined);

        // compute rigid transformation
        Vector3s v_sens1_sens2 = computeRigidTrans(points_sens1, points_sens2, first_defined, last_defined);

        ASSERT_MATRIX_APPROX(pose2T2D(v_sens1_sens2), T_sens1_sens2, 1e-6);
        auto sq_error = computeSqDist(points_sens1,
                                      0,
                                      points_sens1.cols()-1,
                                      first_defined,
                                      last_defined,
                                      pose2T2D(v_sens1_sens2)*points_sens2,
                                      0,
                                      points_sens2.cols()-1,
                                      first_defined,
                                      last_defined,
                                      points_sens1.cols()-1,
                                      points_sens1.cols());
        ASSERT_TRUE((sq_error < 1e-6).all());

        // compute rigid transformation
        Vector3s v_sens1_map = computeRigidTrans(points_sens1, points_map);

        ASSERT_MATRIX_APPROX(pose2T2D(v_sens1_map), T_map_sens1.inverse(), 1e-6);
        sq_error = computeSqDist(points_sens1,
                                  0,
                                  points_sens1.cols()-1,
                                  first_defined,
                                  last_defined,
                                  pose2T2D(v_sens1_map)*points_map,
                                  0,
                                  points_map.cols()-1,
                                  first_defined,
                                  last_defined,
                                  points_sens1.cols()-1,
                                  points_sens1.cols());
        ASSERT_TRUE((sq_error < 1e-6).all());
    }
}

TEST_F(Polyline2DTest, computeRigidTrans3points2defined)
{
    bool first_defined = false;
    bool last_defined = true;

    // Random points
    for (auto i = 0; i < 1e3; i++)
    {
        SetUpPoints(3, first_defined, last_defined);

        // compute rigid transformation
        Vector3s v_sens1_sens2 = computeRigidTrans(points_sens1, points_sens2, first_defined, last_defined);

        ASSERT_MATRIX_APPROX(pose2T2D(v_sens1_sens2), T_sens1_sens2, 1e-6);
        auto sq_error = computeSqDist(points_sens1,
                                      0,
                                      points_sens1.cols()-1,
                                      first_defined,
                                      last_defined,
                                      pose2T2D(v_sens1_sens2)*points_sens2,
                                      0,
                                      points_sens2.cols()-1,
                                      first_defined,
                                      last_defined,
                                      points_sens1.cols()-1,
                                      points_sens1.cols());
        ASSERT_TRUE((sq_error < 1e-6).all());

        // compute rigid transformation
        Vector3s v_sens1_map = computeRigidTrans(points_sens1, points_map, first_defined, last_defined);

        ASSERT_MATRIX_APPROX(pose2T2D(v_sens1_map), T_map_sens1.inverse(), 1e-6);
        sq_error = computeSqDist(points_sens1,
                                  0,
                                  points_sens1.cols()-1,
                                  first_defined,
                                  last_defined,
                                  pose2T2D(v_sens1_map)*points_map,
                                  0,
                                  points_map.cols()-1,
                                  first_defined,
                                  last_defined,
                                  points_sens1.cols()-1,
                                  points_sens1.cols());
        ASSERT_TRUE((sq_error < 1e-6).all());
    }
}

TEST_F(Polyline2DTest, computeRigidTrans3points1defined)
{
    bool first_defined = false;
    bool last_defined = false;

    // Random points
    for (auto i = 0; i < 1e3; i++)
    {
        SetUpPoints(3, first_defined, last_defined);

        // compute rigid transformation
        Vector3s v_sens1_sens2 = computeRigidTrans(points_sens1, points_sens2, first_defined, last_defined);

        ASSERT_MATRIX_APPROX(pose2T2D(v_sens1_sens2), T_sens1_sens2, 1e-6);
        auto sq_error = computeSqDist(points_sens1,
                                      0,
                                      points_sens1.cols()-1,
                                      first_defined,
                                      last_defined,
                                      pose2T2D(v_sens1_sens2)*points_sens2,
                                      0,
                                      points_sens2.cols()-1,
                                      first_defined,
                                      last_defined,
                                      points_sens1.cols()-1,
                                      points_sens1.cols());
        ASSERT_TRUE((sq_error < 1e-6).all());

        // compute rigid transformation
        Vector3s v_sens1_map = computeRigidTrans(points_sens1, points_map, first_defined, last_defined);

        ASSERT_MATRIX_APPROX(pose2T2D(v_sens1_map), T_map_sens1.inverse(), 1e-6);
        sq_error = computeSqDist(points_sens1,
                                  0,
                                  points_sens1.cols()-1,
                                  first_defined,
                                  last_defined,
                                  pose2T2D(v_sens1_map)*points_map,
                                  0,
                                  points_map.cols()-1,
                                  first_defined,
                                  last_defined,
                                  points_sens1.cols()-1,
                                  points_sens1.cols());
        ASSERT_TRUE((sq_error < 1e-6).all());
    }
}

TEST_F(Polyline2DTest, computeRigidTrans2points1defined)
{
    bool first_defined = false;
    bool last_defined = true;

    // Random points
    for (auto i = 0; i < 1e3; i++)
    {
        SetUpPoints(2, first_defined, last_defined);

        // compute rigid transformation
        Vector3s v_sens1_sens2 = computeRigidTrans(points_sens1, points_sens2, first_defined, last_defined);

        ASSERT_MATRIX_APPROX(pose2T2D(v_sens1_sens2), T_sens1_sens2, 1e-6);
        auto sq_error = computeSqDist(points_sens1,
                                      0,
                                      points_sens1.cols()-1,
                                      first_defined,
                                      last_defined,
                                      pose2T2D(v_sens1_sens2)*points_sens2,
                                      0,
                                      points_sens2.cols()-1,
                                      first_defined,
                                      last_defined,
                                      points_sens1.cols()-1,
                                      points_sens1.cols());
        ASSERT_TRUE((sq_error < 1e-6).all());

        // compute rigid transformation
        Vector3s v_sens1_map = computeRigidTrans(points_sens1, points_map, first_defined, last_defined);

        ASSERT_MATRIX_APPROX(pose2T2D(v_sens1_map), T_map_sens1.inverse(), 1e-6);
        sq_error = computeSqDist(points_sens1,
                                  0,
                                  points_sens1.cols()-1,
                                  first_defined,
                                  last_defined,
                                  pose2T2D(v_sens1_map)*points_map,
                                  0,
                                  points_map.cols()-1,
                                  first_defined,
                                  last_defined,
                                  points_sens1.cols()-1,
                                  points_sens1.cols());
        ASSERT_TRUE((sq_error < 1e-6).all());
    }
}

TEST_F(Polyline2DTest, computeRigidTrans2points1definedBack)
{
    bool first_defined = true;
    bool last_defined = false;

    // Random points
    for (auto i = 0; i < 1e3; i++)
    {
        SetUpPoints(2, first_defined, last_defined);

        // compute rigid transformation
        Vector3s v_sens1_sens2 = computeRigidTrans(points_sens1, points_sens2, first_defined, last_defined);

        ASSERT_MATRIX_APPROX(pose2T2D(v_sens1_sens2), T_sens1_sens2, 1e-6);
        auto sq_error = computeSqDist(points_sens1,
                                      0,
                                      points_sens1.cols()-1,
                                      first_defined,
                                      last_defined,
                                      pose2T2D(v_sens1_sens2)*points_sens2,
                                      0,
                                      points_sens2.cols()-1,
                                      first_defined,
                                      last_defined,
                                      points_sens1.cols()-1,
                                      points_sens1.cols());
        ASSERT_TRUE((sq_error < 1e-6).all());

        // compute rigid transformation
        Vector3s v_sens1_map = computeRigidTrans(points_sens1, points_map, first_defined, last_defined);

        ASSERT_MATRIX_APPROX(pose2T2D(v_sens1_map), T_map_sens1.inverse(), 1e-6);
        sq_error = computeSqDist(points_sens1,
                                  0,
                                  points_sens1.cols()-1,
                                  first_defined,
                                  last_defined,
                                  pose2T2D(v_sens1_map)*points_map,
                                  0,
                                  points_map.cols()-1,
                                  first_defined,
                                  last_defined,
                                  points_sens1.cols()-1,
                                  points_sens1.cols());
        ASSERT_TRUE((sq_error < 1e-6).all());
    }
}

int main(int argc, char **argv)
{
  testing::InitGoogleTest(&argc, argv);
  return RUN_ALL_TESTS();
}
